from datetime import datetime, timedelta

from django.contrib import admin

from sat_automations.acs_journal.models import ACSJournal
from sat_automations.utils.common.time import SATTimeZones, bound_time
from sat_automations.utils.mixins.admin import ExportCsvMixin

# Register your models here.


class MessageLTCFilter(admin.SimpleListFilter):
    title = "Local Time"
    parameter_name = "message_utc"
    # Now is today at midnight
    now = bound_time(datetime.now(), nearest_day=True)
    yesterday = bound_time(datetime.now() - timedelta(days=1), nearest_day=True)
    thirty_days = bound_time(datetime.now() - timedelta(days=30), nearest_day=True)
    sixty_days = bound_time(datetime.now() - timedelta(days=60), nearest_day=True)

    def lookups(self, request, model_admin):
        return (
            ("now", "Today"),
            ("yesterday", "Yesterday"),
            ("thirty_days", "30 Days"),
            ("sixty_days", "60 Days"),
        )

    def queryset(self, request, queryset):
        if self.value() == "now":
            return queryset.filter(message_utc__gte=self.now)
        if self.value() == "yesterday":
            return queryset.filter(message_utc__gte=self.yesterday, message_utc__lt=self.now)
        if self.value() == "thirty_days":
            return queryset.filter(message_utc__gte=self.thirty_days)
        if self.value() == "sixty_days":
            return queryset.filter(message_utc__gte=self.sixty_days)


class ACSJournalAdmin(ExportCsvMixin, admin.ModelAdmin):
    list_display = [
        "message_type",
        "server_utc_str",
        "message_utc_str",
        "message_ltc",
        "campus_id",
        "card_num",
        "primary_object_name",
        "secondary_object_name",
        "state_code",
        "primary_type",
        "secondary_type",
    ]
    readonly_fields = [
        "message_type",
        "server_utc",
        "message_utc",
        "campus_id",
        "card_num",
        "primary_object_name",
        "primary_object_type",
        "primary_object_uuid",
        "secondary_object_name",
        "secondary_object_type",
        "secondary_object_uuid",
        "state_code",
        "search_vector",
    ]
    list_filter = ["state_code", MessageLTCFilter]
    search_fields = ["campus_id", "search_vector"]

    class Meta:
        model = ACSJournal

    actions = ExportCsvMixin.actions

    def primary_type(self, obj):
        return obj.primary_type()

    def secondary_type(self, obj):
        return obj.secondary_type()

    def server_utc_str(self, obj):
        return obj.server_utc.strftime("%Y-%m-%d %H:%M:%S %Z")

    def message_utc_str(self, obj):
        return obj.message_utc.strftime("%Y-%m-%d %H:%M:%S %Z")

    def message_ltc(self, obj):
        new_dt = obj.message_utc.astimezone(SATTimeZones.EPT.value)
        return new_dt.strftime("%Y-%m-%d %H:%M:%S %Z")

    def get_ordering(self, request):
        return ["-server_utc"]

    def has_module_permission(self, request, obj=None):
        if not request.user.is_staff:
            return False
        return True

    def has_add_permission(self, request, obj=None):
        if not request.user.is_superuser:
            return False
        return True

    def has_delete_permission(self, request, obj=None):
        return False


# admin.site.register(ACSJournal, ACSJournalAdmin)
