import re

from django.contrib.postgres.indexes import (
    BrinIndex,
    GinIndex,  # add the Postgres recommended GIN index
)
from django.contrib.postgres.search import SearchVectorField
from django.db import models
from model_utils.models import TimeStampedModel

from sat_automations.utils.common.time import SATTimeZones

END_OF_DOTTED_PATH = re.compile(r"\.([^.]*$)")

# Create your models here.


class ACSJournal(TimeStampedModel):
    campus_id = models.CharField(max_length=16, null=True, blank=True)
    card_num = models.CharField(max_length=16, null=True, blank=True)
    message_type = models.CharField(max_length=128, null=True, blank=True)
    # Message UTC is the time the message was created by the panel. This is the time the event occurred.
    message_utc = models.DateTimeField()
    # Server UTC is the time the message was received by the CCURE server.
    # This dt has microseconds and is probably the most accurate time we have, for doing repairs.
    server_utc = models.DateTimeField(null=True, blank=True)
    state_code = models.CharField(max_length=32, null=True, blank=True)
    primary_object_name = models.CharField(max_length=256, null=True, blank=True)
    primary_object_type = models.CharField(max_length=128, null=True, blank=True)
    primary_object_uuid = models.UUIDField(null=True, blank=True)
    secondary_object_name = models.CharField(max_length=256, null=True, blank=True)
    secondary_object_type = models.CharField(max_length=128, null=True, blank=True)
    secondary_object_uuid = models.UUIDField(null=True, blank=True)
    # The search vector field is used to create a full-text index on the fields primary_object_name and
    # secondary_object_name. The 0004 migration creates a trigger to update the search_vector field when the
    # primary_object_name or secondary_object_name fields are created or updated (updated should really never happen).
    # This field should be used when filtering on the primary_object_name or secondary_object_name fields. So instead of
    # ACSJournal.objects.filter(primary_object_name__<search/contains/startswith> ='CHC') you would use
    # ACSJournal.objects.filter(search_vector='CHC') This will use the full-text index and be much faster.
    # This is a Postgres specific feature.
    # When the project is updated to Django 5, we can use the new GeneratedField to remove the trigger.
    # https://docs.djangoproject.com/en/5.0/releases/5.0/#database-generated-model-field
    search_vector = SearchVectorField(null=True, blank=True)

    class Meta:
        verbose_name = "ACSJournal"
        verbose_name_plural = "ACSJournal"
        indexes = [
            models.Index(
                fields=[
                    "campus_id",
                    "state_code",
                    "message_type",
                    "primary_object_uuid",
                    "secondary_object_uuid",
                ]
            ),
            BrinIndex(fields=["message_utc", "server_utc"], autosummarize=True),
            GinIndex(fields=["search_vector"]),
        ]

    def __str__(self):
        return f"{self.message_utc.astimezone(SATTimeZones.EPT.value)}: {self.primary_object_name} ({self.state_code})"

    def primary_type(self):
        if self.primary_object_type:
            return END_OF_DOTTED_PATH.search(self.primary_object_type).group(1)
        return ""

    def secondary_type(self):
        if self.secondary_object_type:
            return END_OF_DOTTED_PATH.search(self.secondary_object_type).group(1)
        return ""
