import os
from datetime import datetime, timedelta

import pytz
from celery import shared_task
from sat.db import ConnectionType, get_db_connection
from sat.logs import SATLogger

from sat_automations.acs_journal.models import ACSJournal
from sat_automations.manage_automations.models import SlackChannel
from sat_automations.utils.authenticate import GoogleAuthenticate
from sat_automations.utils.common.connection import CCURE_CONNECTION
from sat_automations.utils.common.queries import get_journal_range
from sat_automations.utils.common.time import SATTimeZones, date_as_datetime, yield_times
from sat_automations.utils.managers import BulkCreateManager

logger = SATLogger(__name__)


class AccessControlJournal:
    def __init__(self) -> None:
        if os.getenv("ENVIRONMENT") != "dev":
            self.authenticator = GoogleAuthenticate("access_control_journal")
            self.authenticator.authenticate()
            if self.authenticator.response_json:
                self.access_token = self.authenticator.response_json.get("token")
            else:
                logger.error(
                    f"Unable to authenticate with auth service. Error: {self.authenticator.error_message}"
                )
                raise ConnectionError("Unable to authenticate with auth service.")
        self.ccure_connection = get_db_connection(ConnectionType.SQL, conn_string=CCURE_CONNECTION)
        self.bulk_create_manager = BulkCreateManager(chunk_size=10000)
        self.slack = SlackChannel.objects.filter(channel_name="access_control_journal").first()

    def query_journal(self, start_date: str, end_date: str):
        # The microseconds in these dates are expressed as a 6 digit integer.
        # MSSQL uses a 3 digit integer, thus we need to truncate the microseconds to 3 digits.
        query = get_journal_range(start_date[:-3], end_date[:-3])
        cursor = self.ccure_connection.cursor()
        results = cursor.execute(query)
        return results.fetchall()

    def _process_journal(self, start_date: datetime, end_date: datetime):
        """
        Process the journal for a range of datetimes.
        :param start_date: datetime in UTC
        :param end_date: datetime in UTC
        :return:
        """
        for begin, end in yield_times(start_date, end_date):
            results = self.query_journal(begin, end)
            logger.info(f"Processing range for {begin} to {end} with {len(results)} records.")
            for row in results:
                message_utc = row[3].replace(tzinfo=pytz.utc)
                server_utc = row[4].replace(tzinfo=pytz.utc)
                acs_journal = ACSJournal(
                    campus_id=row[0],
                    card_num=row[1],
                    message_type=row[2],
                    message_utc=message_utc,
                    server_utc=server_utc,
                    state_code=row[5],
                    primary_object_name=row[6],
                    primary_object_type=row[7],
                    primary_object_uuid=row[8],
                    secondary_object_name=row[9],
                    secondary_object_type=row[10],
                    secondary_object_uuid=row[11],
                )
                self.bulk_create_manager.add(acs_journal)
        self.bulk_create_manager.done()
        logger.info("Processing complete.")

    def process_hourly_journal(self, days=1):
        logger.info("Finding last journal entry.")
        end = datetime.now(tz=SATTimeZones.UTC.value)
        # Find the last journal entry, by looping through the days until we find a record.
        while not ACSJournal.objects.filter(server_utc__gt=(end - timedelta(days=days))).exists():
            days += 1
        logger.info(f"Found last journal entry on date: {end - timedelta(days=days)}.")
        start = (
            ACSJournal.objects.filter(server_utc__gt=(end - timedelta(days=days)))
            .order_by("-server_utc")
            .first()
        )
        logger.info(f"Processing journal from {start.server_utc} to {end}.")
        self._process_journal(start.server_utc, end)

    def process_from_start_of_journal(self, start="2022-11-13"):
        """
        Process the rows from the start of the journal. As of the time of this writing, the first record in the
        CCURE journal is 2022-11-13.
        :return:
        """
        start = date_as_datetime(start)
        end = datetime.now(tz=SATTimeZones.UTC.value)
        self._process_journal(start_date=start, end_date=end)


@shared_task
def process_access_control_journal():
    acj = AccessControlJournal()
    try:
        acj.process_hourly_journal()
        logger.info("Access Control Journal processing complete.")
    except ValueError as e:
        logger.error(f"Error processing access control journal: {e}")
        acj.slack.send_message(f"Error processing access control journal: {e}")
