import pytest
from django.urls import reverse
from time_machine import travel

from sat_automations.acs_journal.models import ACSJournal
from sat_automations.acs_journal.views import DEFAULT_QUERY_DAYS
from sat_automations.utils.common.time import datetimestr_as_datetime

pytestmark = pytest.mark.django_db


@pytest.fixture
def patch_liaison_clearances(mocker):
    mocker.patch(
        "sat_automations.acs_journal.views.JournalViewSet.get_liaison_clearances",
    )


@pytest.fixture
def patch_clearance_items_allowed(mocker):
    def _patch_clearance_items_allowed(items):
        guid_list = []
        for i, item in enumerate(items):
            guids = {}
            if i % 2 == 0:
                guids["door_guid"] = item
            else:
                guids["elevator_guid"] = item
            guid_list.append(guids)
        mocker.patch(
            "sat_automations.acs_journal.views.JournalViewSet.get_clearance_items",
            return_value=guid_list,
        )

    return _patch_clearance_items_allowed


def test_journal_list_view_default(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    journal, primary_uuids, secondary_uuids = random_acs_journal(days=DEFAULT_QUERY_DAYS)
    patch_clearance_items_allowed([primary_uuids] + [secondary_uuids])
    journal.save()
    url = reverse("acs-journal-list")
    response = client.get(url, format="json")
    assert response.status_code == 200
    assert response.json()["count"] == 1
    obj_fields = [x.name for x in journal._meta.fields]
    for key in response.json()["results"][0].keys():
        assert key in obj_fields


def test_journal_list_view_outside_default_days(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    journal, primary_uuids, secondary_uuids = random_acs_journal(days=20)
    patch_clearance_items_allowed([primary_uuids] + [secondary_uuids])
    journal.save()
    url = reverse("acs-journal-list")
    response = client.get(url, format="json")
    assert response.status_code == 200
    assert response.json()["count"] == 0


def test_journal_list_view_limit_by_uuid(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    journals, primary_uuids, secondary_uuids = random_acs_journal(
        count=4, days=DEFAULT_QUERY_DAYS, increment_by="minutes"
    )
    # Remove a uuid "assigned" to a liaison
    [x.save() for x in journals]
    patch_clearance_items_allowed(primary_uuids[-1:] + secondary_uuids[-1:])
    url = reverse("acs-journal-list")
    response = client.get(url, format="json")
    assert response.status_code == 200
    assert ACSJournal.objects.all().count() == 4
    assert response.json()["count"] == 1


def test_no_clearances_assigned_get_all(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    journals, _, _ = random_acs_journal(count=4, days=DEFAULT_QUERY_DAYS, increment_by="minutes")
    patch_clearance_items_allowed([])
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json")
    assert response.status_code == 200
    assert response.json()["count"] == 4


def test_journal_search_or(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    # There should be 4 results for this search
    journals1, primary_uuids, secondary_uuids = random_acs_journal(count=2)
    journals2, primary_uuids2, secondary_uuids2 = random_acs_journal(
        count=2, primary_name_prefix="test"
    )
    journals = journals1 + journals2
    patch_clearance_items_allowed(
        primary_uuids + secondary_uuids + primary_uuids2 + secondary_uuids2
    )
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"search_or": "test CHC"})
    assert response.status_code == 200
    assert response.json()["count"] == 4


def test_journal_search_and(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    # There should be 0 results for this search
    journals1, primary_uuids, secondary_uuids = random_acs_journal(count=2)
    journals2, primary_uuids2, secondary_uuids2 = random_acs_journal(
        count=2, primary_name_prefix="test"
    )
    journals = journals1 + journals2
    patch_clearance_items_allowed(
        primary_uuids + secondary_uuids + primary_uuids2 + secondary_uuids2
    )
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"search_and": "test CHC"})
    assert response.status_code == 200
    assert response.json()["count"] == 0


def test_journal_search_and_w_results(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    # There should be 4 results for this search
    journals1, primary_uuids, secondary_uuids = random_acs_journal(count=2)
    journals2, primary_uuids2, secondary_uuids2 = random_acs_journal(
        count=2, primary_name_prefix="test"
    )
    journals = journals1 + journals2
    patch_clearance_items_allowed(
        primary_uuids + secondary_uuids + primary_uuids2 + secondary_uuids2
    )
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"search_term": "CHC CHS"})
    assert response.status_code == 200
    assert response.json()["count"] == 4


def test_journal_search_term(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    # There should be 2 results for this search
    journals1, primary_uuids, secondary_uuids = random_acs_journal(count=2)
    journals2, primary_uuids2, secondary_uuids2 = random_acs_journal(
        count=2, primary_name_prefix="test"
    )
    journals = journals1 + journals2
    patch_clearance_items_allowed(
        primary_uuids + secondary_uuids + primary_uuids2 + secondary_uuids2
    )
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"search": "CHC"})
    assert response.status_code == 200
    assert response.json()["count"] == 2


def test_journal_date_range(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    # This should return 2 results.
    # This tries to replicate the server state which will be a UTC datetime, being searched with an EPT datetime.
    # Users of the system will be thinking in EPT when they search.
    server_utc = "2022-01-03 05:00:00"
    search_ept = "2022-01-03 00:00:00"
    end_ept = "2022-01-04 00:00:00"
    start_dt = datetimestr_as_datetime(server_utc)
    traveller = travel(start_dt)
    traveller.start()
    journals, primary_uuids, secondary_uuids = random_acs_journal(
        count=2, start_dt=start_dt, days=0
    )
    patch_clearance_items_allowed(primary_uuids + secondary_uuids)
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"start": search_ept, "end": end_ept})
    assert response.status_code == 200
    assert response.json()["count"] == 2
    traveller.stop()


def test_journal_date_range_too_long(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    """
    It should return one result. Two items exist between the given search_ept and end_ept dates,
    but the search_ept date will be modified to be within 72 hours of the end_ept.
    """
    server_utc = "2022-01-03 05:00:00"
    search_ept = "2021-12-25 00:00:00"
    end_ept = "2022-01-06 17:00:00"
    start_dt = datetimestr_as_datetime(server_utc)
    traveller = travel(start_dt)
    traveller.start()
    journals, primary_uuids, secondary_uuids = random_acs_journal(
        count=2, start_dt=start_dt, days=0
    )
    patch_clearance_items_allowed(primary_uuids + secondary_uuids)
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"start": search_ept, "end": end_ept})
    assert response.status_code == 200
    assert response.json()["count"] == 1
    traveller.stop()


def test_journal_date_range_plus_term(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    server_utc = "2022-01-03 05:00:00"
    search_ept = "2022-01-03 00:00:00"
    end_ept = "2022-01-04 00:00:00"
    start_dt = datetimestr_as_datetime(server_utc)
    traveller = travel(start_dt)
    traveller.start()
    # Generate the first two in the range
    journals1, primary_uuids, secondary_uuids = random_acs_journal(
        count=2, start_dt=start_dt, days=0
    )
    patch_clearance_items_allowed(primary_uuids + secondary_uuids)
    # Generate the next two 2 days prior to the range
    journals2, primary_uuids2, secondary_uuids2 = random_acs_journal(
        count=2, start_dt=start_dt, days=2
    )
    journals = journals1 + journals2
    patch_clearance_items_allowed(
        primary_uuids + secondary_uuids + primary_uuids2 + secondary_uuids2
    )
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(
        url, format="json", data={"start": search_ept, "end": end_ept, "search": "CHC"}
    )
    assert response.status_code == 200
    assert response.json()["count"] == 2
    traveller.stop()


def test_journal_state_code(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    # This should return 2 results.
    journals, primary_uuids, secondary_uuids = random_acs_journal(count=3, increment_by="minutes")
    patch_clearance_items_allowed(primary_uuids + secondary_uuids)
    for i, journal in enumerate(journals):
        journal.state_code = "Inactive"
        if i == 0:
            journal.state_code = "Admit"
        journal.save()
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"state_code": "Inactive"})
    assert response.status_code == 200
    assert response.json()["count"] == 2


def test_custom_page_size(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    journals, primary_uuids, secondary_uuids = random_acs_journal(count=10, increment_by="minutes")
    patch_clearance_items_allowed(primary_uuids + secondary_uuids)
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json", data={"page_size": 5})
    assert response.status_code == 200
    assert response.json()["count"] == 10
    assert len(response.json()["results"]) == 5


def test_default_page_size(
    client, random_acs_journal, patch_liaison_clearances, patch_clearance_items_allowed
):
    journals, primary_uuids, secondary_uuids = random_acs_journal(count=150, increment_by="minutes")
    patch_clearance_items_allowed(primary_uuids + secondary_uuids)
    [x.save() for x in journals]
    url = reverse("acs-journal-list")
    response = client.get(url, format="json")
    assert response.status_code == 200
    assert response.json()["count"] == 150
    assert len(response.json()["results"]) == 100
