import datetime
import operator
import os
from datetime import timedelta
from functools import reduce

import requests
from django.conf import settings
from django.contrib.postgres.search import SearchQuery
from django.db.models import Q, QuerySet
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from ncsu.authorization.authorization_roles import READ_ROLES
from rest_framework import serializers, viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import APIException, AuthenticationFailed
from rest_framework.response import Response
from sat.db import ConnectionType, get_db_connection
from sat.logs import SATLogger

from sat_automations.acs_journal.models import ACSJournal
from sat_automations.utils.common.connection import CCURE_CONNECTION
from sat_automations.utils.common.paginators import ConfigurablePageNumberPagination
from sat_automations.utils.common.queries import return_ccure_clearance_items_query
from sat_automations.utils.common.time import SATTimeZones, bound_time, datetimestr_as_datetime

# Default query time
DEFAULT_QUERY_DAYS = 3
# CACHE Timeout in seconds: 1 hour
JOURNAL_CACHE_TTL = 60 * 60
if os.getenv("DUMMY_CACHE"):
    JOURNAL_CACHE_TTL = 0

logger = SATLogger(__name__)


class JournalSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = ACSJournal
        fields = [
            "id",
            "campus_id",
            "card_num",
            "message_type",
            "message_utc",
            "server_utc",
            "state_code",
            "primary_object_type",
            "primary_object_name",
            "primary_object_uuid",
            "secondary_object_type",
            "secondary_object_name",
            "secondary_object_uuid",
            "created",
        ]


class JournalViewSet(viewsets.ReadOnlyModelViewSet):
    """
    This is a specialized viewset. The current size of the data is around 32 million rows. Meaning that
    we need to be very careful about how we filter and paginate the data. The default queryset is the last
    3 months of data.

    The idea is to use filtering to limit the data to a reasonable size using only the indexed fields and
    the vector search field. The vector search field is a full-text index on the primary_object_name and the
    secondary_object_name

    Supported Request Parameters:
    -----------------------------
    - `liaison_campus_id`: The campus_id of the liaison. This will be used to limit the returned data to the clearance
    items a liaison manages. This is a required field.
    - `search_or`: A space separated list of terms to search for. This will be an OR search.
    - `search_and`: A space separated list of terms to search for. This will be an AND search.
    - `search`: A single term to search for. This will be a simple search.
    - `start`: A datetime string to start the range. This will be converted to UTC. (YYYY-MM-DD HH:MM:SS)
    - `end`: A datetime string to end the range. This will be converted to UTC. (YYYY-MM-DD HH:MM:SS)
    - `state_code`: A state code to filter on.
       - Active
       - Admit
       - Canceled
       - DoorClosed
       - DoorForced
       - DoorHeld
       - Inactive
       - Locked
       - Reject
       - Scheduled
       - Unknown
       - Unlocked


    Usage:
    ------
    - `/api/v1/acs-journal/?liaison_campus_id=<campus_id>` # Will return the last 7 days of data. Limited by the clearances of the liaison.
    - `/api/v1/acs-journal/?liaison_campus_id=<campus_id>&page=20` # Will return the 20th page of data.
    - `/api/v1/acs-journal/?liaison_campus_id=<campus_id>&search_or=CHC+CHS` # This will return all records that contain CHC or CHS
    - `/api/v1/acs-journal/?liaison_campus_id=<campus_id>&search_and=CHC+CHS` # This will return all records that contain CHC and CHS
    - `/api/v1/acs-journal/?liaison_campus_id=<campus_id>&search=CHC` # This will return all records that contain CHC
    - `/api/v1/acs-journal/?liaison_campus_id=<campus_id>&start=2022-01-01+00:00:00&end=2022-01-31+23:59:59` # This will return all records between the start and end dates.
    - `/api/v1/acs-journal/?liaison_campus_id=<campus_id>&state_code=Active` # This will return all records that have a state code of Active
    """

    queryset = ACSJournal.objects.all()
    serializer_class = JournalSerializer
    pagination_class = ConfigurablePageNumberPagination

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clearance_items_allowed = None

    def _filters(self, request) -> QuerySet:
        search_or = request.query_params.get("search_or")
        search_and = request.query_params.get("search_and")
        search_term = request.query_params.get("search")
        start = request.query_params.get("start")
        end = request.query_params.get("end")
        state_code = request.query_params.get("state_code")
        campus_id = request.query_params.get("campus_id")

        # Convert to utc datetime
        # This little bit of foolery is because the dates from the journal are stored in UTC, but
        # our users will be thinking in EPT. So we need to: first, tell the system that the date is
        # an EPT datetime, then second, convert the dates to UTC before we filter.
        if start:
            start = datetimestr_as_datetime(start, SATTimeZones.EPT).astimezone(
                SATTimeZones.UTC.value
            )
        if end:
            end = datetimestr_as_datetime(end, SATTimeZones.EPT).astimezone(SATTimeZones.UTC.value)
        if start and end:
            if end - start > timedelta(days=DEFAULT_QUERY_DAYS):
                start = end - timedelta(days=DEFAULT_QUERY_DAYS)
        elif start:
            end = start + timedelta(days=DEFAULT_QUERY_DAYS)
        elif end:
            start = end - timedelta(days=DEFAULT_QUERY_DAYS)

        # use the correct queryset
        if start or end:
            qs = self._time_range_queryset(start, end)
        else:
            qs = self._default_queryset()

        if campus_id:
            qs = qs.filter(campus_id=campus_id)
        if any([search_term, search_or, search_and]):
            if search_term:
                qs = qs.filter(search_vector=search_term)
            if search_or:
                qs = self._search_vector_queryset(search_or, qs, boolean_type="|")
            if search_and:
                qs = self._search_vector_queryset(search_and, qs)

        qs = self._state_code_queryset(state_code, qs)

        if self.clearance_items_allowed:
            # Apply the ClearanceItems filter
            qs = qs.filter(self.clearance_items_allowed[0] | self.clearance_items_allowed[1])
        return qs

    @staticmethod
    def _state_code_queryset(state_code: str, qs: QuerySet) -> QuerySet:
        if state_code:
            return qs.filter(state_code=state_code)
        else:
            return qs.filter(state_code__in=["Admit", "Reject"])

    @staticmethod
    def _search_vector_queryset(search_term: str, qs: QuerySet, boolean_type="&") -> QuerySet:
        terms = search_term.split(" ")
        query_str = ""
        if len(terms) > 1:
            query_str += ""
            for term in terms:
                query_str += f"'{term}' {boolean_type} "
            query_str = query_str[:-3]
        else:
            query_str = terms[0]
        search_term = SearchQuery(query_str, search_type="raw")
        return qs.filter(search_vector=search_term)

    def _time_range_queryset(self, start: datetime.datetime, end: datetime.datetime) -> QuerySet:
        return self.get_queryset().filter(message_utc__lte=end, message_utc__gte=start)

    @staticmethod
    def query_url_get(request, url: str) -> requests.Response:
        response = requests.get(
            url, headers={"Authorization": f"{request.headers['Authorization']}"}
        )
        if response.status_code == 401:
            raise AuthenticationFailed(
                "Unauthorized: Most likely a token timeout. Please refresh your browser."
            )
        return response

    def get_liaison_clearances(self, request):
        """
        Get the liaison clearances for the current user.
        :param request: Request
        :return: Response
        """
        url = (
            f"{settings.CLEARANCE_SERVICE_URL}/liaison?campus_id={request.GET['liaison_campus_id']}"
        )
        try:
            response = self.query_url_get(request, url)
            if response.status_code != 200:
                return []
            return [x["id"] for x in response.json()["clearances"]]
        except AuthenticationFailed as e:
            raise e

    @staticmethod
    def get_clearance_items(items: list[int]):
        c9k_db_conn = get_db_connection(ConnectionType.SQL, conn_string=CCURE_CONNECTION)
        cursor = c9k_db_conn.cursor()
        query = return_ccure_clearance_items_query(items)
        logger.debug(f"Query: {query}")
        cursor.execute(query)
        column_names = [column[0] for column in cursor.description]
        result_rows = [dict(zip(column_names, row)) for row in cursor.fetchall()]
        cursor.close()
        return result_rows

    @staticmethod
    def _build_q_query(items: list[str]) -> tuple[Q, Q]:
        primary = reduce(operator.or_, (Q(primary_object_uuid=x) for x in items))
        secondary = reduce(operator.or_, (Q(secondary_object_uuid=x) for x in items))
        return primary, secondary

    def _default_queryset(self, days: int = DEFAULT_QUERY_DAYS) -> QuerySet:
        from_dt = bound_time(timezone.now() - timedelta(days=days), nearest_day=True).replace(
            tzinfo=SATTimeZones.EPT.value
        )
        to_dt = bound_time(timezone.now()).replace(tzinfo=SATTimeZones.EPT.value)
        logger.debug(f"Default days filtered from ({days}) days")
        return self.get_queryset().filter(message_utc__lte=to_dt, message_utc__gte=from_dt)

    @staticmethod
    def door_elevator_guids(items):
        itms = []
        for i in items:
            if item := i.get("door_guid"):
                itms.append(item)
            if item := i.get("elevator_guid"):
                itms.append(item)
        return itms

    @method_decorator(cache_page(JOURNAL_CACHE_TTL))
    def list(self, request, *args, **kwargs):
        # Since the data is so enormous, we need to limit a bare list request to a
        # reasonable number of records.
        # Default will be the last 3 days of data.
        # Page urls are cached for 10 minutes.
        logger.info(f"Request: {request.META.get('QUERY_STRING')}")
        clr_items = []
        try:
            if clearance_ids := self.get_liaison_clearances(request):
                clearance_items = self.get_clearance_items(clearance_ids)
                clr_items = self.door_elevator_guids(clearance_items)
        except AuthenticationFailed as e:
            logger.error(f"AuthenticationFailed: {e}")
            raise e
        except Exception as e:
            logger.error(f"Error getting clearance items: {e}")
            raise APIException(f"Error getting clearance items. {e}")
        logger.info(f"Clearance Items: {len(clr_items)}")
        if clr_items:
            # Until our Auth system has a good RBAC setup, we will assume that a person using
            # the tool is not an admin when clr_items are populated.
            # If `clr_items` is empty, then the person is an admin and can
            # see all records.
            self.clearance_items_allowed = self._build_q_query(clr_items)
        filtered = self._filters(request)
        queryset = self.filter_queryset(filtered)
        logger.debug(f"Queryset explain: {queryset.explain()}")
        ordered_qs = queryset.order_by("-message_utc")
        page = self.paginate_queryset(ordered_qs)
        try:
            if page is not None:
                serializer = JournalSerializer(page, many=True)
                return self.get_paginated_response(serializer.data)
            serializer = self.get_serializer(ordered_qs, many=True)
            return Response(serializer.data)
        except AttributeError as e:
            logger.error(f"Serializer error: {e}")
            raise APIException(f"Serializer error: {e}")
        except Exception as e:
            logger.error(f"Serializer error: {e}")
            raise APIException(f"Serializer error: {e}")

    @action(detail=False)
    def get_entries_by_id(self, request):
        url = f"{settings.AUTH_SERVICE_URL}/authorize/roles"
        response = self.query_url_get(request, url)
        if roles := response.json().get("roles"):
            if not any(role in READ_ROLES for role in roles):
                raise APIException(code=401, detail="Not authorized to read this data.")
        if campus_id := request.query_params.get("campus_id"):
            filtered = self.get_queryset().filter(campus_id=campus_id).order_by("-message_utc")[:10]
            queryset = self.filter_queryset(filtered)
            serializer = self.get_serializer(queryset, many=True)
            return Response(serializer.data)
        raise APIException(code=400, detail="No campus_id supplied.")
