from datetime import datetime
from typing import Optional

from celery import shared_task
from sat.gravity_forms import Entry, GravityForms
from sat.logs import SATLogger

from sat_automations.affiliate_automation.models import Cardholder, CardRequest, Sponsor
from sat_automations.utils.common.time import datetimestr_as_datetime

logger = SATLogger(__name__)
CARD_REQUEST_FORM_ID = 2
CARDHOLDER_INFO_FORM_ID = 3

ITEMS_PER_PAGE = 50

card_types = {
    "Visitor": "VI",
    "Affiliate": "AF",
    "Corporate Partner": "CP",
    "Contractor": "CO",
    "Visitor (Self Pay)": "VS",
}


class CardRequestScrape:
    def __init__(self):
        self.gravity_forms = GravityForms()

    def get_form_entries(
        self,
        form_id: int,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        page: Optional[int] = 1,
    ) -> list[Entry]:
        """get card requests for the given timeframe"""
        filters = []
        if start_time:
            filters.append(("date_created", str(start_time), ">"))
        if end_time:
            filters.append(("date_created", str(end_time), "<"))
        date_filter = self.gravity_forms.field_filters(keys=filters)
        card_requests = self.gravity_forms.search_entry(
            field_filters=date_filter, form_id=form_id, page=page, page_size=ITEMS_PER_PAGE
        )
        return card_requests

    def save_card_requests(
        self, start_time: Optional[datetime] = None, end_time: Optional[datetime] = None
    ):
        page = 1
        while True:
            card_requests = self.get_form_entries(CARD_REQUEST_FORM_ID, start_time, end_time, page)
            logger.info(f"Processing page {page} of card requests")
            for card_dict in card_requests:
                sponsor, created = Sponsor.objects.get_or_create(email=card_dict.details["21"])
                if created:
                    sponsor.name = card_dict.details["22"]
                    sponsor.unity_id = card_dict.details["23"]
                    sponsor.save()
                card_request, _ = CardRequest.objects.get_or_create(
                    sponsor=sponsor,
                    card_type=card_types[card_dict.details["7"]],
                    date_created=datetimestr_as_datetime(card_dict.date_created),
                    details=card_dict.details,
                )
                affiliate_ids = card_dict.details["32"].split(",")
                logger.debug(f"Processing {len(affiliate_ids)} for {sponsor.name}")
                cardholders = []
                for aff_id in affiliate_ids:
                    if not aff_id:
                        continue
                    cardholder_data = self.gravity_forms.get_entry(aff_id)
                    campus_id = cardholder_data.details.get("4") or cardholder_data.details.get("5")
                    if not campus_id:
                        logger.error(
                            f"Affiliate record doesn't have a campus_id: affiliate_id {aff_id}"
                        )
                        continue
                    try:
                        cardholder, _ = Cardholder.objects.get_or_create(
                            email=cardholder_data.details["3"],
                            first_name=cardholder_data.details["2.3"],
                            last_name=cardholder_data.details["2.6"],
                            campus_id=campus_id,
                        )
                        cardholders.append(cardholder)
                    except RuntimeError as e:
                        logger.error(
                            f"Unable to add cardholder: {cardholder_data.details}, due to {e}"
                        )
                        raise e
                    except KeyError as ke:
                        logger.error(f"Missing a key in the cardholder data: {ke}")
                        raise ke
                logger.info(f"Saving card request for {sponsor.name}")
                card_request.cardholders.set(cardholders)
                card_request.save()
            if len(card_requests) < ITEMS_PER_PAGE:
                break
            page += 1


@shared_task(name="initial_scrape")
def initial_scrape():
    """
    This is the initial scrape that will be run to get all the card requests. This
    should only need to be run once, or if there is a need to re-scrape all the card
    data. For example in the case of a catastrophic failure database failure.
    :return:
    """
    scrape = CardRequestScrape()
    scrape.save_card_requests()


@shared_task(name="daily_scrape")
def daily_scrape():
    """
    This is the daily scrape that will be run to get all the card requests from the previous day.
    :return:
    """
    # Get the last Card Request
    scrape = CardRequestScrape()
    start_time = CardRequest.objects.last().date_created
    logger.info(f"Starting daily scrape from {start_time}")
    scrape.save_card_requests(start_time=start_time)
