from datetime import datetime

import pytest
from sat.gravity_forms import Entry

from sat_automations.affiliate_automation.models import Cardholder, CardRequest, Sponsor
from sat_automations.affiliate_automation.tasks import CardRequestScrape, daily_scrape

pytestmark = pytest.mark.django_db
now = datetime.now()

CARD_REQUEST_RESULTS = [
    {
        "id": "19679",
        "form_id": "2",
        "post_id": None,
        "date_created": "2024-04-16 12:45:38",
        "date_updated": "2024-04-16 12:45:38",
        "is_starred": "0",
        "is_read": "0",
        "ip": "152.1.238.148",
        "source_url": "https://onecard.ncsu.edu/get-card/card-request-form/",
        "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "currency": "USD",
        "payment_status": None,
        "payment_date": None,
        "payment_amount": None,
        "payment_method": None,
        "transaction_id": None,
        "is_fulfilled": None,
        "created_by": "7670",
        "transaction_type": None,
        "status": "active",
        "22": "Melissa Taylor",
        "23": "mawahl",
        "21": "mawahl@ncsu.edu",
        "32": "19500",
        "7": "Contractor",
        "26": "Quality Builders, Inc. of Raleigh",
        "33": "(919) 821-9001",
    },
    {
        "id": "19677",
        "form_id": "2",
        "post_id": None,
        "date_created": "2024-04-15 22:44:06",
        "date_updated": "2024-04-15 22:44:06",
        "is_starred": "0",
        "is_read": "0",
        "ip": "136.56.207.93",
        "source_url": "https://onecard.ncsu.edu/get-card/card-request-form/",
        "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "currency": "USD",
        "payment_status": None,
        "payment_date": None,
        "payment_amount": None,
        "payment_method": None,
        "transaction_id": None,
        "is_fulfilled": None,
        "created_by": "7769",
        "transaction_type": None,
        "status": "active",
        "22": "Ashok Gopalarathnam",
        "23": "agopalar",
        "21": "agopalar@ncsu.edu",
        "32": "19500",
        "7": "Visitor",
        "13": "Engineering",
        "27": "140940",
        "25": "687107",
        "12": "Elizabeth Walker",
        "14": "7910",
    },
    {
        "id": "19679",
        "form_id": "2",
        "post_id": None,
        "date_created": "2024-04-16 12:45:38",
        "date_updated": "2024-04-16 12:45:38",
        "is_starred": "0",
        "is_read": "0",
        "ip": "152.1.238.148",
        "source_url": "https://onecard.ncsu.edu/get-card/card-request-form/",
        "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "currency": "USD",
        "payment_status": None,
        "payment_date": None,
        "payment_amount": None,
        "payment_method": None,
        "transaction_id": None,
        "is_fulfilled": None,
        "created_by": "7670",
        "transaction_type": None,
        "status": "active",
        "22": "Melissa Taylor",
        "23": "mawahl",
        "21": "mawahl@ncsu.edu",
        "32": "19500",
        "7": "Contractor",
        "26": "Quality Builders, Inc. of Raleigh",
        "33": "(919) 821-9001",
    },
]

CARDHOLDER_RESULTS = {
    "id": "19500",
    "form_id": "3",
    "post_id": None,
    "date_created": "2023-12-11 18:12:54",
    "date_updated": "2023-12-11 18:12:54",
    "is_starred": "0",
    "is_read": "0",
    "ip": "10.152.4.73",
    "source_url": "https://onecard.ncsu.edu/get-card/card-request-form/",
    "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "currency": "USD",
    "payment_status": None,
    "payment_date": None,
    "payment_amount": None,
    "payment_method": None,
    "transaction_id": None,
    "is_fulfilled": None,
    "created_by": "7749",
    "transaction_type": None,
    "status": "active",
    "2.3": "MICHAEL",
    "2.6": "STEWART",
    "3": "tomc@bfpe.com",
    "4": "100037981",
    "gpnf_entry_nested_form_field": "32",
    "gpnf_entry_parent": "19501",
    "gpnf_entry_parent_form": "2",
    "2.2": "",
    "2.4": "",
    "2.8": "",
    "5": "",
    "gravityformsrecaptcha_score": False,
}


@pytest.fixture
def patch_card_scrape(mocker):
    return mocker.patch(
        "sat_automations.affiliate_automation.tasks.CardRequestScrape.get_form_entries"
    )


@pytest.fixture
def patch_gravity_forms(mocker):
    return mocker.patch("sat_automations.affiliate_automation.tasks.GravityForms.__init__")


@pytest.fixture
def patch_entry(mocker):
    return mocker.patch("sat_automations.affiliate_automation.tasks.GravityForms.get_entry")


def test_save_card_requests(patch_card_scrape, patch_gravity_forms, patch_entry):
    patch_gravity_forms.return_value = None
    patch_card_scrape.return_value = [Entry(**x) for x in CARD_REQUEST_RESULTS]
    patch_entry.return_value = Entry(**CARDHOLDER_RESULTS)

    scrape = CardRequestScrape()
    scrape.save_card_requests()

    sponsor_count = Sponsor.objects.count()
    cardholder_count = Cardholder.objects.count()
    card_request_count = CardRequest.objects.count()

    assert sponsor_count == 2
    assert cardholder_count == 1
    assert card_request_count == 2


def test_save_card_requests_saves_campus_id(patch_card_scrape, patch_entry, patch_gravity_forms):
    patch_gravity_forms.return_value = None
    patch_card_scrape.return_value = [Entry(**x) for x in CARD_REQUEST_RESULTS]
    patch_entry.return_value = Entry(**CARDHOLDER_RESULTS)

    scrape = CardRequestScrape()
    scrape.save_card_requests()
    assert Cardholder.objects.first().campus_id == "100037981"


def test_daily_scrape(patch_card_scrape, patch_entry, patch_gravity_forms):
    patch_gravity_forms.return_value = None
    patch_card_scrape.return_value = [Entry(**CARD_REQUEST_RESULTS[1])]
    patch_entry.return_value = Entry(**CARDHOLDER_RESULTS)

    scrape = CardRequestScrape()
    scrape.save_card_requests()

    patch_card_scrape.return_value = [Entry(**CARD_REQUEST_RESULTS[2])]

    daily_scrape()

    sponsor_count = Sponsor.objects.count()
    cardholder_count = Cardholder.objects.count()
    card_request_count = CardRequest.objects.count()

    assert sponsor_count == 2
    assert cardholder_count == 1
    assert card_request_count == 2


# We need to add more tests here.
# What about when the email for the holder is used multiple times
# These tests also don't cover multiple holders in field 32
# Should be done as a group exercise.
