# The views are primarily DRF viewsets, which are a combination of views and serializers.
# Path: sat_automations/api/urls.py
from rest_framework import serializers, viewsets

from sat_automations.affiliate_automation.models import Cardholder, CardRequest, Sponsor


class SponsorSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Sponsor
        fields = [
            "id",
            "created",
            "modified",
            "name",
            "email",
            "unity_id",
        ]


class SponsorViewSet(viewsets.ReadOnlyModelViewSet):
    """
    This viewset automatically provides `list` and `detail` actions.

    Usage:
    ------
    """

    queryset = Sponsor.objects.all()
    serializer_class = SponsorSerializer


class NestedCardRequestSerializer(serializers.HyperlinkedModelSerializer):
    """Used in /card-request/cardholder results.card_requests"""

    sponsor = SponsorSerializer(many=False)

    class Meta:
        model = CardRequest
        fields = [
            "id",
            "date_created",
            "card_type",
            "sponsor",
        ]


class CardholderSerializer(serializers.HyperlinkedModelSerializer):
    card_requests = serializers.SerializerMethodField()

    def get_card_requests(self, cardholder):
        card_requests = CardRequest.objects.filter(cardholders=cardholder)
        return NestedCardRequestSerializer(card_requests, many=True).data

    class Meta:
        model = Cardholder
        fields = [
            "id",
            "first_name",
            "last_name",
            "email",
            "campus_id",
            "ccure_id",
            "prox_id",
            "card_requests",
        ]


class CardholderViewSet(viewsets.ReadOnlyModelViewSet):
    """
    This viewset displays the cardholders from card requests.
    The list can be filtered with query params. Params are case insensitive, but need an exact match

    Supported Request Parameters:
    -----------------------------
    - `first_name`
    - `last_name`
    - `campus_id`
    - `prox_id`

    Usage:
    ------
    - `/card-request/cardholder` # Will return all cardholders in the database
    - `/card-request/cardholder/?first_name=larry`  # returns all cardholders with the first name 'Larry'
    - `/card-request/cardholder/?last_name=Bird`  # returns all cardholders with last name 'Bird'
    - `/card-request/cardholder/?campus_id=001190210` # returns the cardholder with this campus_id
    - `/card-request/cardholder/?prox_id=99999` # returns the cardholder with this prox_id
    - `/card-request/cardholder/?first_name=tim&last_name=duncan` # return cardholders that match all the given params
    """

    queryset = Cardholder.objects.all()
    serializer_class = CardholderSerializer

    def get_queryset(self):
        queryset = super().get_queryset()

        fields = {
            "first_name": "",
            "last_name": "",
            "campus_id": "",
            "prox_id": "",
        }
        search = {}
        for key in fields:
            if param_val := self.request.query_params.get(key):
                search[key] = param_val
        queryset = queryset.filter(**search)
        return queryset


class CardRequestSerializer(serializers.HyperlinkedModelSerializer):
    cardholders = CardholderSerializer(many=True)
    sponsor = SponsorSerializer(many=False)

    class Meta:
        model = CardRequest
        fields = [
            "id",
            "date_created",
            "card_type",
            "details",
            "sponsor",
            "cardholders",
        ]


class CardRequestViewSet(viewsets.ReadOnlyModelViewSet):
    """"""

    queryset = CardRequest.objects.all()
    serializer_class = CardRequestSerializer
