from pyodbc import Connection

from .models import (
    Account,
    Asset,
    FMILocation,
    Project,
    account_from_peoplesoft,
    asset_from_camera,
    asset_from_door,
    asset_from_elevator,
    asset_from_keybox,
    asset_from_sims,
    location_from_fmi,
    project_from_peoplesoft,
)
from .utils import get_counts_for_key

PEEP_CAMERA_TYPE = "8004"


def get_door_information(sql_client: Connection) -> list[Asset]:
    """
    Queries the CCure database for all Doors and returns them as Assets
    """

    door_query = """
    SELECT
      'CCURE' AS [feed],
      'DOORS' AS [dataSource],
      'CCURE_DB04' AS [serverAddr],
      subsystemType =
                    CASE
                    WHEN [ir].[GUID] IS NULL AND [or].[GUID] IS NULL AND [dlr].[GUID] IS NULL
                      THEN 'Monitor Point'
                    WHEN [doors].[Name] LIKE '%MP' OR [doors].[Name] LIKE '%- MP %'
                      THEN 'Monitor Point'
                    WHEN [ir].[ReaderType] = 4
                      THEN 'Door - Wireless'
                    WHEN  [doors].[Name] LIKE '%WDL' OR [doors].[Name] LIKE '%WCR'
                      THEN 'Door - Wireless'
                    WHEN [doors].[Name] LIKE '%LRR%'
                      THEN 'Door - Gate, LRR'
                    WHEN [doors].[Name] LIKE '%Long Range Reader%'
                      THEN 'Door - Gate, LRR'
                    WHEN [doors].[Name] LIKE '%L.R.R%'
                      THEN 'Door - Gate, LRR'
                    WHEN [doors].[Name] LIKE '%Man Gate%'
                      THEN 'Door - Gate, Man'
                    WHEN [doors].[Name] LIKE '%CG - Emergency Rear Basketball Gate%'
                      THEN 'Door - Gate, Traffic'
                    WHEN [doors].[Name] LIKE '%CG Tennis - Main Entrance Gate Tennis Ct.%'
                      THEN 'Door - Gate, Traffic'
                    WHEN [doors].[Name] LIKE '%Traffic Gate%'
                      THEN 'Door - Gate, Traffic'
                    WHEN [doors].[Name] LIKE '%Finger Barn Gate%'
                      THEN 'Door - Gate, Traffic'
                    WHEN [doors].[Name] LIKE '%Turnstile%'
                      THEN 'Door - Turnstile'
                    WHEN [doors].[Name] LIKE 'ELEV%'
                      THEN 'Door - Elevator'
                    ELSE 'Door'
                    END,
      'iStarDoor' AS [deviceType],
      [doors].[GUID] AS [GUID],
      replace([doors].[Name], '  ', ' ') AS [name],
      [doors].[ClassType] AS [type],
      replace(replace(replace(isnull([doors].Description, ''),
                    CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [comments],
      [isc].[GUID] AS [controllerGUID],
      replace(isnull([isc].[Name], ''), '  ', ' ') AS [controllerName],
      [isc].[ClassType] AS [controllerType],
      replace(replace(replace(isnull([isc].Description, ''),
                 CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [controllerComments],
      [p].[GUID] AS [partitionGUID],
      replace(isnull([p].[Name], ''), '  ', ' ') AS [partitionName],
      [p].[ClassType] AS [partitionType],
      replace(replace(replace(isnull([p].Description, ''),
                 CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [partitionComments],
      [ir].[GUID] AS [readerInGUID],
      replace(isnull([ir].[Name], ''), '  ', ' ') AS [readerInName],
      [ir].[ClassType] AS [readerInType],
      replace(replace(replace(isnull([ir].Description, ''),
                 CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [readerInComments],
      [or].[GUID] AS [readerOutGUID],
      replace(isnull([or].[Name], ''), '  ', ' ') AS [readerOutName],
      [or].[ClassType] AS [readerOutType],
      replace(replace(replace(isnull([or].Description, ''),
                 CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [readerOutComments],
      [dlr].[GUID] AS [doorLockRelayGUID],
      replace(isnull([dlr].[Name], ''), '  ', ' ') AS [doorLockRelayName],
      [dlr].[ClassType] AS [doorLockRelayType],
      replace(replace(replace(isnull([dlr].Description, ''),
                 CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [doorLockRelayComments],
      [ds].[GUID] AS [doorSwitchGUID],
      replace(isnull([ds].[Name], ''), '  ', ' ') AS [doorSwitchName],
      [ds].[ClassType] AS [doorSwitchType],
      replace(replace(replace(isnull([ds].Description, ''),
                 CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [doorSwitchComments],
      [rte].[GUID] AS [reqToExitGUID],
      replace(isnull([rte].[Name], ''), '  ', ' ') AS [reqToExitName],
      [rte].[ClassType] AS [reqToExitType],
          replace(replace(replace(isnull([rte].Description, ''),
                 CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [reqToExitComments]
    FROM [ACVSCore].[Access].[Door] [doors]
      LEFT OUTER JOIN [ACVSCore].[Access].[iStarController] [isc] ON [isc].[ObjectID] = [doors].[ControllerID]
      LEFT OUTER JOIN [ACVSCore].[dbo].[Partition] [p] ON [p].[ObjectID] = [doors].[PartitionID]
      LEFT OUTER JOIN [ACVSCore].[Access].[Reader] [ir] ON [ir].[ObjectID] = [doors].[InReaderID]
      LEFT OUTER JOIN [ACVSCore].[Access].[Reader] [or] ON [or].[ObjectID] = [doors].[OutReaderID]
      LEFT OUTER JOIN [ACVSCore].[Access].[Input] [ds] ON [ds].[ObjectID] = [doors].[DoorSwitchID]
      LEFT OUTER JOIN [ACVSCore].[Access].[Input] [rte] ON [rte].[ObjectID] = [doors].[RTEID]
      LEFT OUTER JOIN [ACVSCore].[Access].[Output] [dlr] ON [dlr].[ObjectID] = [doors].[DoorLockID]
      LEFT OUTER JOIN [ACVSCore].[Access].[DoorUDF] [dudf] ON [dudf].ObjectID = [doors].ObjectID
    """

    result_rows = _execute_sql_query(sql_client, door_query)
    output_data = []
    for row in result_rows:
        output_data.append(asset_from_door(row))
    return output_data


def get_elevator_information(sql_client: Connection) -> list[Asset]:
    """Get elevator data from source sql database and convert it to Assets"""

    elevator_query = """
        SELECT
            'CCURE' AS [feed],
            'ELEVATORS' AS [dataSource],
            'CCURE_DB04' AS [serverAddr],
            'Door - Elevator' AS [subsystemType],
            'Elevator' AS [deviceType],
            [elevators].[GUID] AS [GUID],
            replace( [elevators].[Name], '  ', ' ' ) AS [name],
            [elevators].[ClassType] AS [type],
            replace(
                replace( replace( isnull( [elevators].Description, '' ),
                 CHAR ( 10 ), '' ), CHAR ( 13 ), '\n' ),
                ',',
                ';'
        ) AS [comments],
        [isc].[GUID] AS [controllerGUID],
        replace( isnull( [isc].[Name], '' ), '  ', ' ' ) AS [controllerName],
        [isc].[ClassType] AS [controllerType],
        replace(
            replace( replace( isnull( [isc].Description, '' ),
             CHAR ( 10 ), '' ), CHAR ( 13 ), '\n' ),
            ',',
            ';'
        ) AS [controllerComments],
        [p].[GUID] AS [partitionGUID],
        replace( isnull( [p].[Name], '' ), '  ', ' ' ) AS [partitionName],
        [p].[ClassType] AS [partitionType],
        replace(
            replace( replace( isnull( [p].Description, '' ),
             CHAR ( 10 ), '' ), CHAR ( 13 ), '\n' ),
            ',',
            ';'
        ) AS [partitionComments],
        [ir].[GUID] AS [readerInGUID],
        replace( isnull( [ir].[Name], '' ), '  ', ' ' ) AS [readerInName],
        [ir].[ClassType] AS [readerInType],
        replace(
            replace( replace( isnull( [ir].Description, '' ),
             CHAR ( 10 ), '' ), CHAR ( 13 ), '\n' ),
            ',',
            ';'
        ) AS [readerInComments]
        FROM
            [ACVSCore].[Access].[Elevator] [elevators]
            LEFT OUTER JOIN [ACVSCore].[Access].[iStarController] [isc] ON [isc].[ObjectID] = [elevators].[ControllerID]
            LEFT OUTER JOIN [ACVSCore].[dbo].[Partition] [p] ON [p].[ObjectID] = [elevators].[PartitionID]
            LEFT OUTER JOIN [ACVSCore].[Access].[Reader] [ir] ON [ir].[ObjectID] = [elevators].[ReaderID]
    """

    result_rows = _execute_sql_query(sql_client, elevator_query)
    output_data = []
    for row in result_rows:
        output_data.append(asset_from_elevator(row))
    return output_data


def get_keybox_information(sql_client: Connection) -> list[Asset]:
    """Get keybox asset data from source SQL database and convert it to Assets"""
    keybox_query = """
        SELECT
            'CCURE' AS [feed],
            'KEY_CABINET' AS [dataSource],
            'CCURE_DEISTER' AS [serverAddr],
            subsystemType = 'Door - Key Cabinet',
            'Key Cabinet' AS [deviceType],
            [kc].[GUID] AS [GUID],
            replace( [kc].[Name], '  ', ' ' ) AS [name],
            [kc].[ClassType] AS [type],
            replace(
                replace( replace( isnull( [kc].Description, '' ),
                         CHAR ( 10 ), '' ),
                CHAR ( 13 ), '\n' ),
                ',',
                ';'
        ) AS [comments],
        [p].[GUID] AS [partitionGUID],
        replace( isnull( [p].[Name], '' ), '  ', ' ' ) AS [partitionName],
        [p].[ClassType] AS [partitionType],
        replace(
            replace( replace( isnull( [p].Description, '' ),
                     CHAR ( 10 ), '' ),
            CHAR ( 13 ), '\n' ),
            ',',
            ';'
        ) AS [partitionComments]
        FROM
            [ACVSCore].[NMS].[proxSafeCabinet] [kc]
            LEFT OUTER JOIN [ACVSCore].[dbo].[Partition] [p]
                ON [p].[ObjectID] = [kc].[PartitionID]
        UNION
        SELECT
            'CCURE' AS [feed],
            'KEY_TERMINAL' AS [dataSource],
            'ehps-c9kdb2.ehps.ncsu.edu\\cc9000db2' AS [serverAddr],
            subsystemType = 'Door - Key Terminal',
            'Key Terminal' AS [deviceType],
            [terminal].[GUID] AS [GUID],
            replace( [terminal].[TerminalDisplayName], '  ', ' ' ) AS [name],
            [terminal].[ClassType] AS [type],
            replace(
                replace( replace( isnull( [terminal].Description, '' ),
                                 CHAR ( 10 ), '' ),
                CHAR ( 13 ), '\n' ),
                ',',
                ';'
        ) AS [comments],
        [p].[GUID] AS [partitionGUID],
        replace( isnull( [p].[Name], '' ), '  ', ' ' ) AS [partitionName],
        [p].[ClassType] AS [partitionType],
        replace(
            replace( replace( isnull( [p].Description, '' ),
                             CHAR ( 10 ), '' ),
            CHAR ( 13 ), '\n' ),
            ',',
            ';'
        ) AS [partitionComments]
        FROM
            [ACVSCore].[NMS].[proxSafeTerminal] [terminal]
            LEFT OUTER JOIN [ACVSCore].[dbo].[Partition] [p]
                ON [p].[ObjectID] = [terminal].[PartitionID]
    """

    result_rows = _execute_sql_query(sql_client, keybox_query)
    output_data = []
    for row in result_rows:
        output_data.append(asset_from_keybox(row))
    return output_data


def filter_peep_cameras(raw_result: list[dict]):
    return [_ for _ in raw_result if str(_["subsystemType"]) != PEEP_CAMERA_TYPE]


def filter_cameras_by_mac_groups(raw_result: list[dict], mac_groups: dict):
    """
    Filters different groups of cameras that share a mac address based on the number of cameras in the group

    There are some camera models that contain multiple different views in one physical camera.
    The GENETEC management system sees each of these views as a separate object,
    and returns them all with different IDs.
    The Asset system would then see them as different unique assets,
    and bill customers based on all of them.

    Customers should only be charged once for these types of cameras,
    so only one asset should come through.
    The asset that does come through should be consistent for all groups.
    If the cameras are named `camera-360V1`, `camera-360V2`, `camera-360V3`,
    then `camera-360V1` should come through for every camera name.

    There is also a separate type of camera that shares MAC addresses, the ENC cameras.
    These are large groupings ( >= 6 cameras) of analogue cameras,
    tied into the GENETEC system via a single network box.
    These SHOULD be billed as separate cameras,
    but since they are legacy installations that are being discontinued,
    they will simply be completely ignored by the pipeline and taken out of service manually.
    """

    sorted_raw_result = sorted(raw_result, key=lambda element: element["name"])

    already_picked_unique_macs = []
    output_rows = []

    for row in sorted_raw_result:
        row_mac = row["mac"]
        if mac_groups[row_mac] == 1:
            output_rows.append(row)
        elif mac_groups[row_mac] == 2:
            if (
                "enc" in row["name"].lower()
            ):  #  ENC camera groups should be completely ignored, they are managed manually
                pass
            else:
                if row_mac not in already_picked_unique_macs:
                    output_rows.append(row)
                    already_picked_unique_macs.append(row_mac)
        elif 3 <= mac_groups[row_mac] <= 5:
            if row_mac not in already_picked_unique_macs:
                output_rows.append(row)
                already_picked_unique_macs.append(row_mac)
        elif mac_groups[row_mac] >= 6:
            pass
    return output_rows


def get_camera_information(sql_client: Connection) -> list[Asset]:
    """Get camera data from source SQL db and convert it to Assets"""
    # TODO Could be some stuff we should remove here, like `FIXTHIS` being hardcoded to the `serverAddr` field
    camera_query = """SELECT
        'GENETEC' AS [feed],
              'GENETEC' AS [dataSource],
              'FIXTHIS' AS [serverAddr],
              CASE
              WHEN (CHARINDEX('PTZ', [ENT].[Name]) != 0)
                THEN 'Camera - PTZ'
                    WHEN   [ENT].Name like '%180VL%'
                        THEN 'Camera - 180 View Left'
                    WHEN   [ENT].Name like '%180VC%'
                        THEN 'Camera - 180 View Left'
                    WHEN   [ENT].Name like '%180VR%'
                        THEN 'Camera - 180 View Left'
                    WHEN   [ENT].Name like '%360V1%'
                        THEN 'Camera - 360 View 1'
                    WHEN   [ENT].Name like '%360V2%'
                        THEN 'Camera - 360 View 2'
                    WHEN   [ENT].Name like '%360V3%'
                        THEN 'Camera - 360 View 3'
                    WHEN   [ENT].Name like '%360V4%'
                        THEN 'Camera - 360 View 4'
                    WHEN   [ENT].Name like '%360QV%'
                        THEN 'Camera - 360 Quad View'
              ELSE 'Camera - Fixed'
              END AS [subsystemType],
        'Video Unit' AS [deviceType],
        [ENT].[GUID] AS [GUID],
        [ENT].Name as [name],
        [ENT].Type as [type],
        [ENT].Description as [comments],
        [ENT].Type,
        LEFT((CAST([UNT].Data as XML)).value('(/VU/PhyAdd)[1]', 'nvarchar(1000)'), 2) + ':'
        + SUBSTRING((CAST([UNT].Data as XML)).value('(/VU/PhyAdd)[1]', 'nvarchar(1000)'), 3, 2) + ':'
        + SUBSTRING((CAST([UNT].Data as XML)).value('(/VU/PhyAdd)[1]', 'nvarchar(1000)'), 5, 2) + ':'
        + SUBSTRING((CAST([UNT].Data as XML)).value('(/VU/PhyAdd)[1]', 'nvarchar(1000)'), 7, 2) + ':'
        + SUBSTRING((CAST([UNT].Data as XML)).value('(/VU/PhyAdd)[1]', 'nvarchar(1000)'), 9, 2) + ':'
        + SUBSTRING((CAST([UNT].Data as XML)).value('(/VU/PhyAdd)[1]', 'nvarchar(1000)'), 11, 2) as [mac],
        [UNT].IP as [ip],
        (CAST([UNT].Data as XML)).value('(/VU/CPort)[1]', 'nvarchar(1000)') as [port],
        [UNT].[Guid] as [videoUnitGUID],
        (CAST([UNT].Data as XML)).value('(/VU/UDI/E/Name)[1]', 'nvarchar(1000)') as [videoUnitName],
        (CAST([UNT].Data as XML)).value('(/VU/Mfct)[1]', 'nvarchar(1000)') as [videoUnitManufacturer],
        (CAST([UNT].Data as XML)).value('(/VU/Model)[1]', 'nvarchar(max)') as [videoUnitModel],
        [ENT].CreationTime as [firstEvent],
        [UNT].FirmwareVersion as [videoUnitFirmware]
        FROM Directory.dbo.Entity AS [ENT]
            LEFT OUTER JOIN Directory.dbo.Device [DEV] ON [DEV].[Guid] = [ENT].[Guid]
          LEFT OUTER JOIN Directory.dbo.Unit [UNT] ON [DEV].[Unit] = [UNT].[Guid]
          where [ENT].Type = '6'
            AND [ENT].SubType='5'
          ORDER BY [UNT].IP
    """

    result_rows = _execute_sql_query(sql_client, camera_query)
    mac_counts = get_counts_for_key("mac", result_rows)

    cameras_peep_excluded = filter_peep_cameras(result_rows)

    output_rows = filter_cameras_by_mac_groups(cameras_peep_excluded, mac_counts)

    return [asset_from_camera(row) for row in output_rows]


def get_sims_information(sql_client: Connection) -> list[Asset]:
    sims_query = """
        SELECT
          'SIMS' AS [feed],
          'SIMS III' AS [dataSource],
          '10.118.228.18\\SIMSData' AS [serverAddr],
          'Burg Panel' AS [subsystemType],
          'SIMS' AS [deviceType],
          [aMain].[AccountNumber] AS [GUID],
          replace([aMain].[PremiseName], '  ', ' ') AS [name],
          [cType].[ControlType] AS [type],
          replace(replace(replace(isnull([aMain].[PremiseNameContinued], ''),
                          CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [comments],
          [aType].[Type] AS [locationType],
          replace(replace(replace(isnull([aMain].[PremiseCrossStreet], ''),
                          CHAR(10), ''), CHAR(13), '\n'), ',', ';') AS [simsLocation],
          '' AS [ip],
          '' AS [mac],
          [aMain].[FirstEventDate] as [firstEvent]
        FROM [SIMSData].[dbo].[AccountMain] [aMain]
          LEFT OUTER JOIN [SIMSData].[dbo].[AccountType] [aType] ON [aType].[DisplayOrder] = [aMain].[AccountTypeID]
          LEFT OUTER JOIN [SIMSData].[dbo].[ControlType] [cType] ON [cType].[DisplayOrder] = [aMain].[ControlTypeID]
        where [aMain].[AccountNumber] LIKE '01-00%'
        ORDER BY [aMain].[PremiseName]
    """

    result_rows = _execute_sql_query(sql_client, sims_query)
    return [asset_from_sims(row) for row in result_rows]


def _execute_sql_query(sql_client: Connection, sql_query: str) -> list[dict]:
    """Executes an SQL query and returns the result as a list of dicts keyed off the column names"""
    cursor = sql_client.cursor()
    cursor.execute(sql_query)
    column_names = [column[0] for column in cursor.description]
    result_rows = [dict(zip(column_names, row)) for row in cursor.fetchall()]
    cursor.close()
    return result_rows


def get_locations(sql_client: Connection) -> list[FMILocation]:
    fmi_query = """SELECT * FROM NCSU_SAT_BUILDINGS"""
    result_rows = _execute_sql_query(sql_client, fmi_query)
    return [location_from_fmi(row) for row in result_rows]


def get_accounts(sql_client: Connection) -> list[Account]:
    account_query = """SELECT * FROM PS_NC_ACCOUNT_VW"""
    result_rows = _execute_sql_query(sql_client, account_query)
    return [account_from_peoplesoft(row) for row in result_rows]


def get_projects(sql_client: Connection) -> list[Project]:
    project_query = """SELECT * FROM PS_NC_PROJ_DESC_VW"""
    result_rows = _execute_sql_query(sql_client, project_query)
    return [project_from_peoplesoft(row) for row in result_rows]
