import csv
import datetime

from django.db import models
from model_utils.models import TimeStampedModel

from sat_automations.asset_automation.utils_2n import normalize_mac_address
from sat_automations.asset_automation.utils_models import (
    get_building_abbreviation_from_name,
    get_room_number_from_name,
)


class Account(TimeStampedModel):
    account = models.CharField(max_length=512, unique=False, default=None, null=True, blank=True)
    effective_status = models.CharField(max_length=12, blank=True, null=True)
    description = models.CharField(max_length=512, blank=True, null=True)


class Asset(models.Model):
    """
    Represents a security asset

    Contains the superset of all fields needed by every supported asset type
    """

    # Core IDs
    dynamics_id = models.CharField(
        max_length=512, unique=True, default=None, db_index=True, null=True, blank=True
    )
    source_guid = models.CharField(max_length=512, unique=True, db_index=True)

    # Core asset attributes
    ip = models.CharField(max_length=512, blank=True, null=True)
    mac = models.CharField(max_length=512, blank=True, null=True)
    name = models.CharField(max_length=512, blank=True, null=True)
    comments = models.CharField(max_length=512, blank=True, null=True)

    # Core Billing
    inventory_subsystem_type_id = models.IntegerField(null=True, blank=True)
    inventory_system_type_id = models.IntegerField(null=True, blank=True)
    suffix = models.CharField(max_length=512, blank=True, null=True)

    sims_name = models.CharField(max_length=512, blank=True, null=True)

    controller_comments = models.CharField(max_length=512, blank=True, null=True)
    controller_guid = models.CharField(max_length=512, blank=True, null=True)
    controller_name = models.CharField(max_length=512, blank=True, null=True)
    controller_type = models.CharField(max_length=512, blank=True, null=True)

    data_source = models.CharField(max_length=512, blank=True, null=True)
    device_type = models.CharField(max_length=512, blank=True, null=True)
    subsystem_type = models.CharField(max_length=512, blank=True, null=True)
    other_type = models.CharField(max_length=512, blank=True, null=True)

    # Building info, derived from prefix(es) in the asset name
    # We can either just include the building abbreviation (or number) in the generated flat file,
    # or we can implement the mapping logic here to include all the building info (Latitude, Longitude, address, etc.).
    # TODO Communicate with Longview about which approach they can deal with
    room_code = models.CharField(max_length=512, blank=True, null=True)
    building_abbreviation = models.CharField(max_length=10, blank=True, null=True)

    def __repr__(self):
        return f"{self.source_guid}: {self.name}"


class FMILocation(TimeStampedModel):
    building_abbr = models.CharField(max_length=100, blank=True)
    building_number = models.CharField(max_length=100, blank=True)
    building_name = models.CharField(max_length=100, blank=True)
    address = models.CharField(max_length=100, blank=True, null=True)
    city = models.CharField(max_length=100, blank=True, null=True)
    county = models.CharField(max_length=100, blank=True, null=True)
    state = models.CharField(max_length=100, blank=True, null=True)
    zip = models.CharField(max_length=100, blank=True, null=True)
    country = models.CharField(max_length=100, blank=True, null=True)
    residence_type = models.CharField(max_length=100, blank=True, null=True)
    use = models.CharField(max_length=100, blank=True, null=True)
    condition = models.CharField(max_length=100, blank=True, null=True)
    status = models.CharField(max_length=100, blank=True, null=True)
    campus = models.CharField(max_length=100, blank=True, null=True)
    latitude = models.CharField(max_length=100, blank=True, null=True)
    longitude = models.CharField(max_length=100, blank=True, null=True)
    site_code = models.CharField(max_length=100, blank=True, null=True)
    funding_type = models.CharField(max_length=100, blank=True, null=True)

    class Meta:
        unique_together = ("building_abbr", "building_number")


class Project(TimeStampedModel):
    project_id = models.CharField(max_length=512, unique=False, default=None, null=True, blank=True)
    effective_date = models.DateField()
    nc_project_reference = models.CharField(max_length=512, blank=True, null=True)
    nc_project_segment_reference = models.CharField(max_length=512, blank=True, null=True)
    nc_effective_status = models.CharField(max_length=12, blank=True, null=True)
    description = models.CharField(max_length=512, blank=True, null=True)
    start_date = models.DateField(blank=True, null=True)
    end_date = models.DateField(blank=True, null=True)
    nc_responsible = models.CharField(max_length=512, blank=True, null=True)
    fund_code = models.CharField(max_length=512, blank=True, null=True)
    department_id = models.CharField(max_length=512, blank=True, null=True)
    program_code = models.CharField(max_length=512, blank=True, null=True)
    class_field = models.CharField(max_length=512, blank=True, null=True)
    nc_description = models.CharField(max_length=1024, blank=True, null=True)


def account_from_peoplesoft(account_data: dict) -> Account:
    return Account(
        account=account_data["ACCOUNT"],
        effective_status=account_data["EFF_STATUS"],
        description=account_data["DESCR"],
    )


def project_from_peoplesoft(project_data: dict) -> Project:
    return Project(
        project_id=project_data["PROJECT_ID"],
        effective_date=project_data["EFFDT"],
        nc_project_reference=project_data["NC_PROJECT_PROJREF"],
        nc_project_segment_reference=project_data["NC_PROJECT_SEGREF"],
        nc_effective_status=project_data["NC_EFF_STATUS"],
        description=project_data["DESCR"],
        start_date=project_data["START_DT"],
        end_date=project_data["END_DT"],
        nc_responsible=project_data["NC_RESPONSIBLE"],
        fund_code=project_data["FUND_CODE"],
        department_id=project_data["DEPTID"],
        program_code=project_data["PROGRAM_CODE"],
        class_field=project_data["CLASS_FLD"],
        nc_description=project_data["NC_DESCR600"],
    )


def asset_from_keybox(keybox: dict) -> Asset:
    room_code = get_room_number_from_name(keybox["name"])
    building_abbreviation = get_building_abbreviation_from_name(keybox["name"])

    return Asset(
        source_guid=keybox["GUID"],
        name=keybox["name"],
        comments=keybox["comments"],
        subsystem_type=keybox["subsystemType"],
        # feed=keybox['feed'],
        device_type=keybox["deviceType"],
        data_source=keybox["dataSource"],
        other_type=keybox["type"],
        room_code=room_code,
        building_abbreviation=building_abbreviation,
    )


def asset_from_door(door: dict) -> Asset:
    room_code = get_room_number_from_name(door["name"])
    building_abbreviation = get_building_abbreviation_from_name(door["name"])

    return Asset(
        source_guid=door["GUID"],
        name=door["name"],
        comments=door["comments"],
        controller_comments=door["controllerComments"],
        controller_guid=door["controllerGUID"],
        controller_name=door["controllerName"],
        controller_type=door["controllerType"],
        data_source=door["dataSource"],
        device_type=door["deviceType"],
        subsystem_type=door["subsystemType"],
        other_type=door["type"],
        building_abbreviation=building_abbreviation,
        room_code=room_code,
    )


def asset_from_elevator(elevator: dict) -> Asset:
    room_code = get_room_number_from_name(elevator["name"])
    building_abbreviation = get_building_abbreviation_from_name(elevator["name"])
    return Asset(
        source_guid=elevator["GUID"],
        name=elevator["name"],
        comments=elevator["comments"],
        controller_comments=elevator["controllerComments"],
        controller_guid=elevator["controllerGUID"],
        controller_name=elevator["controllerName"],
        controller_type=elevator["controllerType"],
        data_source=elevator["dataSource"],
        device_type=elevator["deviceType"],
        subsystem_type=elevator["subsystemType"],
        other_type=elevator["type"],
        room_code=room_code,
        building_abbreviation=building_abbreviation,
    )


def asset_from_camera(camera: dict) -> Asset:
    room_code = get_room_number_from_name(camera["name"])
    building_abbreviation = get_building_abbreviation_from_name(camera["name"])

    return Asset(
        source_guid=camera["GUID"],
        ip=camera["ip"],
        mac=camera["mac"],
        name=camera["name"],
        comments=camera["comments"],
        data_source=camera["dataSource"],
        device_type=camera["deviceType"],
        subsystem_type=camera["subsystemType"],
        other_type=camera["type"],
        room_code=room_code,
        building_abbreviation=building_abbreviation,
    )


def asset_from_sims(sims: dict) -> Asset:
    room_code = get_room_number_from_name(sims["name"])
    building_abbreviation = get_building_abbreviation_from_name(sims["name"])
    return Asset(
        source_guid=sims["GUID"],
        ip=sims["ip"],
        mac=sims["mac"],
        name=sims["name"],
        comments=sims["comments"],
        data_source=sims["dataSource"],
        device_type=sims["deviceType"],
        subsystem_type=sims["subsystemType"],
        other_type=sims["type"],
        room_code=room_code,
        building_abbreviation=building_abbreviation,
    )


def asset_from_2n_intercom(intercom: dict) -> Asset:
    last_edit_date = datetime.datetime.strptime(intercom["last-edit-date"], "%m/%d/%Y")
    comments = f"{intercom['location']}, {intercom['phone']}, Last Touched {last_edit_date.strftime('%b %Y')}"
    return Asset(
        source_guid=intercom["intercomGUID"],
        name=intercom["intercomName"],
        type=intercom["intercomType"],
        feed=intercom["feed"],
        data_source=intercom["dataSource"],
        subsystem_type=intercom["subsystemType"],
        device_type=intercom["deviceType"],
        ip=intercom["ip"],
        mac=intercom["mac"],
        comments=comments,
    )


def asset_from_aiphone_intercom(intercom: dict) -> Asset:
    comment_fields = ("notes", "multiCast", "keyForSoftClient", "videoPortRange")
    comments = " ".join([intercom[field].strip() for field in comment_fields])
    return Asset(
        source_guid=intercom["intercomGUID"],
        name=intercom["intercomName"],
        type=intercom["intercomType"],
        feed=intercom["feed"],
        data_source=intercom["dataSource"],
        subsystem_type=intercom["subsystemType"],
        device_type=intercom["deviceType"],
        ip=intercom["ip"],
        mac=intercom["mac"],
        comments=comments,
    )


def asset_from_commend_intercom(intercom: dict) -> Asset:
    return Asset(
        source_guid=intercom["intercomGUID"],
        name=intercom["intercomName"],
        type=intercom["intercomType"],
        feed=intercom["feed"],
        data_source=intercom["dataSource"],
        subsystem_type=intercom["subsystemType"],
        device_type=intercom["deviceType"],
        ip=intercom["ip"],
        mac=intercom["mac"],
        comments=intercom["notes"].strip(),
    )


def account_to_dynamic_365(account: Account) -> dict:
    return {
        "Account": account.account,
        "Effective Status": account.effective_status,
        "Description": account.description,
    }


def asset_to_dynamic_365(asset: Asset) -> dict:
    """Converts an Asset object into a dictionary formatted for the dynamics 365 upload"""
    return {
        "ID": asset.dynamics_id,  # TODO Not sure what this is supposed to be now? The ID they generate?
        "GUID": asset.source_guid,
        "Data source": asset.data_source,  # TODO potentially rename to Parent System Name
        "IP Address": asset.ip,
        "MAC Address": asset.mac,
        "Name": asset.name,
        "Suffix": asset.suffix,
        "Comments": asset.comments,
        "Inventory Subsystem Type ID": asset.inventory_subsystem_type_id,
        "Inventory System Type ID": asset.inventory_system_type_id,
        "Device Type": asset.device_type,
        "Subsystem Type": asset.subsystem_type,
        "Controller Comments": asset.controller_comments,
        "Controller GUID": asset.controller_guid,
        "Controller Name": asset.controller_name,
        "Controller Type": asset.controller_type,
        "SIMS Name": asset.sims_name,
        "Building Abbreviation": asset.building_abbreviation,
        "Room(s)": asset.room_code,
        "In-Service Date": None,
        "Warranty Exp Date": None,
        "Out-Of-Service Date": None,
    }


def location_to_dynamic_365(location: FMILocation) -> dict:
    """Converts a location object to a dictionary format for the dynamics 365 CSV upload"""
    return {
        "Building Number": location.building_number,
        "Building Abbreviation": location.building_abbr,
        "Name": location.building_name,
        "Address": location.address,
        "City": location.city,
        "County": location.county,
        "State": location.state,
        "Zip Code": location.zip,
        "Country": location.country,
        "Residence Type": location.residence_type,
        "Use": location.use,
        "Condition": location.condition,
        "Status": location.status,
        "Campus": location.campus,
        "Latitude": location.latitude,
        "Longitude": location.longitude,
        "Site Code": location.site_code,
        "Funding Type": location.funding_type,
    }


def project_to_dynamic_365(project: Project) -> dict:
    return {
        "ID": project.project_id,
        "Effective Date": project.effective_date.isoformat(),
        "NC Project Reference": project.nc_project_reference,
        "NC Effective Status": project.nc_effective_status,
        "Description": project.description,
        "Start Date": project.start_date.isoformat(),
        "End Date": project.end_date.isoformat(),
        "NC Responsible": project.nc_responsible,
        "Fund Code": project.fund_code,
        "Department ID": project.department_id,
        "Program Code": project.program_code,
        "Class Field": project.class_field,
        "NC Description": project.nc_description,
    }


def assets_to_csv(assets: list, file_object) -> None:
    """Converts a list of Assets to a file like object, formatted as a CSV"""
    dynamics_formatted_assets = [asset_to_dynamic_365(asset) for asset in assets]

    writer = csv.DictWriter(file_object, fieldnames=dynamics_formatted_assets[0].keys())
    writer.writeheader()
    writer.writerows(dynamics_formatted_assets)
    file_object.seek(0)


def get_assets_from_object_store() -> list[Asset]:
    return Asset.objects.all()


def assets_to_object_store(assets: list[Asset]) -> None:
    Asset.objects.bulk_create(assets)


def asset_from_commander_intercom_api(intercom: dict) -> Asset:
    room_code = get_room_number_from_name(intercom["name"])
    building_abbreviation = get_building_abbreviation_from_name(intercom["name"])

    return Asset(
        source_guid=intercom["Hardware"]["SerialNumber"],
        name=intercom["Name"],
        ip=intercom["IpAddress"],
        mac=normalize_mac_address(intercom["Hardware"]["MacAddress"]),
        feed="INTERCOM",
        data_source="INTERCOM",
        subsystem_type="Intercom 2N",
        room_code=room_code,
        building_abbreviation=building_abbreviation,
    )


def location_from_fmi(location: dict) -> FMILocation:
    return FMILocation(
        address=location["ADDRESS"],
        building_abbr=location["BLDG_ABBR"],
        building_name=location["BLDG_NAME"],
        building_number=location["BLDG_NUM"],
        city=location["CITY"],
        county=location["BLDG_COUNTY_NAME"],
        state=location["STATE"],
        zip=location["ZIP"],
        country=location["BLDG_COUNTRY"],
        residence_type=location["RESID_NAME"],
        use=location["USE_NAME"],
        condition=location["COND_NAME"],
        status=location["STATUS_DESCR"],
        funding_type=location["FUND_NAME"],
        campus=location["CAMPUS"],
        latitude=location["LATITUDE"],
        longitude=location["LONGITUDE"],
    )
