import csv

import cx_Oracle
from azure.storage.blob import BlobServiceClient
from celery import shared_task
from sat.db import ConnectionType, SatDBException, get_db_connection
from sat.logs import SATLogger

from sat_automations.asset_automation import helper_queries
from sat_automations.asset_automation.models import (
    Account,
    Asset,
    FMILocation,
    Project,
    asset_to_dynamic_365,
    location_to_dynamic_365,
)
from sat_automations.asset_automation.utils import (
    CsvOpener,
    diff_assets,
    diff_locations,
    upload_accounts_to_azure,
    upload_assets_to_azure,
    upload_locations_to_azure,
    upload_projects_to_azure,
)
from sat_automations.utils.common.connection import (
    AZURE_BLOB_ACCOUNT_URL,
    AZURE_BLOB_SHARED_ACCESS_KEY,
    CCURE_CONNECTION,
    DEISTER_CONNECTION,
    FMI_CONNECTION,
    FORPT_CONNECTION,
    GENETEC_CONNECTION,
    SIMS_CONNECTION,
)

logger = SATLogger(__name__)


NEW_LOCATIONS_FILE = "new-locations-{}.csv"
UPDATED_LOCATIONS_FILE = "updated-locations-{}.csv"
DELETED_LOCATIONS_FILE = "deleted-locations-{}.csv"

NEW_ASSETS_FILE = "new-assets-{}.csv"
UPDATED_ASSETS_FILE = "updated-assets-{}.csv"
DELETED_ASSETS_FILE = "deleted-assets-{}.csv"

NEW_ACCOUNTS_FILE = "new-accounts-{}.csv"

NEW_PROJECTS_FILE = "new-projects-{}.csv"


class AssetAutomation:
    def __init__(self) -> None:
        self.blob_service_client = BlobServiceClient(
            AZURE_BLOB_ACCOUNT_URL, AZURE_BLOB_SHARED_ACCESS_KEY
        )
        containers = [_ for _ in self.blob_service_client.list_containers()]
        ads_container = containers[0]["name"]
        self.azure_client = self.blob_service_client.get_container_client(ads_container)

        try:
            self.ccure_conn = get_db_connection(ConnectionType.SQL, conn_string=CCURE_CONNECTION)
            self.diester_conn = get_db_connection(
                ConnectionType.SQL, conn_string=DEISTER_CONNECTION
            )
            self.genetec_conn = get_db_connection(
                ConnectionType.SQL, conn_string=GENETEC_CONNECTION
            )
            self.sims_conn = get_db_connection(ConnectionType.SQL, conn_string=SIMS_CONNECTION)
            self.fmi_conn = get_db_connection(ConnectionType.SQL, conn_string=FMI_CONNECTION)

            # Manually setting up the cx_Oracle connection here, the connection initialization in
            # the sat.db module tries to use a conn-string instead of a conn-dict,
            # and the cx_Oracle.connect method doesn't support that behavior.
            self.forpt_conn = cx_Oracle.connect(
                user=FORPT_CONNECTION["user"],
                password=FORPT_CONNECTION["password"],
                dsn=FORPT_CONNECTION["dsn"],
                encoding="UTF-8",
            )

        except SatDBException as error:
            logger.error(f"{error}")
            raise ConnectionError(f"Unable to connect to database: {error}")


@shared_task
def update_object_store():
    asset_automation = AssetAutomation()
    total_source_assets = list()
    total_source_assets += helper_queries.get_door_information(asset_automation.ccure_conn)
    total_source_assets += helper_queries.get_elevator_information(asset_automation.ccure_conn)
    total_source_assets += helper_queries.get_keybox_information(asset_automation.diester_conn)
    total_source_assets += helper_queries.get_camera_information(asset_automation.genetec_conn)
    total_source_assets += helper_queries.get_sims_information(asset_automation.sims_conn)

    datastore_assets = Asset.objects.all()
    assets_to_create, assets_to_update, assets_to_delete = diff_assets(
        total_source_assets, datastore_assets
    )
    azure_container_client = asset_automation.blob_service_client.get_container_client(
        "incomingassets"
    )

    if assets_to_create:
        logger.info(f"Identified {len(assets_to_create)} assets to add to the datastore")
        Asset.objects.bulk_create(assets_to_create)
        logger.info("Added new assets to the Django datastore")
        upload_assets_to_azure(assets_to_create, azure_container_client)

    if assets_to_update:
        logger.info(
            f"Identified {len(assets_to_update)} assets with values that need to be updated"
        )
        update_field_names = [
            field.attname for field in Asset._meta.get_fields() if field.attname != "id"
        ]
        Asset.objects.bulk_update(assets_to_update, fields=update_field_names)
        upload_assets_to_azure(assets_to_update, azure_container_client)
        logger.info("Updated assets in Django datastore")

    if assets_to_delete:
        logger.info(f"Identified {len(assets_to_delete)} assets that need to be deleted")
        ids_to_delete = [asset.id for asset in assets_to_delete]
        Asset.objects.filter(pk__in=ids_to_delete).delete()
        logger.info("Deleted assets from Django datastore")


@shared_task
def update_location_store():
    asset_automation = AssetAutomation()
    source_locations = helper_queries.get_locations(asset_automation.fmi_conn)

    datastore_locations = FMILocation.objects.all()

    locations_to_create, locations_to_update, locations_to_delete = diff_locations(
        source_locations, datastore_locations
    )
    location_container_client = asset_automation.blob_service_client.get_container_client(
        "incominglocations"
    )

    if locations_to_create:
        logger.info(f"Identified {len(locations_to_create)} locations to add to the datastore")
        FMILocation.objects.bulk_create(locations_to_create)
        upload_locations_to_azure(
            locations_to_create, location_container_client, NEW_LOCATIONS_FILE
        )
        logger.info("Added new locations to the Django datastore")

    if locations_to_update:
        logger.info(
            f"Identified {len(locations_to_update)} locations with values that need to be updated"
        )
        upload_locations_to_azure(
            locations_to_update, location_container_client, UPDATED_LOCATIONS_FILE
        )
        update_field_names = [
            field.attname for field in FMILocation._meta.get_fields() if field.attname != "id"
        ]
        FMILocation.objects.bulk_update(locations_to_update, fields=update_field_names)
        logger.info("Updated locations in Django datastore")

    if locations_to_delete:
        logger.info(f"Identified {len(locations_to_delete)} locations that need to be deleted")
        upload_locations_to_azure(
            locations_to_delete, location_container_client, DELETED_LOCATIONS_FILE
        )
        ids_to_delete = [location.id for location in locations_to_delete]
        FMILocation.objects.filter(pk__in=ids_to_delete).delete()
        logger.info("Deleted locations from Django datastore")


@shared_task
def update_projects():
    asset_automation = AssetAutomation()
    logger.info("Retrieving projects from Peoplesoft")
    new_projects = helper_queries.get_projects(asset_automation.forpt_conn)
    Project.objects.all().delete()
    Project.objects.bulk_create(new_projects)
    logger.info(f"Updated datastore with {len(new_projects)} projects")
    azure_container_client = asset_automation.blob_service_client.get_container_client(
        "incomingprojects"
    )
    upload_projects_to_azure(new_projects, azure_container_client, NEW_PROJECTS_FILE)
    logger.info(f"Uploaded {len(new_projects)} projects to azure blob")


@shared_task
def update_accounts():
    asset_automation = AssetAutomation()
    logger.info("Retrieving accounts from Peoplesoft")
    new_accounts = helper_queries.get_accounts(asset_automation.forpt_conn)
    Account.objects.all().delete()
    Account.objects.bulk_create(new_accounts)
    logger.info(f"Updated datastore with {len(new_accounts)} accounts")
    azure_container_client = asset_automation.blob_service_client.get_container_client(
        "incomingaccounts"
    )
    upload_accounts_to_azure(new_accounts, azure_container_client, NEW_ACCOUNTS_FILE)
    logger.info(f"Uploaded {len(new_accounts)} accounts to azure blob")


def generate_csv_file(objects: list[dict], csv_handler: CsvOpener):
    upload_field_names = tuple(objects[0].keys())
    csv_writer = csv.DictWriter(csv_handler, upload_field_names)
    csv_writer.writeheader()
    csv_writer.writerows(objects)


@shared_task
def generate_local_csv():
    asset_automation = AssetAutomation()
    total_source_assets = list()
    total_source_assets += helper_queries.get_door_information(asset_automation.ccure_conn)
    total_source_assets += helper_queries.get_elevator_information(asset_automation.ccure_conn)
    total_source_assets += helper_queries.get_keybox_information(asset_automation.diester_conn)
    total_source_assets += helper_queries.get_camera_information(asset_automation.genetec_conn)
    total_source_assets += helper_queries.get_sims_information(asset_automation.sims_conn)

    logger.info("Deleting asset datastore")
    Asset.objects.all().delete()
    logger.info("Successfully deleted asset datastore")

    logger.info(
        f"Generating new asset datastore with {len(total_source_assets)} assets pulled from sources"
    )
    Asset.objects.bulk_create(total_source_assets)
    logger.info("Added new assets to the Django datastore")

    logger.info("Convert assets to dynamics format")
    dynamics_assets = [asset_to_dynamic_365(asset) for asset in total_source_assets]

    logger.info("Write assets to local csv file")
    with CsvOpener("local-assets.csv") as asset_csv:
        generate_csv_file(dynamics_assets, asset_csv)

    logger.info("Deleting Location datastore")
    FMILocation.objects.all().delete()

    logger.info("Retrieving source locations")
    source_locations = helper_queries.get_locations(asset_automation.fmi_conn)
    logger.info("Creating location datastore")
    FMILocation.objects.bulk_create(source_locations)

    logger.info("Transforming locations to dynamics format")
    dynamics_locations = [location_to_dynamic_365(location) for location in source_locations]

    logger.info("Write locations to csv file")
    with CsvOpener("local-locations.csv") as location_csv:
        generate_csv_file(dynamics_locations, location_csv)
