import pytest

from sat_automations.asset_automation.models import Asset, FMILocation
from sat_automations.asset_automation.utils import diff_assets, diff_locations

# Mark all tests here as requiring the setup and teardown of a test database
pytestmark = pytest.mark.django_db


def test_only_new_asset():
    """A single source asset and no datastore assets should only result in that asset in the create-list"""
    asset = Asset(source_guid="1")
    source_assets = [asset]
    store_assets = []

    create_assets, update_assets, delete_assets = diff_assets(source_assets, store_assets)

    assert create_assets == source_assets
    assert update_assets == []
    assert delete_assets == []


def test_only_existing_asset():
    """A single source asset and a single datastore asset with the same guid should result in no changes"""

    old_asset = Asset(source_guid="1", name="test")
    store_assets = [old_asset]

    new_asset = Asset(source_guid="1", name="test")
    source_assets = [new_asset]

    create_assets, update_assets, delete_assets = diff_assets(source_assets, store_assets)

    assert create_assets == []
    assert update_assets == []
    assert delete_assets == []


def test_update_asset():
    """A single source asset and a single datastore asset with the same guid but different data should result in an update"""
    old_asset = Asset(source_guid="1", name="old")
    updated_asset = Asset(source_guid="1", name="new")
    source_assets = [updated_asset]
    store_assets = [old_asset]

    create_assets, update_assets, delete_assets = diff_assets(source_assets, store_assets)

    assert create_assets == []
    assert update_assets == [updated_asset]
    assert delete_assets == []


def test_delete_asset():
    """If there is a datastore asset but no source asset, the datastore asset should be in the delete-list"""
    asset = Asset(source_guid="1")
    source_assets = []
    store_assets = [asset]

    create_assets, update_assets, delete_assets = diff_assets(source_assets, store_assets)

    assert create_assets == []
    assert update_assets == []
    assert delete_assets == [asset]


def test_new_location_only():
    location = FMILocation()
    source_locations = [location]
    store_locations = []

    create_assets, update_assets, delete_assets = diff_locations(source_locations, store_locations)

    assert create_assets == source_locations
    assert update_assets == []
    assert delete_assets == []


def test_existing_location_no_update():
    source_location = FMILocation(building_abbr="TEST", building_number="1234")
    store_location = FMILocation(building_abbr="TEST", building_number="1234")
    source_locations = [source_location]
    store_locations = [store_location]

    create_assets, update_assets, delete_assets = diff_locations(source_locations, store_locations)

    assert not create_assets
    assert not update_assets
    assert not delete_assets


def test_existing_location_update():
    source_location = FMILocation(
        building_abbr="TEST", building_number="1234", building_name="new-test-name"
    )
    store_location = FMILocation(
        building_abbr="TEST", building_number="1234", building_name="old-test-name"
    )
    source_locations = [source_location]
    store_locations = [store_location]

    create_assets, update_assets, delete_assets = diff_locations(source_locations, store_locations)

    assert not create_assets
    assert update_assets == source_locations
    assert not delete_assets


def test_delete_location():
    """Location only in the store should be queued for deletion"""

    store_location = FMILocation(
        building_abbr="TEST", building_number="1234", building_name="old-test-name"
    )
    source_locations = []
    store_locations = [store_location]

    create_assets, update_assets, delete_assets = diff_locations(source_locations, store_locations)

    assert not create_assets
    assert not update_assets
    assert delete_assets == store_locations


def test_location_abbreviation_clash():
    """
    If a new location matches the abbreviation of a stored location, throw a warning and don't include in the creates
    """

    source_location_1 = FMILocation(building_abbr="TEST", building_number="1")
    source_location_2 = FMILocation(building_abbr="TEST", building_number="2")

    store_location = FMILocation(building_abbr="TEST", building_number="1")

    source_locations = [source_location_1, source_location_2]
    store_locations = [store_location]

    create_assets, update_assets, delete_assets = diff_locations(source_locations, store_locations)

    assert not create_assets
    assert not update_assets
    assert not delete_assets


def test_outgoing_location_abbreviation_clash():
    """
    How to handle an abbreviation clash between a new location and a stored location that no longer shows up in the source


    TODO Not sure which of the following this should actually expect:
    1. Just rely on the next scheduled run to bring over the new building?
    2. Implement the business logic here to remove the building from the comparison and datastore first?
    3. Run the whole location update function twice for each run?

    Currently implements 1
    """

    source_location_2 = FMILocation(building_abbr="TEST", building_number="2")

    store_location = FMILocation(building_abbr="TEST", building_number="1")

    source_locations = [source_location_2]
    store_locations = [store_location]

    create_assets, update_assets, delete_assets = diff_locations(source_locations, store_locations)

    assert not create_assets
    assert not update_assets
    assert delete_assets == store_locations
