from io import StringIO
from unittest.mock import MagicMock

import pytest

from sat_automations.asset_automation.models import Asset, assets_to_csv
from sat_automations.asset_automation.utils import upload_assets_to_azure

pytestmark = pytest.mark.django_db


def test_normal_asset_to_csv():
    """A valid asset should output a file object in a CSV format"""
    test_assets = [Asset(name="test-name", source_guid="test-guid")]
    test_file_object = StringIO()
    assets_to_csv(test_assets, test_file_object)
    output_string = test_file_object.read()
    assert output_string.split(",")[1] == "GUID"
    assert output_string.split("\r\n")[1].split(",")[1] == "test-guid"


def test_new_asset_from_source_has_blank_id_field():
    """A new valid asset without an assigned dynamics id should have an empty ID field"""
    test_assets = [Asset(name="test-name", source_guid="test-guid")]
    test_file_object = StringIO()
    assets_to_csv(test_assets, test_file_object)
    output_string = test_file_object.read()
    assert output_string.split(",")[0] == "ID"
    assert output_string.split("\r\n")[1].split(",")[0] == ""


def test_asset_in_dynamics_has_id():
    """An existing asset that has already been uploaded to dynamics should map dynamics_id to the first column"""
    test_assets = [Asset(name="test-name", source_guid="test-guid")]
    test_file_object = StringIO()
    assets_to_csv(test_assets, test_file_object)
    output_string = test_file_object.read()
    assert output_string.split(",")[0] == "ID"


@pytest.fixture
def azure_clients():
    container_client = MagicMock()
    blob_client = MagicMock()
    container_client.get_blob_client.return_value = blob_client
    return container_client, blob_client


def test_asset_uploaded(azure_clients):
    """The blob client's upload method should be called if function is called correctly"""
    container_client, blob_client = azure_clients

    test_assets = [Asset(name="test-name", source_guid="test-guid")]
    upload_assets_to_azure(test_assets, container_client)
    blob_client.upload_blob.assert_called()


def test_assets_chunked(azure_clients):
    """If more assets than the chunk size are uploaded, the upload method should be called twice"""
    container_client, blob_client = azure_clients

    test_assets = [
        Asset(name="test-name-1", source_guid="test-guid-1"),
        Asset(name="test-name-2", source_guid="test-guid-2"),
        Asset(name="test-name-3", source_guid="test-guid-3"),
    ]
    upload_assets_to_azure(test_assets, container_client, asset_batch_size=2)
    assert blob_client.upload_blob.call_count == 2
