import copy
import datetime
import decimal
from unittest.mock import MagicMock

import pytest

from sat_automations.asset_automation import helper_queries as hq
from sat_automations.asset_automation.models import Asset, FMILocation

pytestmark = pytest.mark.django_db

DOOR_QUERY_RETURN = (
    "test-feed",
    "test-dataSource",
    "test-serverAddr",
    "test-subsystemType",
    "test-deviceType",
    "test-GUID",
    "test-name",
    "test-type",
    "test-comments",
    "test-controllerGUID",
    "test-controllerName",
    "test-controllerType",
    "test-controllerComments",
    "test-partitionGUID",
    "test-partitionName",
    "test-partitionType",
    "test-partitionComments",
    "test-readerInGUID",
    "test-readerInName",
    "test-readerInType",
    "test-readerInComments",
    "test-readerOutGUID",
    "test-readerOutName",
    "test-readerOutType",
    "test-readerOutComments",
    "test-doorLockRelayGUID",
    "test-doorLockRelayName",
    "test-doorLockRelayType",
    "test-doorLockRelayComments",
    "test-doorSwitchGUID",
    "test-doorSwitchName",
    "test-doorSwitchType",
    "test-doorSwitchComments",
    "test-reqToExitGUID",
    "test-reqToExitName",
    "test-reqToExitType",
    "test-reqToExitComments",
)

ELEVATOR_QUERY_RETURN = (
    "test-feed",
    "test-dataSource",
    "test-serverAddr",
    "test-subsystemType",
    "test-deviceType",
    "test-GUID",
    "test-name",
    "test-type",
    "test-comments",
    "test-controllerGUID",
    "test-controllerName",
    "test-controllerType",
    "test-controllerComments",
    "test-partitionGUID",
    "test-partitionName",
    "test-partitionType",
    "test-partitionComments",
    "test-readerInGUID",
    "test-readerInName",
    "test-readerInType",
    "test-readerInComments",
)

KEYBOX_QUERY_RETURN = (
    "test-feed",
    "test-dataSource",
    "test-serverAddr",
    "test-subsystemType",
    "test-deviceType",
    "test-GUID",
    "test-name",
    "test-type",
    "test-comments",
    "test-partitionGUID",
    "test-partitionName",
    "test-partitionType",
    "test-partitionComments",
)

CAMERA_QUERY_RETURN = (
    "test-feed",
    "test-dataSource",
    "test-serverAddr",
    "test-subsystemType",
    "test-deviceType",
    "test-GUID",
    "test-name",
    "test-type",
    "test-comments",
    "test-Type",
    "test-mac",
    "test-ip",
    "test-port",
    "test-videoUnitGUID",
    "test-videoUnitName",
    "test-videoUnitManufacturer",
    "test-videoUnitModel",
    "test-firstEvent",
    "test-videoUnitFirmware",
)

SIMS_QUERY_RETURN = (
    "test-feed",
    "test-dataSource",
    "test-serverAddr",
    "test-subsystemType",
    "test-deviceType",
    "test-GUID",
    "test-name",
    "test-type",
    "test-comments",
    "test-locationType",
    "test-simsLocation",
    "test-ip",
    "test-mac",
    "test-firstEvent",
)

FMI_QUERY_RETURN = (
    "test-number",
    "test-abbreviation",
    "Weaver Lab Gasification Research Shelter",
    "3112 Faucette Dr.",
    "Raleigh",
    "NC",
    "27607",
    "092",
    "Wake",
    "USA",
    decimal.Decimal("19067"),
    1191.0,
    1206.0,
    "2013",
    "",
    "CENTRAL   ",
    "NCSU NC State University",
    "A ",
    "Active Building",
    decimal.Decimal("35.7842898000"),
    decimal.Decimal("-78.6790157630"),
    "1 ",
    "Non-Residential",
    "1 ",
    "Owned in fee simple",
    "1 ",
    "1 - Satisfactory",
    "9    ",
    "Other",
    "2 ",
    "State Appropriated",
    "",
    None,
    None,
    None,
    None,
    None,
    datetime.datetime(1900, 1, 1, 0, 0),
)

DOOR_QUERY_DESCRIPTION = (
    ("feed", str, None, 5, 5, 0, False),
    ("dataSource", str, None, 5, 5, 0, False),
    ("serverAddr", str, None, 10, 10, 0, False),
    ("subsystemType", str, None, 20, 20, 0, False),
    ("deviceType", str, None, 9, 9, 0, False),
    ("GUID", str, None, 36, 36, 0, False),
    ("name", str, None, 4000, 4000, 0, True),
    ("type", str, None, 300, 300, 0, False),
    ("comments", str, None, 4000, 4000, 0, True),
    ("controllerGUID", str, None, 36, 36, 0, True),
    ("controllerName", str, None, 4000, 4000, 0, True),
    ("controllerType", str, None, 300, 300, 0, True),
    ("controllerComments", str, None, 4000, 4000, 0, True),
    ("partitionGUID", str, None, 36, 36, 0, True),
    ("partitionName", str, None, 4000, 4000, 0, True),
    ("partitionType", str, None, 300, 300, 0, True),
    ("partitionComments", str, None, 4000, 4000, 0, True),
    ("readerInGUID", str, None, 36, 36, 0, True),
    ("readerInName", str, None, 4000, 4000, 0, True),
    ("readerInType", str, None, 300, 300, 0, True),
    ("readerInComments", str, None, 4000, 4000, 0, True),
    ("readerOutGUID", str, None, 36, 36, 0, True),
    ("readerOutName", str, None, 4000, 4000, 0, True),
    ("readerOutType", str, None, 300, 300, 0, True),
    ("readerOutComments", str, None, 4000, 4000, 0, True),
    ("doorLockRelayGUID", str, None, 36, 36, 0, True),
    ("doorLockRelayName", str, None, 4000, 4000, 0, True),
    ("doorLockRelayType", str, None, 300, 300, 0, True),
    ("doorLockRelayComments", str, None, 4000, 4000, 0, True),
    ("doorSwitchGUID", str, None, 36, 36, 0, True),
    ("doorSwitchName", str, None, 4000, 4000, 0, True),
    ("doorSwitchType", str, None, 300, 300, 0, True),
    ("doorSwitchComments", str, None, 4000, 4000, 0, True),
    ("reqToExitGUID", str, None, 36, 36, 0, True),
    ("reqToExitName", str, None, 4000, 4000, 0, True),
    ("reqToExitType", str, None, 300, 300, 0, True),
    ("reqToExitComments", str, None, 4000, 4000, 0, True),
)

ELEVATOR_QUERY_DESCRIPTION = (
    ("feed", str, None, 5, 5, 0, False),
    ("dataSource", str, None, 9, 9, 0, False),
    ("serverAddr", str, None, 10, 10, 0, False),
    ("subsystemType", str, None, 15, 15, 0, False),
    ("deviceType", str, None, 8, 8, 0, False),
    ("GUID", str, None, 36, 36, 0, False),
    ("name", str, None, 4000, 4000, 0, True),
    ("type", str, None, 300, 300, 0, False),
    ("comments", str, None, 4000, 4000, 0, True),
    ("controllerGUID", str, None, 36, 36, 0, True),
    ("controllerName", str, None, 4000, 4000, 0, True),
    ("controllerType", str, None, 300, 300, 0, True),
    ("controllerComments", str, None, 4000, 4000, 0, True),
    ("partitionGUID", str, None, 36, 36, 0, True),
    ("partitionName", str, None, 4000, 4000, 0, True),
    ("partitionType", str, None, 300, 300, 0, True),
    ("partitionComments", str, None, 4000, 4000, 0, True),
    ("readerInGUID", str, None, 36, 36, 0, True),
    ("readerInName", str, None, 4000, 4000, 0, True),
    ("readerInType", str, None, 300, 300, 0, True),
    ("readerInComments", str, None, 4000, 4000, 0, True),
)

KEYBOX_QUERY_DESCRIPTION = (
    ("feed", str, None, 5, 5, 0, False),
    ("dataSource", str, None, 12, 12, 0, False),
    ("serverAddr", str, None, 35, 35, 0, False),
    ("subsystemType", str, None, 19, 19, 0, False),
    ("deviceType", str, None, 12, 12, 0, False),
    ("GUID", str, None, 36, 36, 0, True),
    ("name", str, None, 4000, 4000, 0, True),
    ("type", str, None, 300, 300, 0, True),
    ("comments", str, None, 4000, 4000, 0, True),
    ("partitionGUID", str, None, 36, 36, 0, True),
    ("partitionName", str, None, 4000, 4000, 0, True),
    ("partitionType", str, None, 300, 300, 0, True),
    ("partitionComments", str, None, 4000, 4000, 0, True),
)

CAMERA_QUERY_DESCRIPTION = (
    ("feed", str, None, 7, 7, 0, False),
    ("dataSource", str, None, 7, 7, 0, False),
    ("serverAddr", str, None, 7, 7, 0, False),
    ("subsystemType", str, None, 22, 22, 0, False),
    ("deviceType", str, None, 10, 10, 0, False),
    ("GUID", str, None, 36, 36, 0, False),
    ("name", str, None, 100, 100, 0, False),
    ("type", int, None, 3, 3, 0, False),
    ("comments", str, None, 100, 100, 0, False),
    ("Type", int, None, 3, 3, 0, False),
    ("mac", str, None, 17, 17, 0, True),
    ("ip", str, None, 100, 100, 0, True),
    ("port", str, None, 1000, 1000, 0, True),
    ("videoUnitGUID", str, None, 36, 36, 0, True),
    ("videoUnitName", str, None, 1000, 1000, 0, True),
    ("videoUnitManufacturer", str, None, 1000, 1000, 0, True),
    ("videoUnitModel", str, None, 0, 0, 0, True),
    ("firstEvent", datetime.datetime, None, 23, 23, 3, False),
    ("videoUnitFirmware", str, None, 100, 100, 0, True),
)

SIMS_QUERY_DESCRIPTION = (
    ("feed", str, None, 4, 4, 0, False),
    ("dataSource", str, None, 8, 8, 0, False),
    ("serverAddr", str, None, 22, 22, 0, False),
    ("subsystemType", str, None, 10, 10, 0, False),
    ("deviceType", str, None, 4, 4, 0, False),
    ("GUID", str, None, 16, 16, 0, False),
    ("name", str, None, 4000, 4000, 0, True),
    ("type", str, None, 13, 13, 0, True),
    ("comments", str, None, 4000, 4000, 0, True),
    ("locationType", str, None, 13, 13, 0, True),
    ("simsLocation", str, None, 4000, 4000, 0, True),
    ("ip", str, None, 1, 1, 0, False),
    ("mac", str, None, 1, 1, 0, False),
    ("firstEvent", datetime.datetime, None, 16, 16, 0, False),
)

FMI_QUERY_DESCRIPTION = (
    ("BLDG_NUM", str, None, 10, 10, 0, True),
    ("BLDG_ABBR", str, None, 4, 4, 0, True),
    ("BLDG_NAME", str, None, 200, 200, 0, True),
    ("ADDRESS", str, None, 50, 50, 0, True),
    ("CITY", str, None, 25, 25, 0, True),
    ("STATE", str, None, 60, 60, 0, True),
    ("ZIP", str, None, 5, 5, 0, True),
    ("BLDG_COUNTY_CD", str, None, 3, 3, 0, True),
    ("BLDG_COUNTY_NAME", str, None, 50, 50, 0, True),
    ("BLDG_COUNTRY", str, None, 20, 20, 0, True),
    ("REPLACCOST", decimal.Decimal, None, 28, 28, 0, True),
    ("ASSIGNABLE", float, None, 53, 53, 0, True),
    ("GROSS", float, None, 53, 53, 0, True),
    ("CONSYEAR", str, None, 4, 4, 0, True),
    ("CAMPUS", str, None, 40, 40, 0, True),
    ("SITECODE", str, None, 10, 10, 0, True),
    ("COMPLEX_NAME", str, None, 50, 50, 0, True),
    ("STATUSCODE", str, None, 2, 2, 0, True),
    ("STATUS_DESCR", str, None, 50, 50, 0, True),
    ("LATITUDE", decimal.Decimal, None, 15, 15, 10, True),
    ("LONGITUDE", decimal.Decimal, None, 15, 15, 10, True),
    ("RESIDCODE", str, None, 2, 2, 0, True),
    ("RESID_NAME", str, None, 50, 50, 0, True),
    ("OWNERCODE", str, None, 2, 2, 0, True),
    ("OWNER_NAME", str, None, 100, 100, 0, True),
    ("CONDCODE", str, None, 2, 2, 0, True),
    ("COND_NAME", str, None, 50, 50, 0, True),
    ("USECODE", str, None, 5, 5, 0, True),
    ("USE_NAME", str, None, 50, 50, 0, True),
    ("FUNDCODE", str, None, 2, 2, 0, True),
    ("FUND_NAME", str, None, 50, 50, 0, True),
    ("HKZONE", str, None, 5, 5, 0, True),
    ("HKMANAGER", str, None, 60, 60, 0, True),
    ("HKMGREMAIL", str, None, 30, 30, 0, True),
    ("HKMGRRADIO", str, None, 30, 30, 0, True),
    ("HKMGROFFICE", str, None, 15, 15, 0, True),
    ("HKMGRCELL", str, None, 15, 15, 0, True),
    ("DATEADD", datetime.datetime, None, 23, 23, 3, True),
)


def get_mock_sql():
    mock_sql_client = MagicMock()
    mock_sql_cursor = MagicMock()
    mock_sql_client.cursor.return_value = mock_sql_cursor
    return mock_sql_client, mock_sql_cursor


@pytest.fixture
def door_mock_sql():
    mock_sql_client, mock_sql_cursor = get_mock_sql()
    mock_sql_cursor.description = DOOR_QUERY_DESCRIPTION
    mock_sql_cursor.fetchall.return_value = (DOOR_QUERY_RETURN,)
    return mock_sql_client, mock_sql_cursor


def test_valid_door_is_asset(door_mock_sql):
    mock_sql_client, mock_sql_cursor = door_mock_sql
    asset = hq.get_door_information(mock_sql_client)[0]
    assert type(asset) is Asset


def test_valid_door_has_correct_data(door_mock_sql):
    """Valid door should have it's fields mapped correctly"""
    mock_sql_client, mock_sql_cursor = door_mock_sql
    asset = hq.get_door_information(mock_sql_client)[0]
    assert asset.source_guid == "test-GUID"
    assert asset.name == "test-name"


@pytest.fixture
def elevator_mock_sql():
    mock_sql_client, mock_sql_cursor = get_mock_sql()
    mock_sql_cursor.description = ELEVATOR_QUERY_DESCRIPTION
    mock_sql_cursor.fetchall.return_value = (ELEVATOR_QUERY_RETURN,)
    return mock_sql_client, mock_sql_cursor


def test_valid_elevator_is_asset(elevator_mock_sql):
    mock_sql_client, mock_sql_cursor = elevator_mock_sql
    asset = hq.get_elevator_information(mock_sql_client)[0]
    assert type(asset) is Asset


def test_valid_elevator_has_correct_data(elevator_mock_sql):
    mock_sql_client, mock_sql_cursor = elevator_mock_sql
    asset = hq.get_elevator_information(mock_sql_client)[0]
    assert asset.source_guid == "test-GUID"
    assert asset.name == "test-name"


@pytest.fixture
def keybox_mock_sql():
    mock_sql_client, mock_sql_cursor = get_mock_sql()
    mock_sql_cursor.description = KEYBOX_QUERY_DESCRIPTION
    mock_sql_cursor.fetchall.return_value = (KEYBOX_QUERY_RETURN,)
    return mock_sql_client, mock_sql_cursor


def test_valid_keybox_is_asset(keybox_mock_sql):
    mock_sql_client, mock_sql_cursor = keybox_mock_sql
    asset = hq.get_keybox_information(mock_sql_client)[0]
    assert type(asset) is Asset


def test_valid_keybox_has_correct_data(keybox_mock_sql):
    mock_sql_client, mock_sql_cursor = keybox_mock_sql
    asset = hq.get_keybox_information(mock_sql_client)[0]
    assert asset.source_guid == "test-GUID"
    assert asset.name == "test-name"


@pytest.fixture
def camera_mock_sql():
    mock_sql_client, mock_sql_cursor = get_mock_sql()
    mock_sql_cursor.description = CAMERA_QUERY_DESCRIPTION
    mock_sql_cursor.fetchall.return_value = (CAMERA_QUERY_RETURN,)
    return mock_sql_client, mock_sql_cursor


def test_valid_camera_is_asset(camera_mock_sql):
    mock_sql_client, mock_sql_cursor = camera_mock_sql
    asset = hq.get_camera_information(mock_sql_client)[0]
    assert type(asset) is Asset


def test_valid_camera_has_correct_data(camera_mock_sql):
    mock_sql_client, mock_sql_cursor = camera_mock_sql
    asset = hq.get_camera_information(mock_sql_client)[0]
    assert asset.source_guid == "test-GUID"
    assert asset.name == "test-name"


def test_allow_through_unique_mac_cameras(camera_mock_sql):
    """If only one asset exists for a mac address, it should be allowed through for billing"""


@pytest.fixture
def camera_data():
    keys = [_[0] for _ in CAMERA_QUERY_DESCRIPTION]
    return dict(zip(keys, CAMERA_QUERY_RETURN))


def test_exclude_peep_cameras(camera_data):
    """
    Cameras on intercoms aren't billed (the intercom as a whole is), so they should be excluded
    """
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    test_name_1 = "test-camera-ENC1"
    test_name_2 = "test-camera-ENC2"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2

    camera_data_2["subsystemType"] = 8004

    camera_list = [camera_data_1, camera_data_2]

    filtered_cameras = hq.filter_peep_cameras(camera_list)
    assert test_name_1 in [camera["name"] for camera in filtered_cameras]
    assert test_name_2 not in [camera["name"] for camera in filtered_cameras]


def test_allow_through_2_macs_enc(camera_data):
    """
    If two cameras share the same mac address, but have names like `ENC`, both should be allowed through for billing

    The ENC naming scheme is used for analog cameras that share a common network box.
    These should be charged for separately, so they need to make it through to the WOS separately.
    """
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    test_name_1 = "test-camera-ENC1"
    test_name_2 = "test-camera-ENC2"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2

    camera_list = [camera_data_1, camera_data_2]
    mac_groups = {"test-mac": 2}

    filtered_cameras = hq.filter_cameras_by_mac_groups(camera_list, mac_groups)
    assert test_name_1 not in [camera["name"] for camera in filtered_cameras]
    assert test_name_2 not in [camera["name"] for camera in filtered_cameras]


def test_filter_2_macs(camera_data):
    """If two cameras share the same mac address (and don't have `ENC` in their description), only the first should make it through for billing"""
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    test_name_1 = "test-camera-360V1"
    test_name_2 = "test-camera-360V2"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2

    camera_list = [camera_data_1, camera_data_2]
    mac_groups = {"test-mac": 2}

    filtered_cameras = hq.filter_cameras_by_mac_groups(camera_list, mac_groups)
    assert test_name_1 in [camera["name"] for camera in filtered_cameras]
    assert test_name_2 not in [camera["name"] for camera in filtered_cameras]


def test_filter_3_macs(camera_data):
    """If 3 assets have a common mac address, only the first should make it through to billing

    There are types of cameras that contain multiple views.
    Genetec sees each of these views as a separate row, so they come through as unique assets.
    But we should only bill the entire unit as a single asset.
    So the code should reduce groups of mac addresses between 2 and 5 down to a single asset.
    """
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    camera_data_3 = copy.deepcopy(camera_data)
    test_name_1 = "test-camera-360V1"
    test_name_2 = "test-camera-360V2"
    test_name_3 = "test-camera-360V3"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2
    camera_data_3["name"] = test_name_3

    camera_list = [camera_data_1, camera_data_2, camera_data_3]
    mac_groups = {"test-mac": 3}

    filtered_cameras = hq.filter_cameras_by_mac_groups(camera_list, mac_groups)
    assert test_name_1 in [camera["name"] for camera in filtered_cameras]
    assert test_name_2 not in [camera["name"] for camera in filtered_cameras]
    assert test_name_3 not in [camera["name"] for camera in filtered_cameras]


def test_filter_3_macs_alt_naming(camera_data):
    """If 3 assets have a common mac address, only the first should make it through to billing

    There are types of cameras that contain multiple views.
    Genetec sees each of these views as a separate row, so they come through as unique assets.
    But we should only bill the entire unit as a single asset.
    So the code should reduce groups of mac addresses between 2 and 5 down to a single asset.
    """
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    camera_data_3 = copy.deepcopy(camera_data)
    test_name_1 = "test-camera-180VC"
    test_name_2 = "test-camera-180VL"
    test_name_3 = "test-camera-180VR"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2
    camera_data_3["name"] = test_name_3

    camera_list = [camera_data_2, camera_data_1, camera_data_3]
    mac_groups = {"test-mac": 3}

    filtered_cameras = hq.filter_cameras_by_mac_groups(camera_list, mac_groups)
    assert test_name_1 in [camera["name"] for camera in filtered_cameras]
    assert test_name_2 not in [camera["name"] for camera in filtered_cameras]
    assert test_name_3 not in [camera["name"] for camera in filtered_cameras]


def test_order_of_filtered_macs(camera_data):
    """The output based on name scheme should be consistent across different names and camera groups"""
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    camera_data_3 = copy.deepcopy(camera_data)
    test_name_1 = "test-camera-360V1"
    test_name_2 = "test-camera-360V2"
    test_name_3 = "test-camera-360V3"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2
    camera_data_3["name"] = test_name_3

    camera_list = [camera_data_3, camera_data_2, camera_data_1]
    mac_groups = {"test-mac": 3}

    filtered_cameras = hq.filter_cameras_by_mac_groups(camera_list, mac_groups)
    assert test_name_1 in [camera["name"] for camera in filtered_cameras]
    assert test_name_2 not in [camera["name"] for camera in filtered_cameras]
    assert test_name_3 not in [camera["name"] for camera in filtered_cameras]


def test_filter_5_macs(camera_data):
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    camera_data_3 = copy.deepcopy(camera_data)
    camera_data_4 = copy.deepcopy(camera_data)
    camera_data_5 = copy.deepcopy(camera_data)
    test_name_1 = "test-camera-360V1"
    test_name_2 = "test-camera-360V2"
    test_name_3 = "test-camera-360V3"
    test_name_4 = "test-camera-360V4"
    test_name_5 = "test-camera-360V5"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2
    camera_data_3["name"] = test_name_3
    camera_data_4["name"] = test_name_4
    camera_data_5["name"] = test_name_5

    camera_list = [camera_data_1, camera_data_2, camera_data_3, camera_data_4, camera_data_5]
    mac_groups = {"test-mac": 5}

    filtered_cameras = hq.filter_cameras_by_mac_groups(camera_list, mac_groups)
    assert test_name_1 in [camera["name"] for camera in filtered_cameras]
    assert test_name_2 not in [camera["name"] for camera in filtered_cameras]
    assert test_name_3 not in [camera["name"] for camera in filtered_cameras]
    assert test_name_4 not in [camera["name"] for camera in filtered_cameras]
    assert test_name_5 not in [camera["name"] for camera in filtered_cameras]


def test_ignore_6_macs(camera_data):
    """If 6 cameras share the same mac address, none should make it through for billing.

    Any groups of 6 or more assets sharing macs are only going to be legacy analog control box cameras,
    which are groups of actually separate cameras feeding into a common network box.
    These will be managed manually, so the pipeline should just ignore them,
    neither marking them for removal nor passing them through.
    """
    camera_data_1 = copy.deepcopy(camera_data)
    camera_data_2 = copy.deepcopy(camera_data)
    camera_data_3 = copy.deepcopy(camera_data)
    camera_data_4 = copy.deepcopy(camera_data)
    camera_data_5 = copy.deepcopy(camera_data)
    camera_data_6 = copy.deepcopy(camera_data)

    test_name_1 = "test-camera-ENC1"
    test_name_2 = "test-camera-ENC2"
    test_name_3 = "test-camera-ENC3"
    test_name_4 = "test-camera-ENC4"
    test_name_5 = "test-camera-ENC5"
    test_name_6 = "test-camera-ENC6"

    camera_data_1["name"] = test_name_1
    camera_data_2["name"] = test_name_2
    camera_data_3["name"] = test_name_3
    camera_data_4["name"] = test_name_4
    camera_data_5["name"] = test_name_5
    camera_data_6["name"] = test_name_6

    camera_list = [
        camera_data_1,
        camera_data_2,
        camera_data_3,
        camera_data_4,
        camera_data_5,
        camera_data_6,
    ]
    mac_groups = {"test-mac": 6}

    filtered_cameras = hq.filter_cameras_by_mac_groups(camera_list, mac_groups)
    assert filtered_cameras == []


@pytest.fixture
def sims_mock_sql():
    mock_sql_client, mock_sql_cursor = get_mock_sql()
    mock_sql_cursor.description = SIMS_QUERY_DESCRIPTION
    mock_sql_cursor.fetchall.return_value = (SIMS_QUERY_RETURN,)
    return mock_sql_client, mock_sql_cursor


def test_valid_sims_is_asset(sims_mock_sql):
    mock_sql_client, mock_sql_cursor = sims_mock_sql
    asset = hq.get_sims_information(mock_sql_client)[0]
    assert type(asset) is Asset


def test_valid_sims_has_correct_data(sims_mock_sql):
    mock_sql_client, mock_sql_cursor = sims_mock_sql
    asset = hq.get_sims_information(mock_sql_client)[0]
    assert asset.source_guid == "test-GUID"
    assert asset.name == "test-name"


@pytest.fixture
def fmi_mock_sql():
    mock_sql_client, mock_sql_cursor = get_mock_sql()
    mock_sql_cursor.description = FMI_QUERY_DESCRIPTION
    mock_sql_cursor.fetchall.return_value = (FMI_QUERY_RETURN,)
    return mock_sql_client, mock_sql_cursor


def test_valid_fmi_is_location(fmi_mock_sql):
    mock_sql_client, mock_sql_cursor = fmi_mock_sql
    location = hq.get_locations(mock_sql_client)[0]
    assert type(location) is FMILocation


def test_valid_fmi_has_correct_data(fmi_mock_sql):
    mock_sql_client, mock_sql_cursor = fmi_mock_sql
    location = hq.get_locations(mock_sql_client)[0]
    assert location.building_number == "test-number"
    assert location.building_abbr == "test-abbreviation"
