from unittest.mock import MagicMock

import pytest

from sat_automations.asset_automation.models import Asset, FMILocation
from sat_automations.asset_automation.tasks import (
    generate_local_csv,
    update_location_store,
    update_object_store,
)

pytestmark = pytest.mark.django_db


@pytest.fixture
def patch_db_setup(mocker):
    return mocker.patch("sat_automations.asset_automation.tasks.AssetAutomation")


@pytest.fixture
def patch_helper_queries(mocker):
    hq_patch = mocker.patch("sat_automations.asset_automation.tasks.helper_queries")
    hq_patch.get_door_information.return_value = [Asset(source_guid="1")]
    hq_patch.get_elevator_information.return_value = [Asset(source_guid="2")]
    hq_patch.get_keybox_information.return_value = [Asset(source_guid="3")]
    hq_patch.get_camera_information.return_value = [Asset(source_guid="4")]
    hq_patch.get_sims_information.return_value = [Asset(source_guid="5")]

    hq_patch.get_locations.return_value = [
        FMILocation(building_abbr="TEST", building_number="8675")
    ]
    return hq_patch


@pytest.fixture
def patch_csv_opener(mocker):
    csv_patch = mocker.patch("sat_automations.asset_automation.tasks.CsvOpener")
    return csv_patch


def test_generate_fresh_object_store(patch_db_setup, patch_helper_queries):
    update_object_store()
    result_assets = Asset.objects.all()
    assert 5 == len(result_assets)


def test_old_assets_removed(patch_db_setup, patch_helper_queries):
    """Generate fresh object store task should totally remove any previous assets"""
    old_asset_1 = Asset(source_guid=1)
    old_asset_1.save()

    update_object_store()
    result_assets = Asset.objects.all()
    assert 5 == len(result_assets)


def test_generate_location_store(patch_db_setup, patch_helper_queries):
    update_location_store()
    result_locations = FMILocation.objects.all()
    assert 1 == len(result_locations)


def test_old_locations_removed(patch_db_setup, patch_helper_queries):
    old_location = FMILocation(building_abbr="TEST", building_number="8675")
    old_location.save()

    update_location_store()
    result_locations = FMILocation.objects.all()
    assert 1 == len(result_locations)


def test_update_asset_store(patch_db_setup, patch_helper_queries):
    old_asset = Asset(source_guid="1", name="test-name-1")
    old_asset.save()

    patch_helper_queries.get_door_information.return_value = [
        Asset(source_guid="1", name="test-name-2")
    ]
    update_object_store()
    result_assets = Asset.objects.filter(source_guid="1").all()
    assert "test-name-2" == result_assets[0].name


def test_update_location_store(patch_db_setup, patch_helper_queries):
    old_location = FMILocation(building_abbr="TEST", building_number="8675")
    old_location.save()

    updated_building_name = "TEST2"
    patch_helper_queries.get_locations.return_value = [
        FMILocation(building_abbr=updated_building_name, building_number="8675")
    ]
    update_location_store()
    result_locations = FMILocation.objects.filter(building_number="8675").all()
    assert updated_building_name == result_locations[0].building_abbr


def test_local_csv(patch_db_setup, patch_helper_queries, patch_csv_opener):
    csv_mock = MagicMock()
    patch_csv_opener.return_value = csv_mock
    csv_enter_mock = MagicMock()
    csv_mock.__enter__.return_value = csv_enter_mock

    location = FMILocation(building_abbr="TEST-ABBR", building_number="1")
    patch_helper_queries.get_locations.return_value = [location]

    asset = Asset(source_guid="1", name="test-name-1")
    patch_helper_queries.get_door_information.return_value = [asset]

    generate_local_csv()

    # The CSV file mock should have asset and location headers, with 5 assets and 1 location
    # The mock doesn't differentiate based on being called with a different file name
    assert "ID,GUID" in csv_enter_mock.method_calls[0].args[0]
    assert "test-name-1" in csv_enter_mock.method_calls[1].args[0]
    assert "Building Number", "Building Abbreviation" in csv_enter_mock.method_calls[-2].args[0]
    assert "TEST-ABBR" in csv_enter_mock.method_calls[-1].args[0]
