from unittest.mock import MagicMock

import pytest

from sat_automations.asset_automation.utils_2n import (
    get_2n_data,
    login_to_commander_instance,
    normalize_mac_address,
)
from sat_automations.asset_automation.utils_models import (
    get_building_abbreviation_from_name,
    get_room_number_from_name,
)


@pytest.fixture
def mock_commander_session():
    return MagicMock()


@pytest.fixture
def mock_response():
    response = MagicMock()
    response.json.return_value = {"test": "test"}
    return response


def test_commander_login(mock_commander_session):
    """"""
    assert login_to_commander_instance(mock_commander_session, "test", "test")


def test_get_2n_data(mock_commander_session, mock_response):
    mock_commander_session.get.return_value = mock_response
    actual_result = get_2n_data(mock_commander_session)
    expected_result = mock_response.json.return_value
    assert expected_result == actual_result


def test_normalize_non_separated_mac_address():
    """MAC addresses with no seperator should be standardized to colon separation"""
    test_mac = "000BAA0B3CBE"
    expected_output = "00:0B:AA:0B:3C:BE"
    test_output = normalize_mac_address(test_mac)
    assert expected_output == test_output


def test_normalize_dash_mac_address():
    """Dash separated mac addresses should be standardized to colon"""
    test_mac = "00-0B-AA-0B-3C-BE"
    expected_output = "00:0B:AA:0B:3C:BE"
    test_output = normalize_mac_address(test_mac)
    assert expected_output == test_output


def test_normalize_absent_mac_address():
    """All absent macs should be empty strings"""
    test_mac = None
    expected_output = ""
    test_output = normalize_mac_address(test_mac)
    assert expected_output == test_output


def test_normalize_lower_mac_address():
    """Any lower case letters shoudld be capitalized"""
    test_mac = "00:0b:aa:0B:3C:BE"
    expected_output = "00:0B:AA:0B:3C:BE"
    test_output = normalize_mac_address(test_mac)
    assert expected_output == test_output


def test_dash_abbreviation():
    test_name = "MKA-01-E1029B-Vending-ENC12"
    expected_abbreviation = "MKA"
    actual_abbreviation = get_building_abbreviation_from_name(test_name)
    assert expected_abbreviation == actual_abbreviation


def test_space_abbreviation():
    test_name = "MCK - Cafeteria Camera"
    expected_abbreviation = "MCK"
    actual_abbreviation = get_building_abbreviation_from_name(test_name)
    assert expected_abbreviation == actual_abbreviation


def test_no_abbreviation():
    test_name = "Cafeteria Camera"
    expected_abbreviation = None
    actual_abbreviation = get_building_abbreviation_from_name(test_name)
    assert expected_abbreviation == actual_abbreviation


def test_room_number_with_floor():
    test_name = "DHL-02-2144-Data Science Academy (DSA) Corridor-DEPT"
    expected_room_number = "2144"
    actual_room_number = get_room_number_from_name(test_name)
    assert expected_room_number == actual_room_number


def test_room_number_no_floor():
    test_name = "DHL-2144-Data Science Academy (DSA) Corridor-DEPT"
    expected_room_number = "2144"
    actual_room_number = get_room_number_from_name(test_name)
    assert expected_room_number == actual_room_number


def test_alphanumeric_room_number():
    test_name = "DHL-E1002-Data Science Academy (DSA) Corridor-DEPT"
    expected_room_number = "E1002"
    actual_room_number = get_room_number_from_name(test_name)
    assert expected_room_number == actual_room_number


def test_short_alphanumeric_room_number():
    test_name = "DHL-E10-Data Science Academy (DSA) Corridor-DEPT"
    expected_room_number = "E10"
    actual_room_number = get_room_number_from_name(test_name)
    assert expected_room_number == actual_room_number


def test_short_alphanumeric_room_number_with_floor():
    test_name = "DHL-02-E10-Data Science Academy (DSA) Corridor-DEPT"
    expected_room_number = "E10"
    actual_room_number = get_room_number_from_name(test_name)
    assert expected_room_number == actual_room_number


def test_no_room_number():
    test_name = "DHL-Data Science Academy (DSA) Corridor-DEPT"
    expected_room_number = None
    actual_room_number = get_room_number_from_name(test_name)
    assert expected_room_number == actual_room_number


def test_no_room_number_with_floor():
    test_name = "DHL-02-Data Science Academy (DSA) Corridor-DEPT"
    expected_room_number = None
    actual_room_number = get_room_number_from_name(test_name)
    assert expected_room_number == actual_room_number
