import copy
import csv
import logging
import os
from io import BytesIO, StringIO
from itertools import islice

from azure.storage.blob import BlobServiceClient, ContainerClient

from .models import (
    Account,
    Asset,
    FMILocation,
    Project,
    account_to_dynamic_365,
    asset_to_dynamic_365,
    location_to_dynamic_365,
    project_to_dynamic_365,
)

LOGGER = logging.getLogger(__name__)


def get_counts_for_key(target_key, target_data):
    """Function that returns the number of times each value occurs for a given key in a dataset"""
    count_dict = dict()
    for row in target_data:
        if row[target_key] not in count_dict:
            count_dict[row[target_key]] = 1
        else:
            count_dict[row[target_key]] += 1
    return count_dict


def get_blob_service_account():
    account_url = os.environ["AZURE_STORAGE_ACCOUNT_URL"]
    shared_access_key = os.environ["AZURE_ACCOUNT_KEY"]
    return BlobServiceClient(account_url, credential=shared_access_key)


def delete_blob_files(container_client: ContainerClient, file_name_prefix: str):
    """
    Deletes all existing blob files for a specific type of object

    This approach provides better data guarantees than trying to only update files with a change,
    since assets being added out of order would bump all the assets at the end of a chunk into the next file.
    """
    old_blobs = container_client.list_blob_names(name_starts_with=file_name_prefix)
    for blob in old_blobs:
        blob_client = container_client.get_blob_client(blob=blob)
        blob_client.delete_blob()


def delete_asset_blobs(container_client: ContainerClient):
    """Deletes all existing blob files for Assets

    This approach provides better data guarantees than trying to only update files with a change,
    since assets being added out of order would bump all the assets at the end of a chunk into the next file.
    """
    old_blobs = container_client.list_blob_names(name_starts_with="asset")
    for blob in old_blobs:
        blob_client = container_client.get_blob_client(blob=blob)
        blob_client.delete_blob()


def asset_batcher(assets, batch_size):
    """Groups an iterator of assets up into chunks based on the batch_size"""
    batch_n = 0
    iterator = iter(assets)
    while batch := list(islice(iterator, batch_size)):
        batch_n += 1
        yield batch, batch_n


def upload_items_to_azure(
    items: list,
    upload_field_names: tuple,
    file_name_template: str,
    item_type_tag: str,
    container_client: ContainerClient,
    batch_size: int = 500,
) -> None:
    """Takes data in CSV ready format and uploads it to azure in batches"""
    for item_batch, item_batch_n in asset_batcher(items, batch_size):
        string_stream = StringIO()
        bytes_stream = BytesIO()

        csv_writer = csv.DictWriter(string_stream, upload_field_names)
        csv_writer.writeheader()
        csv_writer.writerows(item_batch)

        bytes_stream.write(string_stream.getvalue().encode())
        bytes_stream.seek(0)

        blob_name = file_name_template.format(item_batch_n)

        blob_client = container_client.get_blob_client(blob=blob_name)
        tags = {"adsType": item_type_tag}
        blob_client.upload_blob(bytes_stream, blob_type="BlockBlob")
        blob_client.set_blob_tags(tags)


def upload_locations_to_azure(
    locations: list[FMILocation], container_client: ContainerClient, file_name_template: str
):
    delete_asset_blobs(container_client)
    dynamic_formatted_locations = [location_to_dynamic_365(location) for location in locations]
    location_field_names = tuple(dynamic_formatted_locations[0].keys())

    upload_items_to_azure(
        dynamic_formatted_locations,
        location_field_names,
        file_name_template,
        "locations",
        container_client,
    )


def upload_accounts_to_azure(
    accounts: list[Account], container_client: ContainerClient, file_name_template: str
):
    delete_blob_files(container_client, file_name_template)
    dynamic_formatted_accounts = [account_to_dynamic_365(account) for account in accounts]
    account_field_names = tuple(dynamic_formatted_accounts[0].keys())

    upload_items_to_azure(
        dynamic_formatted_accounts,
        account_field_names,
        file_name_template,
        "accounts",
        container_client,
    )


def upload_projects_to_azure(
    projects: list[Project], container_client: ContainerClient, file_name_template: str
):
    delete_blob_files(container_client, file_name_template)
    dynamic_formatted_projects = [project_to_dynamic_365(project) for project in projects]
    project_field_names = tuple(dynamic_formatted_projects[0].keys())

    upload_items_to_azure(
        dynamic_formatted_projects,
        project_field_names,
        file_name_template,
        "projects",
        container_client,
    )


def upload_assets_to_azure(
    assets: list[Asset], container_client: ContainerClient, asset_batch_size=500
) -> None:
    """Converts list of assets to CSV files in batches and uploads them to an Azure blob container"""

    delete_asset_blobs(container_client)

    dynamic_formatted_assets = [asset_to_dynamic_365(asset) for asset in assets]
    upload_field_names = tuple(dynamic_formatted_assets[0].keys())

    for asset_batch, asset_batch_n in asset_batcher(
        dynamic_formatted_assets, batch_size=asset_batch_size
    ):
        # Need to use both types of IO stream, csv library requires StringIO,
        # and Azure library requires BytesIO
        string_stream = StringIO()
        bytes_stream = BytesIO()

        csv_writer = csv.DictWriter(string_stream, upload_field_names)
        csv_writer.writeheader()
        csv_writer.writerows(asset_batch)

        bytes_stream.write(string_stream.getvalue().encode())
        bytes_stream.seek(0)

        blob_client = container_client.get_blob_client(blob=f"assets-{asset_batch_n}.csv")
        tags = {"adsType": "asset"}
        blob_client.upload_blob(bytes_stream, blob_type="BlockBlob")
        blob_client.set_blob_tags(tags)


def diff_assets(source_assets: list[Asset], store_assets: list[Asset]):
    """Decides which assets need to be added, updated, or deleted by comparing source assets and store assets using their Source GUIDs"""

    guid_keyed_store_assets = {asset.source_guid: asset for asset in store_assets}

    source_ids = {asset.source_guid for asset in source_assets}
    store_ids = {asset.source_guid for asset in store_assets}

    new_assets = [
        source_asset for source_asset in source_assets if source_asset.source_guid not in store_ids
    ]
    raw_update_assets = [
        source_asset for source_asset in source_assets if source_asset.source_guid in store_ids
    ]
    delete_assets = [
        store_asset for store_asset in store_assets if store_asset.source_guid not in source_ids
    ]

    update_assets = []
    for asset in raw_update_assets:
        matching_store_asset = guid_keyed_store_assets[asset.source_guid]
        store_dict = matching_store_asset.__dict__
        source_dict = asset.__dict__

        store_compare_dict = {
            key: value for key, value in store_dict.items() if key not in ["_state", "id"]
        }
        source_compare_dict = {
            key: value for key, value in source_dict.items() if key not in ["_state", "id"]
        }

        if store_compare_dict != source_compare_dict:
            asset.id = matching_store_asset.id
            update_assets.append(asset)

    return new_assets, update_assets, delete_assets


def diff_locations(source_locations: list[FMILocation], store_locations: list[FMILocation]):
    """
    Get list of locations to create, update, or delete keying off both building number and building abbreviation

    Have to use both because the source FMI database uses building number as the unique identifier,
    but the actual assets are mapped to the building they're in based on the building abbreviation included in the
    asset name.

    Creates any locations that are in the source but not in the store.
    Updates any locations in the source and the store, but with value differences
    Deletes any locations in the store but not in the source
    """

    number_keyed_store_locations = {
        store_location.building_number: store_location for store_location in store_locations
    }

    {location.building_abbr for location in source_locations}
    source_numbers = {location.building_number for location in source_locations}

    store_abbreviations = {location.building_abbr for location in store_locations}
    store_numbers = {location.building_number for location in store_locations}

    new_locations = [
        source_location
        for source_location in source_locations
        if source_location.building_number not in store_numbers
        and source_location.building_abbr not in store_abbreviations
    ]

    errored_new_locations = [
        source_location
        for source_location in source_locations
        if source_location.building_number not in store_numbers
        and source_location.building_abbr in store_abbreviations
    ]
    if errored_new_locations:
        LOGGER.warning(
            f"{len(errored_new_locations)} FMI locations have been detected that have a new building number, but match an existing building's abbreviation"
        )

    raw_update_locations = [
        source_location
        for source_location in source_locations
        if source_location.building_number in store_numbers
    ]
    delete_locations = [
        store_location
        for store_location in store_locations
        if store_location.building_number not in source_numbers
    ]

    update_locations = []
    for source_location in raw_update_locations:
        matching_stored_location = number_keyed_store_locations[source_location.building_number]

        source_location_dict = copy.deepcopy(source_location.__dict__)
        matching_stored_location_dict = copy.deepcopy(matching_stored_location.__dict__)

        keys_not_to_compare = ["_state", "created", "modified", "id"]
        for location_dict in [source_location_dict, matching_stored_location_dict]:
            for key in keys_not_to_compare:
                location_dict.pop(key)

        if source_location_dict != matching_stored_location_dict:
            source_location.id = matching_stored_location.id
            update_locations.append(source_location)

    return new_locations, update_locations, delete_locations


class CsvOpener:
    def __init__(self, file_name):
        self.file_name = file_name

    def __enter__(self):
        self.file = open(self.file_name, "w")
        return self.file

    def __exit__(self, *args):
        self.file.close()
