import random
from datetime import datetime, timedelta

import pytest
from faker import Faker

from sat_automations.acs_journal.models import ACSJournal
from sat_automations.acs_journal.views import DEFAULT_QUERY_DAYS
from sat_automations.affiliate_automation.models import Cardholder, CardRequest, Sponsor
from sat_automations.disablement_automation.models import (
    DisablementTracker,
    DisablementType,
    PeopleSoftStatusType,
)
from sat_automations.disablement_automation.tests.test_disablement_tasks import (
    MATCHED_PEOPLESOFT_RECORDS,
)
from sat_automations.housing_automation.models import (
    ARLog,
    AssignRevokeTracker,
    ClearanceType,
    RoomUseCode,
    StatusType,
)
from sat_automations.housing_automation.tests.fixtures import (
    PEOPLESOFT_RESPONSE_EXAMPLES,
)
from sat_automations.utils.common.models import PeoplesoftPerson
from sat_automations.utils.common.time import SATTimeZones

Faker.seed(random.randint(0, 1000))

fake = Faker()

pytestmark = pytest.mark.django_db


SUPERVISOR = PeoplesoftPerson(**MATCHED_PEOPLESOFT_RECORDS[0].render_person())


class Response:
    status_code = 200


@pytest.fixture
def setup_rucs():
    for ruc in PEOPLESOFT_RESPONSE_EXAMPLES:
        RoomUseCode.objects.get_or_create(name=ruc.get("room_second_description", "Test"))


@pytest.fixture
def mock_sendgrid_client(mocker):
    return mocker.patch("sat_automations.disablement_automation.email.sendgrid_client")


@pytest.fixture
def patch_google_authenticate(mocker):
    def _mock_google_authenticate(app_name):
        mocker.patch(f"sat_automations.{app_name}.tasks.GoogleAuthenticate")

    return _mock_google_authenticate


@pytest.fixture
def patch_peoplesoft_proxy_provider(mocker):
    def _patch_peoplesoft_proxy_provider(app_name):
        mocker.patch(f"sat_automations.{app_name}.tasks.PeopleSoftProxyProvider")

    return _patch_peoplesoft_proxy_provider


@pytest.fixture
def patch_slack_channels(mocker):
    pass


@pytest.fixture
def patch_ccure_connection(mocker):
    def _patch_ccure_connection(app_name):
        mocker.patch(f"sat_automations.{app_name}.tasks.get_db_connection")

    return _patch_ccure_connection


@pytest.fixture
def patch_peoplesoft_provider(mocker):
    def _patch_peoplesoft_provider(app_name):
        mocker.patch(f"sat_automations.{app_name}.tasks.PeopleSoftProxyProvider")

    return _patch_peoplesoft_provider


@pytest.fixture
def random_assign_revoke_log():
    def _random_assign_revoke_log(count=1):
        logs = []
        for _ in range(count):
            logs.append(
                ARLog(
                    campus_id=fake.random_int(min=0, max=999999999),
                    ruc=RoomUseCode.objects.create(name="Test"),
                    action_date=fake.date_between(start_date="-1d", end_date="+15d"),
                    action_type=ClearanceType(random.choice(ClearanceType.choices)[0]),
                    req_status="RESV",
                )
            )
        return logs[0] if count == 1 else logs

    return _random_assign_revoke_log


@pytest.fixture
def random_assign_revoke_tracker():
    return AssignRevokeTracker(
        campus_id=fake.random_int(min=0, max=999999999),
        move_in_date=fake.date_between(start_date="-30d", end_date="-15d"),
        move_out_date=fake.date_between(start_date="-10d", end_date="-5d"),
        building_code=f"{fake.word()} {fake.word()}-{fake.random_int(min=100, max=400)}-{fake.random_letter().upper()}",
        status=StatusType(random.choice(StatusType.choices)[0]),
        action_type=ClearanceType(random.choice(ClearanceType.choices)[0]),
    )


@pytest.fixture
def fake_card_request_sponsor():
    return Sponsor(name="Sponsor123", email="sponsor123@fake.email", unity_id="sponso1")


@pytest.fixture
def fake_card_request_cardholders():
    return [
        Cardholder(
            first_name="First1",
            last_name="Last1",
            email="1@email.one",
            campus_id="001100001",
            ccure_id=5001,
            prox_id=1234,
        ),
        Cardholder(
            first_name="First2",
            last_name="Last2",
            email="2@email.two",
            campus_id="200100002",
            ccure_id=5002,
        ),
    ]


@pytest.fixture
def fake_card_request_request():
    sponsor = Sponsor.objects.create(
        name="Sponsor2", email="sponsor2@fake.email", unity_id="sponso2"
    )
    cardholders = [
        Cardholder.objects.create(
            first_name="First3",
            last_name="Last3",
            email="3@email.three",
            campus_id="001100003",
            ccure_id=5003,
            prox_id=2345,
        ),
        Cardholder.objects.create(
            first_name="First4",
            last_name="Last4",
            email="4@email.four",
            campus_id="200100004",
            ccure_id=5004,
        ),
    ]
    card_request = CardRequest.objects.create(
        date_created=datetime.now(),
        sponsor=sponsor,
        card_type="VI",
        details=None,
    )
    card_request.cardholders.set(cardholders)
    return card_request


@pytest.fixture
def random_disablement_tracker():
    purge_on = (datetime.now() + timedelta(days=90)).date()
    return DisablementTracker(
        campus_id=fake.random_int(min=0, max=999999999),
        ccure_object_id=fake.random_int(min=0, max=999999),
        ccure_guid=fake.uuid4(),
        ccure_name=fake.name(),
        ccure_personnel_id=fake.random_int(min=0, max=9999),
        ccure_last_used=fake.date_between(start_date="-30d", end_date="-15d"),
        disabled_in_ccure=fake.boolean(),
        purge_on=purge_on,
        peoplesoft_status=fake.random.choice(PeopleSoftStatusType.choices)[0],
        peoplesoft_effective_date=fake.random.choice(
            [fake.date_between(start_date="-30d", end_date="-15d"), None]
        ),
        peoplesoft_email=fake.email(),
        message=fake.text(),
    )


@pytest.fixture
def defined_disablement_tracker():
    def _defined_tracker(count=1, purge_on_days=0, email_num=0, disabled=False, student=False):
        trackers = []
        emails_sent = []
        purge_on = None
        effective_date = None
        disable_type = DisablementType.EMPLOYEE
        status = PeopleSoftStatusType.ACTIVE
        if not student:
            effective_date = fake.date_between(start_date="-30d", end_date="-15d")
        if student:
            disable_type = DisablementType.STUDENT
        if purge_on_days:
            purge_on = datetime.now() + timedelta(days=purge_on_days)
            status = PeopleSoftStatusType.INACTIVE
        for _ in range(email_num):
            emails_sent.append(datetime.now().date())
        for _ in range(count):
            trackers.append(
                DisablementTracker.objects.create(
                    campus_id=fake.random_int(min=0, max=999999999),
                    ccure_object_id=fake.random_int(min=0, max=999999),
                    ccure_guid=fake.uuid4(),
                    ccure_name=fake.name(),
                    ccure_personnel_id=fake.random_int(min=0, max=9999),
                    ccure_last_used=fake.date_between(start_date="-30d", end_date="-15d"),
                    disabled_in_ccure=disabled,
                    peoplesoft_status=status,
                    peoplesoft_email=fake.email(),
                    peoplesoft_effective_date=effective_date,
                    supervisor_id=str(fake.random_int(min=0, max=999999999)),
                    disable_type=disable_type,
                    message=fake.text(),
                    purge_on=purge_on,
                    emails_sent=emails_sent,
                )
            )
        return trackers[0] if count == 1 else trackers

    return _defined_tracker


@pytest.fixture
def employee_supervisor(defined_disablement_tracker):
    employee = defined_disablement_tracker()
    employee.disable_type = "EMP"
    employee.save()
    supervisor = SUPERVISOR
    return employee, supervisor


ACS_JOURNAL_CHOICES = {
    "message_type": [
        "ManualAction",
        "CardAdmitted",
        "KeypadCommandActivity",
        "CardRejected",
        "ObjectChangedState",
    ],
    "state_code": [
        "Admit",
        "Reject",
    ],
    "primary_object_type": [
        "SoftwareHouse.NextGen.Common.SecurityObjects.KeypadCommand",
        "SoftwareHouse.NextGen.Common.SecurityObjects.ManualAction",
        "SoftwareHouse.NextGen.Common.SecurityObjects.iStarDoor",
        "SoftwareHouse.NextGen.Common.SecurityObjects.Personnel",
        "",
    ],
    "secondary_object_type": [
        "SoftwareHouse.NextGen.Common.SecurityObjects.iStarDoor",
        "SoftwareHouse.NextGen.Common.SecurityObjects.Elevator",
        "",
    ],
}


@pytest.fixture
def random_acs_journal():
    def _random_acs_journal(
        count=1,
        start_dt=datetime.now(),
        increment_by="days",
        primary_name_prefix="CHC",
        secondary_name_prefix="CHS",
        days=DEFAULT_QUERY_DAYS,
    ):
        journals = []
        start = start_dt - timedelta(days=days)
        primary_uuids = [fake.uuid4() for _ in range(count)]
        secondary_uuids = [fake.uuid4() for _ in range(count)]
        for c in range(count):
            journals.append(
                ACSJournal.objects.create(
                    campus_id=f"{fake.random_int(min=0, max=999999999)}".zfill(9),
                    card_num=f"{fake.random_int(min=0, max=999999)}".zfill(6),
                    message_type=fake.random.choice(ACS_JOURNAL_CHOICES.get("message_type")),
                    message_utc=start.astimezone(SATTimeZones.UTC.value),
                    server_utc=start.astimezone(SATTimeZones.UTC.value),
                    state_code=fake.random.choice(ACS_JOURNAL_CHOICES.get("state_code")),
                    primary_object_name=f"{primary_name_prefix}-{fake.name()}",
                    primary_object_type=fake.random.choice(
                        ACS_JOURNAL_CHOICES.get("primary_object_type")
                    ),
                    primary_object_uuid=primary_uuids[c],
                    secondary_object_name=f"{secondary_name_prefix}-{fake.name()}",
                    secondary_object_type=fake.random.choice(
                        ACS_JOURNAL_CHOICES.get("secondary_object_type")
                    ),
                    secondary_object_uuid=secondary_uuids[c],
                )
            )
            if increment_by == "days":
                start += timedelta(days=1)
            if increment_by == "hours":
                start += timedelta(hours=1)
            if increment_by == "minutes":
                start += timedelta(minutes=1)
            if increment_by == "seconds":
                start += timedelta(seconds=1)
            if increment_by == "microseconds":
                start += timedelta(microseconds=1)
        if count == 1:
            return journals[0], primary_uuids[0], secondary_uuids[0]
        return journals, primary_uuids, secondary_uuids

    return _random_acs_journal


@pytest.fixture
def random_ccure_journal():
    def _random_ccure_journal(count=1, start_dt=datetime.now(), increment_by="days", uuids=None):
        primary_uuid = fake.uuid4()
        secondary_uuid = fake.uuid4()
        if uuids:
            primary_uuid, secondary_uuid = uuids
        journals = []
        start = start_dt - timedelta(days=60)
        for c in range(count):
            journals.append(
                (
                    f"{fake.random_int(min=0, max=999999999)}".zfill(9),
                    f"{fake.random_int(min=0, max=999999)}".zfill(6),
                    fake.random.choice(ACS_JOURNAL_CHOICES.get("message_type")),
                    start,
                    start,
                    fake.random.choice(ACS_JOURNAL_CHOICES.get("state_code")),
                    fake.name(),
                    fake.random.choice(ACS_JOURNAL_CHOICES.get("primary_object_type")),
                    primary_uuid,
                    fake.name(),
                    fake.random.choice(ACS_JOURNAL_CHOICES.get("secondary_object_type")),
                    secondary_uuid,
                )
            )

            if increment_by == "days":
                start += timedelta(days=1)
            if increment_by == "hours":
                start += timedelta(hours=1)
            if increment_by == "minutes":
                start += timedelta(minutes=1)
            if increment_by == "seconds":
                start += timedelta(seconds=1)
            if increment_by == "microseconds":
                start += timedelta(microseconds=1)
        return journals[0] if count == 1 else journals

    return _random_ccure_journal
