from datetime import datetime, timedelta

from django.contrib import (
    admin,  # noqa: F401
)

#
from sat_automations.disablement_automation.models import DisablementTracker
from sat_automations.disablement_automation.tasks import process_tracker
from sat_automations.utils.mixins.admin import ExportCsvMixin


class LastUsedFilter(admin.SimpleListFilter):
    title = "Badge Last Used"
    parameter_name = "ccure_last_used"
    now = datetime.now().date()
    sixty_days = (datetime.now() - timedelta(days=60)).date()
    hundred_and_twenty_days = (datetime.now() - timedelta(days=120)).date()
    one_year = (datetime.now() - timedelta(days=365)).date()
    three_years = (datetime.now() - timedelta(days=1095)).date()

    def lookups(self, request, model_admin):
        return (
            ("sixty_days", "within 60 Days"),
            ("hundred_twenty", "within 120 Days"),
            ("gt_year", "within 1 Year"),
            ("lt_year", "older than 1 Year"),
            ("lt_three_years", "older than 3 Years"),
            ("never", "Never Used"),
        )

    def queryset(self, request, queryset):
        if self.value() == "sixty_days":
            return queryset.filter(ccure_last_used__gte=self.sixty_days)
        if self.value() == "hundred_twenty":
            return queryset.filter(ccure_last_used__gte=self.hundred_and_twenty_days)
        if self.value() == "gt_year":
            return queryset.filter(ccure_last_used__gte=self.one_year)
        if self.value() == "lt_year":
            return queryset.filter(ccure_last_used__lte=self.one_year)
        if self.value() == "lt_three_years":
            return queryset.filter(ccure_last_used__lte=self.three_years)
        if self.value() == "never":
            return queryset.filter(ccure_last_used__isnull=True)


class DisablementTrackerAdmin(ExportCsvMixin, admin.ModelAdmin):
    list_display = [
        "ccure_name",
        "peoplesoft_email",
        "campus_id",
        "peoplesoft_effective_date",
        "ccure_last_used",
        "peoplesoft_status",
        "disable_type",
        "peoplesoft_student",
        "peoplesoft_employee",
        "ever_student",
        "purge_on",
        "disabled_in_ccure",
        "modified",
    ]
    readonly_fields = [
        # "ccure_name",
        # "campus_id",
        # "ccure_guid",
        # "ccure_object_id",
        # "peoplesoft_effective_date",
        # "ccure_last_used",
        # "peoplesoft_student",
        # "peoplesoft_employee",
        # "peoplesoft_email",
        # "peoplesoft_unity",
        # "ever_student",
        # "purge_on",
        # "ccure_personnel_id",
        # "disabled_in_ccure",
    ]
    search_fields = ["campus_id", "ccure_name", "peoplesoft_unity"]
    list_filter = [
        "peoplesoft_status",
        "purge_on",
        "peoplesoft_student",
        "peoplesoft_employee",
        "ever_student",
        "disabled_in_ccure",
        "disable_type",
        "modified",
        LastUsedFilter,
    ]

    @admin.action(description="Disable Filtered Records")
    def disable_selected_records(self, request, queryset):
        """Disable filtered records."""
        if queryset.count() == 0:
            self.message_user(
                request,
                "No records were selected. Please select at least one record to disable.",
            )
            return
        task_info = process_tracker.delay([str(x.id) for x in queryset])
        self.message_user(
            request,
            f"Disablement task with UUID ({task_info.id}) to disable {len(queryset)} records has been initiated.",
        )

    actions = [disable_selected_records] + ExportCsvMixin.actions

    class Meta:
        model = DisablementTracker

    def get_ordering(self, request):
        return ["-modified"]

    def has_module_permission(self, request, obj=None):
        if not request.user.is_staff:
            return False
        return True

    def has_add_permission(self, request, obj=None):
        if not request.user.is_superuser:
            return False
        return True

    def has_delete_permission(self, request, obj=None):
        return False


admin.site.register(DisablementTracker, DisablementTrackerAdmin)
