import urllib

from django.conf import settings
from django.utils import timezone
from sat.logs import SATLogger
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail, MailSettings, SandBoxMode

from sat_automations.disablement_automation.models import DisablementTracker, DisablementType
from sat_automations.utils.common.connection import DEFAULT_FROM_EMAIL, SENDGRID_API_KEY
from sat_automations.utils.common.models import PeoplesoftPerson
from sat_automations.utils.common.time import HUMAN_FMT

logger = SATLogger(__name__)

# These are sendgrid template IDs. These templates are designed in the sendgrid dashboard.
EMPLOYEE_TEMPLATE_ID = "d-bc17c7271603449cbe3360198a6a8903"
SUPERVISOR_TEMPLATE_ID = "d-fc229a317b534fe2bb8fe29442a87b2d"
STUDENT_TEMPLATE_ID = "d-315cdb4a70314d03aa5f261f406e9dc2"

if not SENDGRID_API_KEY:
    logger.error("SENDGRID_API_KEY is not set. Email functionality will not work.")
sendgrid_client = SendGridAPIClient(api_key=SENDGRID_API_KEY)


def send_disablement_message(
    recipient: DisablementTracker, supervisor: PeoplesoftPerson = None
) -> bool:
    success = False
    # Currently there are 2 entries in the PS_NC_EHPS_STDT_VW in PeopleSoft that have empty strings
    # in the EMAIL_ADDR field. Gotta love a permissive database.
    if recipient.peoplesoft_email and not recipient.peoplesoft_email.strip():
        return success
    if recipient.disable_type == DisablementType.EMPLOYEE:
        if send_email(_gen_message(recipient, EMPLOYEE_TEMPLATE_ID)):
            success = True
            # If the student got their email, but the supervisor has an empty email
            # we will call this a success.
            if (
                supervisor
                and supervisor.current_employee == "Y"
                and supervisor.email
                and supervisor.email.strip()
            ):
                send_email(_gen_message(recipient, SUPERVISOR_TEMPLATE_ID, supervisor=supervisor))
    else:
        """Send a student email"""
        if send_email(_gen_message(recipient, STUDENT_TEMPLATE_ID)):
            success = True
    return success


def _gen_message(
    recipient: DisablementTracker, template_id: str, supervisor: PeoplesoftPerson = None
):
    to_email = recipient.peoplesoft_email
    template_data = {
        "DISABLEMENT_DATE": (recipient.purge_on - timezone.timedelta(days=83)).strftime(HUMAN_FMT),
    }
    if supervisor:
        to_email = supervisor.email
        template_data["EMPLOYEE_NAME"] = recipient.ccure_name
    message = Mail(
        from_email=f"sat-noreply <{DEFAULT_FROM_EMAIL}>",
        to_emails=to_email,
        subject="Hopefully this won't show up",
        html_content="This also shouldn't show up",
    )
    message.template_id = template_id
    message.dynamic_template_data = template_data
    return message


def send_email(email: Mail) -> bool:
    """
    send_mail wraps Twilio SendGrid's API client, and makes a POST request to
    the api/v3/mail/send endpoint with `email`.
    Args:
        email<sendgrid.helpers.mail.Mail>: single mail object.
    """
    if settings.SENDGRID_SANDBOX:
        email.mail_settings = MailSettings().sandbox_mode = SandBoxMode(True)
    try:
        response = sendgrid_client.send(email)
        if response.status_code < 300:
            logger.debug(f"Email to {email.to} processed")
            return True
        logger.error(f"Email to {email.to} had a status code of {response.status_code}")
        return False
    except urllib.error.HTTPError as e:
        logger.error(f"Email to {email.to} failed to send: {e}")
        return False
    except KeyError as ke:
        logger.error(f"Email {email} has failed to send: {ke}")
        return False
