from datetime import datetime, timedelta

from django.core.management.base import BaseCommand
from faker import Faker

from sat_automations.disablement_automation.models import (
    DisablementTracker,
    DisablementType,
    PeopleSoftStatusType,
)

fake = Faker()

disable_within_seven = datetime.now() + timedelta(days=5)
disable_within_fourteen = datetime.now() + timedelta(days=14)


def defined_tracker(count=1, purge_on_days=0, email_num=0, disabled=False, student=False):
    trackers = []
    emails_sent = []
    purge_on = None
    effective_date = None
    disable_type = DisablementType.EMPLOYEE
    status = PeopleSoftStatusType.ACTIVE
    if not student:
        effective_date = fake.date_between(start_date="-30d", end_date="-15d")
    if student:
        disable_type = DisablementType.STUDENT
    if purge_on_days:
        purge_on = datetime.now() + timedelta(days=purge_on_days)
        status = PeopleSoftStatusType.INACTIVE
    for _ in range(email_num):
        emails_sent.append(datetime.now().date())
    for _ in range(count):
        trackers.append(
            DisablementTracker(
                campus_id=fake.random_int(min=0, max=999999999),
                ccure_object_id=fake.random_int(min=0, max=999999),
                ccure_guid=fake.uuid4(),
                ccure_name=fake.name(),
                ccure_personnel_id=fake.random_int(min=0, max=9999),
                ccure_last_used=fake.date_between(start_date="-30d", end_date="-15d"),
                disabled_in_ccure=disabled,
                peoplesoft_status=status,
                peoplesoft_email=fake.email(),
                peoplesoft_effective_date=effective_date,
                disable_type=disable_type,
                message=fake.text(),
                purge_on=purge_on,
                emails_sent=emails_sent,
            )
        )
    return trackers[0] if count == 1 else trackers


def setup_objects(objs: [DisablementTracker]):
    for obj in objs[0:10]:
        obj.peoplesoft_status = PeopleSoftStatusType.ACTIVE
        obj.save()
    for obj in objs[10:15]:
        obj.peoplesoft_status = PeopleSoftStatusType.INACTIVE
        obj.purge_on = disable_within_seven
        obj.save()
    for obj in objs[15:]:
        obj.peoplesoft_status = PeopleSoftStatusType.INACTIVE
        obj.purge_on = disable_within_fourteen
        obj.save()


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--to-email",
            help="The email address to send the test emails to.",
        )
        parser.add_argument(
            "--supervisor-id",
            help="The supervisor ID to send the test emails to. This should be a campus ID.",
        )

    def handle(self, *args, **options):
        # create tracker items
        to_email = options.get("to_email")
        supervisor_id = options.get("supervisor_id")
        if not all([to_email, supervisor_id]):
            self.stdout.write(
                self.style.WARNING("Missing required arguments. No test emails sent.")
            )
            exit(0)
        # First remove all existing trackers
        DisablementTracker.objects.all().delete()

        # Create a bunch of trackers
        NIP = defined_tracker(count=30)
        for tracker in NIP:
            tracker.peoplesoft_status = PeopleSoftStatusType.NOT_IN_PSFT
            tracker.peoplesoft_effective_date = None
            tracker.save()

        employees = defined_tracker(count=30)
        setup_objects(employees)
        students = defined_tracker(count=30, student=True)
        setup_objects(students)
        for emp in employees:
            emp.peoplesoft_email = to_email
            emp.supervisor_id = supervisor_id
            emp.save()

        for student in students:
            student.peoplesoft_email = to_email
            student.supervisor_id = supervisor_id
            student.save()

        if (
            DisablementTracker.objects.filter(
                purge_on__isnull=False,
                disabled_in_ccure=False,
                emails_sent__len__lte=2,
                peoplesoft_effective_date__isnull=False,
            ).count()
            != 20
        ):
            self.stdout.write(
                self.style.ERROR("Test data not created successfully. Please try again.")
            )
            exit(1)

        if (
            DisablementTracker.objects.filter(
                purge_on__isnull=False,
                disabled_in_ccure=False,
                emails_sent__len__lte=2,
                peoplesoft_effective_date__isnull=True,
            ).count()
            != 20
        ):
            self.stdout.write(
                self.style.ERROR("Test data not created successfully. Please try again.")
            )
            exit(1)

        self.stdout.write(self.style.SUCCESS("Test data created successfully."))
