from django.contrib.postgres.fields import ArrayField
from django.db import models  # noqa: F401
from model_utils.models import TimeStampedModel


class PeopleSoftStatusType(models.TextChoices):
    ACTIVE = "ACT", "Active"
    INACTIVE = "IAC", "Inactive"
    NOT_IN_PSFT = "NIP", "Not in PeopleSoft"


class DisablementType(models.TextChoices):
    EMPLOYEE = "EMP", "Employee"
    STUDENT = "STU", "Student"
    AFFILIATE = "AFF", "Affiliate"


class DisablementTracker(TimeStampedModel):
    """A model to track the results of the Disablement automation."""

    campus_id = models.CharField(max_length=32, unique=True)
    ccure_object_id = models.IntegerField(blank=True, null=True)
    ccure_guid = models.CharField(max_length=40, blank=True, null=True)
    ccure_name = models.CharField(max_length=255, blank=True, null=True)
    ccure_personnel_id = models.IntegerField(blank=True, null=True)
    ccure_last_used = models.DateField(blank=True, null=True)
    peoplesoft_email = models.EmailField(blank=True, null=True)
    peoplesoft_unity = models.CharField(max_length=8, blank=True, null=True)
    peoplesoft_status = models.CharField(
        choices=PeopleSoftStatusType.choices,
        default=PeopleSoftStatusType.ACTIVE,
        max_length=max(len(val) for val in PeopleSoftStatusType.values),
    )
    peoplesoft_effective_date = models.DateField(
        null=True, blank=True, verbose_name="Effective Date"
    )
    peoplesoft_student = models.BooleanField(default=False, verbose_name="Active Student")
    peoplesoft_employee = models.BooleanField(default=False, verbose_name="Active Employee")
    disable_type = models.CharField(
        choices=DisablementType.choices,
        null=True,
        blank=True,
        max_length=max(len(val) for val in DisablementType.values),
    )
    disabled_in_ccure = models.BooleanField(default=False)
    purged_from_ccure = models.BooleanField(default=False)
    ever_student = models.BooleanField(default=False)
    purge_on = models.DateField(blank=True, null=True)
    emails_sent = ArrayField(models.DateField(blank=True, null=True), default=list)
    supervisor_id = models.CharField(max_length=32, null=True, blank=True)
    message = models.TextField(blank=True, null=True)

    class Meta:
        verbose_name = "DisablementTracker"
        verbose_name_plural = "Disablement Trackers"
        indexes = [
            models.Index(
                fields=[
                    "campus_id",
                    "peoplesoft_status",
                    "ccure_name",
                    "peoplesoft_unity",
                    "peoplesoft_email",
                ]
            ),
        ]

    def __str__(self):
        return f"{self.ccure_name}"

    @classmethod
    def from_db(cls, db, field_names, values):
        """Override from_db to store original values."""
        instance = super().from_db(db, field_names, values)
        instance._loaded_values = dict(zip(field_names, values))
        return instance

    @property
    def changed(self) -> bool:
        """Return True if the model has changed since it was loaded from the database."""
        if not hasattr(self, "_loaded_values"):
            return False
        d1 = self._loaded_values
        d2 = {field.name: getattr(self, field.name) for field in self._meta.fields}
        return d1 != d2

    @property
    def is_active(self) -> bool:
        """Return True if either peoplesoft_student or peoplesoft_employee is True."""
        return any(
            [
                self.peoplesoft_student,
                self.peoplesoft_employee,
            ]
        )
