import itertools as it
import time
from dataclasses import dataclass
from datetime import date, datetime, timedelta

import requests
from acslib.base import ACSRequestException
from acslib.ccure import CcureAPI
from acslib.ccure.filters import NFUZZ, PersonnelFilter
from celery import shared_task
from django.core.paginator import Paginator
from django.utils import timezone
from pydantic import ValidationError
from sat.db import ConnectionType, get_db_connection
from sat.logs import SATLogger

from sat_automations.affiliate_automation.models import Cardholder
from sat_automations.disablement_automation.email import send_disablement_message
from sat_automations.disablement_automation.models import (
    DisablementTracker,
    DisablementType,
    PeopleSoftStatusType,
)
from sat_automations.manage_automations.models import SlackChannel
from sat_automations.utils.authenticate import GoogleAuthenticate
from sat_automations.utils.common.connection import (
    CCURE_CONNECTION,
    CLEARANCE_SERVICE_URL,
    PEOPLESOFT_PROXY_URL,
    get_acslib_connection,
)
from sat_automations.utils.common.models import PeoplesoftPerson
from sat_automations.utils.common.queries import CCURE_ALL_RECORDS_QUERY, CCURE_LAST_USED_QUERY
from sat_automations.utils.managers import BulkCreateManager

logger = SATLogger(__name__)


@dataclass
class CCureRecord:
    object_id: str
    personnel_type_id: str
    guid: str
    name: str
    campus_id: str
    is_disabled: bool
    last_used: date = None

    def __str__(self):
        return f"{self.campus_id}: {self.name}"

    def __repr__(self):
        return f"{self.campus_id}: {self.name}"


class Disablement:
    def __init__(self) -> None:
        self.authenticator = GoogleAuthenticate("disablement")
        self.authenticator.authenticate()
        if self.authenticator.response_json:
            self.access_token = self.authenticator.response_json.get("token")
        else:
            logger.error(
                f"Unable to authenticate with auth service. Error: {self.authenticator.error_message}"
            )
            raise ConnectionError("Unable to authenticate with auth service.")
        self.ccure_connection = get_db_connection(ConnectionType.SQL, conn_string=CCURE_CONNECTION)
        self.bulk_create_manager = BulkCreateManager(update_fields=["disabled_in_ccure", "message"])
        self.slack = SlackChannel.objects.filter(channel_name="disablement").first()
        self.process_failures = []
        self.current_ccure_records = {}
        self.current_peoplesoft_records = []
        self.records_for_storage = []
        self.ignore_fields = ["id", "campus_id", "created"]
        self.update_fields = [
            x.name
            for x in DisablementTracker._meta.local_fields
            if x.name not in self.ignore_fields
        ]

    @property
    def acslib_connection(self):
        return CcureAPI(get_acslib_connection())

    def get_peoplesoft_person(self, campus_id: str) -> PeoplesoftPerson | None:
        if not campus_id:
            return []
        person_search = f"{PEOPLESOFT_PROXY_URL}/people?search={campus_id}"
        response = requests.get(
            person_search,
            headers={"Authorization": f"Bearer {self.access_token}"},
            timeout=200,
        )
        if person := response.json():
            try:
                return PeoplesoftPerson(**person[0])
            except ValidationError as e:
                logger.error(f"Failed to validate person due to {e}")

    def send_recipient_emails(self, recipients: list[DisablementTracker]):
        """
        :param recipients:
        :return:
        """
        send_errors = []
        for recipient in recipients:
            supervisor = None
            supervisor_id = recipient.supervisor_id
            # a null supervisor_id is currently being set to a whitespace with a len of 1 for some reason,
            # so we need to make sure a real campus is being provided
            if supervisor_id and len(supervisor_id) > 1:
                try:
                    int(supervisor_id)  # validate supervisor_id is parsable as int
                    supervisor = self.get_peoplesoft_person(supervisor_id)
                    if supervisor:
                        logger.debug(
                            f"Sending email to: {recipient.peoplesoft_email}, and supervisor: {supervisor.email}"
                        )
                    else:
                        logger.debug(f"Sending email to: {recipient.peoplesoft_email}")
                except ValueError:
                    logger.error(f"Supervisor ID is not an integer: {supervisor_id}")
                    logger.debug(f"Sending email to: {recipient.peoplesoft_email}")
            try:
                if send_disablement_message(recipient, supervisor):  # if the email succeeds
                    recipient.emails_sent.append(timezone.now())
                    recipient.save()
                else:
                    send_errors.append(recipient.peoplesoft_email)
            except Exception:
                send_errors.append(recipient.peoplesoft_email)
        if send_errors:
            logger.error(f"Could not send emails to these recipients: {', '.join(send_errors)}")
        return {"send_errors": send_errors}

    def get_ccure_records(self):
        try:
            ccure_cursor = self.ccure_connection.cursor()
            query = CCURE_ALL_RECORDS_QUERY
            return ccure_cursor.execute(query).fetchall()
        except Exception as e:
            logger.error(f"Failed to fetch records due to {str(e)}")

    def get_last_used_date(self):
        try:
            ccure_cursor = self.ccure_connection.cursor()
            query = CCURE_LAST_USED_QUERY
            return ccure_cursor.execute(query).fetchall()
        except Exception as e:
            logger.error(f"Failed to fetch records due to {str(e)}")

    def peoplesoft_records(self, limit=0):
        url = f"{PEOPLESOFT_PROXY_URL}/people/disablement"
        if limit:
            url += f"?limit={limit}"
        response = requests.get(
            url,
            headers={
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
                "Referer": f"{__name__}",
            },
            timeout=200,
        )
        return response.json()

    def ccure_records(self) -> tuple[dict[str, CCureRecord], list[str]]:
        all_records = self.get_ccure_records()
        last_used_records = {x[1]: x[0] for x in self.get_last_used_date()}
        records = {}
        ccure_ids = []
        for record in all_records:
            obj = CCureRecord(*record)
            obj.last_used = last_used_records.get(obj.campus_id)
            records[obj.campus_id] = obj
            ccure_ids.append(obj.campus_id)
        return records, ccure_ids

    def personnel_disable_api_request(self, campus_id: str):
        response = requests.post(
            f"{CLEARANCE_SERVICE_URL}/personnel/disable?campus_id={campus_id}",
            headers={"Authorization": f"Bearer {self.access_token}"},
            timeout=2,
        )
        return response

    def personnel_exists_in_ccure(self, campus_id: str):
        personnel_records = self.acslib_connection.personnel.search(
            [campus_id], search_filter=PersonnelFilter(lookups={"Text1": NFUZZ})
        )
        return bool(personnel_records)

    def personnel_purge_api_request(self, campus_id: str):
        """Removes a person from CCURE."""
        response = requests.post(
            f"{CLEARANCE_SERVICE_URL}/personnel/remove",
            headers={"Authorization": f"Bearer {self.access_token}"},
            json={"campus_id": campus_id},
            timeout=200,
        )
        return response

    def set_person_record(
        self, record: DisablementTracker, person_record: PeoplesoftPerson
    ) -> DisablementTracker:
        """Set record details for people found in Peoplesoft"""
        record.peoplesoft_student = True if person_record.current_student == "Y" else False
        record.peoplesoft_employee = True if person_record.current_employee == "Y" else False
        if person_record.effective_date:
            record.peoplesoft_effective_date = timezone.make_aware(
                datetime.strptime(person_record.effective_date, "%Y-%m-%d %H:%M:%S")
            ).date()

        # Set email and unity
        record.peoplesoft_email = person_record.email
        record.peoplesoft_unity = person_record.unity

        # Set ever_student. Once this is set it should never be unset, since it controls when a person gets
        # disabled in CCURE after they become INACTIVE.
        if not record.ever_student and person_record.was_student == "Y":
            record.ever_student = True

        # Set disable type.
        if record.peoplesoft_effective_date:
            record.disable_type = DisablementType.EMPLOYEE
            record.supervisor_id = person_record.supervisor_id
        else:
            record.disable_type = DisablementType.STUDENT

        if record.is_active:
            record.peoplesoft_status = PeopleSoftStatusType.ACTIVE
            record.purge_on = None
            # if a person is set to active their disablement email tracker should be reset
            record.emails_sent = []
        else:
            record.peoplesoft_status = PeopleSoftStatusType.INACTIVE
            if not record.purge_on:
                # Set the purge_on date to 90 days after the record becomes inactive.
                # This is defined unless the record becomes active again.
                record.purge_on = (datetime.now() + timedelta(days=90)).date()

        cc_rec = self.current_ccure_records[record.campus_id]
        record.ccure_last_used = cc_rec.last_used
        record.disabled_in_ccure = cc_rec.is_disabled

        if record.changed:
            record.modified = timezone.now()
        return record

    @staticmethod
    def new_disablement_tracker(ccure_record: CCureRecord) -> DisablementTracker:
        record = DisablementTracker(campus_id=ccure_record.campus_id)
        record.ccure_object_id = ccure_record.object_id
        record.ccure_guid = ccure_record.guid
        record.ccure_name = ccure_record.name
        record.ccure_personnel_id = ccure_record.personnel_type_id
        record.ccure_last_used = ccure_record.last_used
        record.disabled_in_ccure = ccure_record.is_disabled
        return record

    def store_records(self, records: list[DisablementTracker], batch_size: int = 1000):
        DisablementTracker.objects.bulk_create(
            records,
            batch_size,
            update_conflicts=True,
            update_fields=self.update_fields,
            unique_fields=["campus_id"],
        )

    @staticmethod
    def grouper(inputs, n, fillvalue=None):
        iters = [iter(inputs)] * n
        return it.zip_longest(*iters, fillvalue=fillvalue)

    def store_trackers(self, records: list[DisablementTracker], batch_size: int = 1000):
        remaining_batches = len(records) // batch_size
        final_batch = len(records) % batch_size
        for batch in self.grouper(records, batch_size):
            if remaining_batches == 0:
                # The grouper will return a tuple of None values for the final batch
                # if the number of records is not evenly divisible by the batch size.
                # We need to remove those None values.
                batch = batch[:final_batch]
            self.store_records(batch, batch_size)
            remaining_batches -= 1

    def handle_in_peoplesoft(self, ids_in_peoplesoft: set[str]):
        seen_before_objs = DisablementTracker.objects.filter(campus_id__in=list(ids_in_peoplesoft))
        logger.info("Processing records that are in peoplesoft.")
        for record in seen_before_objs:
            person_record = self.set_person_record(
                record, self.current_peoplesoft_records[record.campus_id]
            )
            self.records_for_storage.append(person_record)

        seen_before_ids = [x.campus_id for x in seen_before_objs]
        not_seen_before_ids = list(ids_in_peoplesoft.difference(set(seen_before_ids)))

        for campus_id in not_seen_before_ids:
            new_record = self.new_disablement_tracker(self.current_ccure_records[campus_id])
            person_record = self.set_person_record(
                new_record, self.current_peoplesoft_records[campus_id]
            )
            self.records_for_storage.append(person_record)

    def handle_not_in_peoplesoft(self, ids_not_in_peoplesoft: set[str]):
        """
        As of this writing, records that are not in peoplesoft will be flagged as such and most
        likely never touched again.

        So all we need to do here is find the list of IDs that are not in peoplesoft and have not yet been
        added to the tracker table and add them. No need to make any changes to existing NIP records.

        :param ids_not_in_peoplesoft: campus IDs from ccure
        :return:
        """
        # affiliates are cardholders who are not in peoplesoft
        affiliates = Cardholder.objects.filter(campus_id__in=ids_not_in_peoplesoft)
        affiliate_campus_ids = {affiliate.campus_id for affiliate in affiliates}

        seen_before_objs = DisablementTracker.objects.filter(
            campus_id__in=list(ids_not_in_peoplesoft)
        )

        for obj in seen_before_objs:
            if obj.campus_id in affiliate_campus_ids:
                obj.disable_type = DisablementType.AFFILIATE

        seen_before_ids = {x.campus_id for x in seen_before_objs}
        not_seen_before_ids = (
            ids_not_in_peoplesoft - seen_before_ids
        )  # IDs in CCURE but not in the Disablement Tracker

        logger.info("Processing records that are not in peoplesoft.")
        for campus_id in not_seen_before_ids:
            new_record = self.new_disablement_tracker(self.current_ccure_records[campus_id])
            new_record.peoplesoft_status = PeopleSoftStatusType.NOT_IN_PSFT
            new_record.disable_type = (
                DisablementType.AFFILIATE if campus_id in affiliate_campus_ids else None
            )
            self.records_for_storage.append(new_record)

    def populate_tracker_records(self):
        logger.info("Started fetching all records from CCURE.")

        start = time.time()
        self.current_ccure_records, ccure_ids = self.ccure_records()
        logger.info(f"Time taken to fetch records from CCURE: {time.time() - start}")
        logger.info(f"Number of records fetched from CCURE: {len(ccure_ids)}")

        # If we see a previously purged record back in CCURE, mark it as not purged.
        previously_purged = DisablementTracker.objects.filter(
            campus_id__in=ccure_ids, purged_from_ccure=True
        )
        for record in previously_purged:
            record.purged_from_ccure = False
            record.save()

        start = time.time()
        logger.info("Started fetching all records from HR.")
        self.current_peoplesoft_records = {
            record["campus_id"]: PeoplesoftPerson(**record) for record in self.peoplesoft_records()
        }
        logger.info(f"Time taken to fetch records from HR: {time.time() - start}")
        start = time.time()
        in_peoplesoft = set(ccure_ids).intersection(set(self.current_peoplesoft_records.keys()))
        logger.info(f"Processing {len(in_peoplesoft)} records in peoplesoft.")
        self.handle_in_peoplesoft(in_peoplesoft)
        logger.info(f"Time taken to handle records in peoplesoft: {time.time() - start}")

        start = time.time()
        not_in_peoplesoft = set(ccure_ids).difference(set(self.current_peoplesoft_records.keys()))
        logger.info(f"Processing {len(not_in_peoplesoft)} records not in peoplesoft.")
        self.handle_not_in_peoplesoft(not_in_peoplesoft)
        logger.info(f"Time taken to handle records not in peoplesoft: {time.time() - start}")

        start = time.time()
        logger.info(f"Saving {len(self.records_for_storage)} records to DB.")
        self.store_trackers(self.records_for_storage)
        logger.info(f"Time taken to save records to DB: {time.time() - start}")

    def process_tracker_records(self, ids: list[str] = None):
        if not ids:
            filtered_list = DisablementTracker.objects.filter(
                peoplesoft_status=PeopleSoftStatusType.INACTIVE,
                disabled_in_ccure=False,
                purged_from_ccure=False,
                purge_on__lte=date.today() + timezone.timedelta(days=83),
            )
        else:
            filtered_list = DisablementTracker.objects.filter(id__in=ids)
        logger.info(f"Disabling {filtered_list.count()} records in CCURE.")
        failed_ids_for_slack = []
        for record in filtered_list:
            if self.personnel_exists_in_ccure(record.campus_id):
                try:
                    response = self.personnel_disable_api_request(record.campus_id)
                    if response.status_code == 200:
                        record.disabled_in_ccure = True
                    else:
                        record.message = f"Failed to disable in CCURE: {response.text}"
                        self.process_failures.append(record.campus_id)
                except ACSRequestException as e:
                    if self.personnel_exists_in_ccure(record.campus_id):
                        logger.info(f"Failed to disable in CCURE: {e.message}")
                        self.process_failures.append(record.campus_id)
                        failed_ids_for_slack.append(record.campus_id)
                    else:
                        record.purged_from_ccure = True
                        # record.save()
                        continue
            else:
                record.message = f"Record does not exist in CCURE: {record.campus_id}"
                record.purged_from_ccure = True
                # record.save()
            self.bulk_create_manager.update(record)
        self.bulk_create_manager.done()

        task_msg = f"Complete with {len(self.process_failures)} failures and  {len(filtered_list) - len(self.process_failures)} successes"
        if self.process_failures:
            if failed_ids_for_slack:
                self.slack.send_message(
                    f"Campus IDs failed to update in CCure: {failed_ids_for_slack}"
                )
            task_msg += f"Failed IDs: {self.process_failures}"
        return task_msg

    def purge_tracker_records(self):
        pass


@shared_task
def populate_tracker():
    logger.info("RUNNING: Disablement")
    disablement = Disablement()
    try:
        disablement.populate_tracker_records()
        logger.info("COMPLETE: Disablement")
    except Exception as ex:
        logger.error(f"FAILED: Disablement: {ex}")
        # disablement.slack.send_message(f"FAILED: Disablement: {ex}")
        raise ex


@shared_task
def process_tracker(ids: list[str] = None):
    logger.info("RUNNING: Process Trackers")
    disablement = Disablement()
    try:
        return disablement.process_tracker_records(ids)
    except Exception as ex:
        logger.error(f"FAILED: Process Trackers: {ex}")
        # disablement.slack.send_message(f"FAILED: Process Trackers: {ex}")
        raise ex


@shared_task
def send_emails_task():
    logger.info("RUNNING: Send Emails")
    disablement = Disablement()
    seven_days_from_now = date.today() + timezone.timedelta(days=7)

    needs_an_email = (
        DisablementTracker.objects.filter(
            purge_on__isnull=False,  # It has a purge_on date
            purge_on__gte=date.today(),  # their purge date is not in the past
            disabled_in_ccure=False,  # it is not disabled in ccure
            emails_sent__len__lt=2,  # It has not had 2 emails sent
        )
        .exclude(peoplesoft_status=PeopleSoftStatusType.NOT_IN_PSFT)
        .order_by("campus_id")
    )
    first_emails = needs_an_email.filter(
        emails_sent__len=0
    )  # All records that have not had an email sent
    second_emails = needs_an_email.filter(emails_sent__len=1, purge_on__lte=seven_days_from_now)
    emails_to_send = first_emails | second_emails
    logger.info(f"Emails to send: {emails_to_send.count()}")
    email_pages = Paginator(emails_to_send, 100)
    send_email_failures = []
    for page in email_pages:
        try:
            logger.info(f"Page: {page.number} of {email_pages.num_pages} pages.")
            response = disablement.send_recipient_emails(page.object_list)
            if send_errors := response.get("send_errors"):
                send_email_failures.extend(send_errors)
        except Exception as e:
            send_email_failures.extend(page.object_list)
            logger.error(f"Could not send emails for {page.object_list}: {e}")
    logger.info("COMPLETE: Send Emails")
    return {"send_email_failures": len(send_email_failures)}


@shared_task
def purge_task():
    logger.info("RUNNING: Purge disabled users")
    disablement = Disablement()

    # Get all trackers where the purge_on date has passed.
    needs_purge = DisablementTracker.objects.filter(
        purge_on__lte=date.today(),  # It has a purge_on date
        emails_sent__len=2,  # It has two emails sent
        purged_from_ccure=False,  # It has not already been purged
    )
    try:
        for user in needs_purge:
            logger.info(f"Purging person {user.campus_id}")
            # Remove this person from CCURE.
            response = disablement.personnel_purge_api_request(user.campus_id)
            if response.status_code != 200:
                logger.error(f"Unable to purge person {user.campus_id} from CCURE.")
            else:
                user.purged_from_ccure = True
                user.save()

    except Exception as ex:
        logger.error(f"FAILED: Purge: {ex}")
        raise ex
