import copy
import random
from dataclasses import dataclass
from datetime import datetime, timedelta

import pytest
import time_machine
from faker import Faker

from sat_automations.affiliate_automation.models import Cardholder
from sat_automations.disablement_automation.config import STUDENT_DISABLEMENT_DAYS
from sat_automations.disablement_automation.models import (
    DisablementTracker,
    DisablementType,
    PeopleSoftStatusType,
)
from sat_automations.disablement_automation.tasks import Disablement, send_emails_task

Faker.seed(random.randint(0, 1000))

fake = Faker()

pytestmark = pytest.mark.django_db


@dataclass
class Person:
    campus_id: str
    first_name: str
    middle_name: str
    last_name: str
    classification: str
    unity: str
    email: str = ""
    # Employees
    empl_status: str = ""
    department: str = ""
    ouc: str = ""
    effective_date: str = ""
    supervisor_id: str = ""
    # Students
    current_student: str = ""
    academic_level: str = ""
    was_student: str = ""

    def __eq__(self, other):
        return self.campus_id == other.campus_id

    def render_person(self):
        return {
            "campus_id": self.campus_id,
            "first_name": self.first_name,
            "middle_name": self.middle_name,
            "last_name": self.last_name,
            "classification": self.classification,
            "effective_date": f"{self.effective_date}",
            "academic_level": self.academic_level,
            "current_student": "Y" if self.current_student else "N",
            "current_employee": "Y" if self.empl_status == "A" else "N",
            "was_student": self.was_student,
            "email": self.email,
            "unity": self.unity,
            "department": self.department,
            "ouc": self.ouc,
            "supervisor_id": self.supervisor_id,
            "reports_to": " ",
        }


PERSONNEL_TYPE_ID = [
    1,
    2,
    3,
    5000,
    5001,
    5003,
    5005,
    5006,
    5007,
    5008,
    5010,
    5011,
    5012,
    5013,
    5014,
    5015,
    5016,
    5017,
    5018,
    5019,
    5020,
    5021,
    5022,
    5023,
]


def build_ccure_records(count=4):
    return [
        (
            str(fake.random_int(min=0, max=99999)),  # object_id
            str(random.choice(PERSONNEL_TYPE_ID)),  # personnel_type_id
            fake.uuid4(),  # guid
            f"{fake.last_name()}, {fake.first_name()}",  # name
            str(fake.random_int(min=0, max=999999999)),  # campus_id
            False,  # disabled
        )
        for _ in range(count)
    ]


FIXED_CCURE_RECORDS = build_ccure_records()


def build_peoplesoft_records(ccure_records: list[tuple] = ()) -> list[Person]:
    """
    By default, this function will return a list of 20 Person objects with
    campus ids that match the CCURE records.
    1. Half of the records will have an effective date within the last 30 days.
    2. The other half will have an effective date older than 60 days.
    3. Half will be current students.
    4. Half will be current employees.
    :param campus_ids:
    :param count:
    :return:
    """
    if not ccure_records:
        ccure_records = [ccure_record for ccure_record in FIXED_CCURE_RECORDS]
    persons = []
    for i, id in enumerate(ccure_records):
        ef_date = ""
        if i % 2 == 0:
            ef_date = datetime.now().isoformat(sep=" ", timespec="seconds")
            student = "Y"
            employee = "A"
            was_student = "Y"
        else:
            student = ""
            employee = "T"
            was_student = ""
        person = Person(
            campus_id=id[4],
            first_name=id[3].split(", ")[1],
            middle_name=fake.random_letter().upper(),
            last_name=id[3].split(", ")[0],
            classification=random.choice(("Faculty", "Staff")),
            email=fake.email(),
            unity=fake.text(max_nb_chars=8, ext_word_list=None),
            current_student=student,
            was_student=was_student,
            empl_status=employee,
            effective_date=ef_date,
            supervisor_id="200568478",
            department=fake.word(),
            ouc="11111",
        )
        persons.append(person)
    return persons


MATCHED_PEOPLESOFT_RECORDS = build_peoplesoft_records()


@pytest.fixture
def patch_ccure_records(mocker):
    mocker.patch("sat_automations.disablement_automation.tasks.Disablement.get_ccure_records")


@pytest.fixture
def patch_peoplesoft_records(mocker):
    mocker.patch("sat_automations.disablement_automation.tasks.Disablement.peoplesoft_records")


@pytest.fixture
def patch_get_person(mocker):
    mocker.patch("sat_automations.disablement_automation.tasks.Disablement.get_peoplesoft_person")


@pytest.fixture
def patch_send_emails(mocker):
    return mocker.patch("sat_automations.disablement_automation.tasks.send_disablement_message")


@pytest.fixture
def standard_patch(
    patch_ccure_records, patch_google_authenticate, patch_ccure_connection, patch_peoplesoft_records
):
    patch_google_authenticate("disablement_automation")
    patch_ccure_connection("disablement_automation")


@pytest.fixture
def email_patch(standard_patch, patch_get_person, patch_send_emails):
    return patch_send_emails


def test_store_ccure_records(standard_patch, patch_get_person, monkeypatch):
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    ccure = FIXED_CCURE_RECORDS
    peoplesoft = [x.render_person() for x in MATCHED_PEOPLESOFT_RECORDS]
    dis.get_ccure_records.return_value = ccure
    dis.peoplesoft_records.return_value = peoplesoft
    dis.get_peoplesoft_person.return_value = []
    dis.populate_tracker_records()
    assert DisablementTracker.objects.all().count() == 4
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.ACTIVE).count()
        == 2
    )
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.INACTIVE).count()
        == 2
    )
    assert (
        DisablementTracker.objects.filter(
            peoplesoft_status=PeopleSoftStatusType.NOT_IN_PSFT
        ).count()
        == 0
    )


def test_no_ccure_records(monkeypatch, standard_patch, caplog):
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    peoplesoft = [x.render_person() for x in MATCHED_PEOPLESOFT_RECORDS]
    dis.get_ccure_records.return_value = []
    dis.peoplesoft_records.return_value = peoplesoft
    dis.populate_tracker_records()
    assert DisablementTracker.objects.all().count() == 0
    assert "Number of records fetched from CCURE: 0" in caplog.text


def test_ccure_not_in_peoplesoft(monkeypatch, standard_patch, patch_get_person):
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    ccure = FIXED_CCURE_RECORDS
    cardholders = [
        Cardholder(
            first_name=cardholder[3].split(", ")[0],
            last_name=cardholder[3].split(", ")[1],
            email="fakeemail@email.email",
            campus_id=cardholder[4],
        )
        for cardholder in ccure
    ]
    Cardholder.objects.bulk_create(cardholders)
    # Remove the first record from the peoplesoft records
    peoplesoft = [x.render_person() for x in MATCHED_PEOPLESOFT_RECORDS[1:4]]
    dis.get_ccure_records.return_value = ccure
    dis.peoplesoft_records.return_value = peoplesoft
    dis.get_peoplesoft_person.return_value = []
    dis.populate_tracker_records()
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.ACTIVE).count()
        == 1
    )
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.INACTIVE).count()
        == 2
    )
    assert (
        DisablementTracker.objects.filter(
            peoplesoft_status=PeopleSoftStatusType.NOT_IN_PSFT
        ).count()
        == 1
    )
    assert DisablementTracker.objects.filter(disable_type=DisablementType.AFFILIATE).count() == 1


def test_employee_inactive(monkeypatch, standard_patch, patch_get_person):
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    ccure = FIXED_CCURE_RECORDS.copy()
    # Add a record with an effective date older than 60 days
    ccure_record = build_ccure_records(1)
    ccure.extend(ccure_record)
    psft_records = build_peoplesoft_records([x for x in ccure])
    psft_records[4].current_student = ""
    psft_records[4].empl_status = "T"
    peoplesoft = [x.render_person() for x in psft_records]
    dis.get_ccure_records.return_value = ccure
    dis.peoplesoft_records.return_value = peoplesoft
    dis.get_peoplesoft_person.return_value = []
    dis.populate_tracker_records()
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.ACTIVE).count()
        == 2
    )
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.INACTIVE).count()
        == 3
    )
    assert (
        DisablementTracker.objects.filter(
            peoplesoft_status=PeopleSoftStatusType.NOT_IN_PSFT
        ).count()
        == 0
    )


def test_inactive_to_active_removes_purge_on(monkeypatch, standard_patch, patch_get_person):
    """
    A record that was previously inactive in peoplesoft is now active. The purge_on date should be removed.
    :param standard_patch:
    :return:
    """

    test_store_ccure_records(standard_patch, patch_get_person, monkeypatch)
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    # Change the status of one of the records in peoplesoft
    record_list = copy.deepcopy(MATCHED_PEOPLESOFT_RECORDS)
    former_inactive = [x for x in record_list if x.empl_status == "T" and x.current_student == ""]
    former_inactive[0].empl_status = "A"
    former_inactive[0].current_student = ""
    # Assert that they currently have a purge_on date
    assert (
        DisablementTracker.objects.get(campus_id=former_inactive[0].campus_id).purge_on is not None
    )

    peoplesoft = [x.render_person() for x in former_inactive]
    dis.get_ccure_records.return_value = FIXED_CCURE_RECORDS
    dis.peoplesoft_records.return_value = peoplesoft
    dis.populate_tracker_records()
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.ACTIVE).count()
        == 3
    )
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.INACTIVE).count()
        == 1
    )
    assert (
        DisablementTracker.objects.filter(
            peoplesoft_status=PeopleSoftStatusType.NOT_IN_PSFT
        ).count()
        == 0
    )
    assert DisablementTracker.objects.get(campus_id=former_inactive[0].campus_id).purge_on is None


def test_active_to_inactive_sets_purge_on(monkeypatch, standard_patch, patch_get_person):
    """
    A record that was previously active in peoplesoft is now inactive. The purge_on date should be set.
    :param standard_patch:
    :return:
    """

    test_store_ccure_records(standard_patch, patch_get_person, monkeypatch)
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    # Change the status of one of the records in peoplesoft
    record_list = copy.deepcopy(MATCHED_PEOPLESOFT_RECORDS)
    former_active = [x for x in record_list if x.empl_status == "A" or x.current_student == "Y"]
    former_active[0].empl_status = "T"
    former_active[0].current_student = ""
    # Assert that they currently do not have a purge_on date
    assert DisablementTracker.objects.get(campus_id=former_active[0].campus_id).purge_on is None

    peoplesoft = [x.render_person() for x in former_active]
    dis.get_ccure_records.return_value = FIXED_CCURE_RECORDS
    dis.peoplesoft_records.return_value = peoplesoft
    dis.populate_tracker_records()
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.ACTIVE).count()
        == 1
    )
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.INACTIVE).count()
        == 3
    )
    assert (
        DisablementTracker.objects.filter(
            peoplesoft_status=PeopleSoftStatusType.NOT_IN_PSFT
        ).count()
        == 0
    )
    assert DisablementTracker.objects.get(campus_id=former_active[0].campus_id).purge_on is not None


def test_update_works_on_changes_from_peoplesoft(monkeypatch, standard_patch, patch_get_person):
    """
    A formerly active record in CCURE is now inactive in peoplesoft. The status should change to inactive.
    :param standard_patch:
    :return:
    """
    test_store_ccure_records(standard_patch, patch_get_person, monkeypatch)
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    # Change the status of one of the records in peoplesoft
    former_active = copy.deepcopy(MATCHED_PEOPLESOFT_RECORDS)
    former_active[0].empl_status = "T"
    former_active[0].current_student = ""
    former_active[0].effective_date = (datetime.now() - timedelta(days=65)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    peoplesoft = [x.render_person() for x in former_active]
    dis.get_ccure_records.return_value = FIXED_CCURE_RECORDS
    dis.peoplesoft_records.return_value = peoplesoft
    dis.populate_tracker_records()
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.ACTIVE).count()
        == 1
    )
    assert (
        DisablementTracker.objects.filter(peoplesoft_status=PeopleSoftStatusType.INACTIVE).count()
        == 3
    )
    assert (
        DisablementTracker.objects.filter(
            peoplesoft_status=PeopleSoftStatusType.NOT_IN_PSFT
        ).count()
        == 0
    )


def test_update_changes_the_modified_datetime(monkeypatch, standard_patch, patch_get_person):
    """
    Any change to the DisablementTracker model should change the modified datetime.
    :param standard_patch:
    :return:
    """
    test_store_ccure_records(standard_patch, patch_get_person, monkeypatch)
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    # Change the status of one of the records in peoplesoft
    former_active = copy.deepcopy(MATCHED_PEOPLESOFT_RECORDS)
    former_active[0].empl_status = "T"
    former_active[0].current_student = ""
    former_active[0].effective_date = (datetime.now() - timedelta(days=65)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    before_change = DisablementTracker.objects.get(campus_id=former_active[0].campus_id)
    peoplesoft = [x.render_person() for x in former_active]
    dis.get_ccure_records.return_value = FIXED_CCURE_RECORDS
    dis.peoplesoft_records.return_value = peoplesoft
    dis.populate_tracker_records()
    after_change = DisablementTracker.objects.get(campus_id=former_active[0].campus_id)
    assert before_change.modified != after_change.modified


def test_process_tracker_success(monkeypatch, mocker, standard_patch, patch_get_person):
    mocker.patch(
        "sat_automations.disablement_automation.tasks.Disablement.personnel_disable_api_request"
    )
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    monkeypatch.setattr(Disablement, "personnel_exists_in_ccure", lambda *_, **__: False)
    test_store_ccure_records(standard_patch, patch_get_person, monkeypatch)
    dis = Disablement()
    dis.personnel_disable_api_request.return_value = mocker.Mock(status_code=200)
    traveller = time_machine.travel(datetime.now() + timedelta(days=STUDENT_DISABLEMENT_DAYS + 1))
    traveller.start()
    dis.process_tracker_records()
    dis.bulk_create_manager.done()
    assert DisablementTracker.objects.filter(disabled_in_ccure=True).count() == 0
    traveller.stop()


def test_process_tracker_failure(monkeypatch, mocker, standard_patch, patch_get_person):
    mocker.patch(
        "sat_automations.disablement_automation.tasks.Disablement.personnel_disable_api_request"
    )
    monkeypatch.setattr(
        Disablement,
        "personnel_exists_in_ccure",
        lambda *_, **__: False,
    )
    test_store_ccure_records(standard_patch, patch_get_person, monkeypatch)
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    dis.slack = mocker.Mock()
    dis.personnel_disable_api_request.return_value = mocker.Mock(status_code=404, text="Not Found")
    traveller = time_machine.travel(datetime.now() + timedelta(days=STUDENT_DISABLEMENT_DAYS + 1))
    traveller.start()
    dis.process_tracker_records()
    dis.bulk_create_manager.done()
    assert DisablementTracker.objects.filter(disabled_in_ccure=True).count() == 0
    for tracker in DisablementTracker.objects.filter(
        peoplesoft_status=PeopleSoftStatusType.INACTIVE, disabled_in_ccure=False
    ):
        assert "Record does not exist in CCURE" in tracker.message
    traveller.stop()


def test_send_emails_default(caplog, defined_disablement_tracker, email_patch):
    defined_disablement_tracker()
    send_emails_task()
    assert "Emails to send: 0" in caplog.messages
    assert "COMPLETE: Send Emails" in caplog.messages


def test_send_emails_employee(caplog, defined_disablement_tracker, email_patch):
    assert DisablementTracker.objects.all().count() == 0
    # This should result in an email being sent to a single employee.
    tracker = defined_disablement_tracker(count=1, purge_on_days=30, email_num=0)
    tracker.disable_type = DisablementType.EMPLOYEE
    tracker.save()
    email_patch.return_value = True
    send_emails_task()
    assert "Emails to send: 1" in caplog.messages
    assert "COMPLETE: Send Emails" in caplog.messages
    tracker.refresh_from_db()
    assert len(tracker.emails_sent) == 1


def test_send_emails_student(caplog, defined_disablement_tracker, email_patch):
    assert DisablementTracker.objects.all().count() == 0
    # This should result in an email being sent to a single student.
    tracker = defined_disablement_tracker(count=1, purge_on_days=30, email_num=0, student=True)
    email_patch.return_value = True
    send_emails_task()
    assert "Emails to send: 1" in caplog.messages
    assert "COMPLETE: Send Emails" in caplog.messages
    tracker.refresh_from_db()
    assert len(tracker.emails_sent) == 1


def test_send_emails_second_email_within_window(caplog, defined_disablement_tracker, email_patch):
    assert DisablementTracker.objects.all().count() == 0
    tracker = defined_disablement_tracker(count=1, purge_on_days=30)
    email_patch.return_value = True
    send_emails_task()
    tracker.refresh_from_db()
    assert len(tracker.emails_sent) == 1
    traveller = time_machine.travel(tracker.purge_on - timedelta(days=6))
    traveller.start()
    send_emails_task()
    tracker.refresh_from_db()
    assert len(tracker.emails_sent) == 2
    traveller.stop()


def test_send_emails_second_email_outside_window(caplog, defined_disablement_tracker, email_patch):
    assert DisablementTracker.objects.all().count() == 0
    tracker = defined_disablement_tracker(count=1, purge_on_days=30)
    email_patch.return_value = True
    send_emails_task()
    tracker.refresh_from_db()
    assert len(tracker.emails_sent) == 1
    traveller = time_machine.travel(tracker.purge_on - timedelta(days=10))
    traveller.start()
    send_emails_task()
    tracker.refresh_from_db()
    assert len(tracker.emails_sent) == 1
    traveller.stop()


def test_non_int_supervisor_id(caplog, defined_disablement_tracker, email_patch):
    assert DisablementTracker.objects.all().count() == 0
    tracker = defined_disablement_tracker(count=1, purge_on_days=30)
    tracker.supervisor_id = "not an int"
    tracker.save()
    email_patch.return_value = True
    send_emails_task()
    assert "Supervisor ID is not an integer: not an int" in caplog.messages
    assert "COMPLETE: Send Emails" in caplog.messages


def test_pagination(caplog, defined_disablement_tracker, email_patch):
    assert DisablementTracker.objects.all().count() == 0
    defined_disablement_tracker(count=101, purge_on_days=30)
    email_patch.return_value = True
    send_emails_task()
    assert "Emails to send: 101" in caplog.messages
    assert "Page: 1 of 2 pages." in caplog.messages
    assert "Page: 2 of 2 pages." in caplog.messages
    assert "COMPLETE: Send Emails" in caplog.messages


def test_emails_sent_array(monkeypatch, standard_patch, patch_get_person, email_patch):
    email_patch.return_value = True
    monkeypatch.setattr(Disablement, "acslib_connection", lambda *_, **__: None)
    dis = Disablement()
    # Change the status of one of the records in peoplesoft
    record_list = copy.deepcopy(MATCHED_PEOPLESOFT_RECORDS)
    former_inactive = [x for x in record_list if x.empl_status == "T" and x.current_student == ""]
    former_inactive[0].empl_status = "A"
    former_inactive[0].current_student = ""
    peoplesoft = [x.render_person() for x in former_inactive]
    dis.get_ccure_records.return_value = FIXED_CCURE_RECORDS
    dis.peoplesoft_records.return_value = peoplesoft
    dis.get_peoplesoft_person.return_value = []

    # when first populated every record in the tracker should have no emails sent
    dis.populate_tracker_records()
    assert DisablementTracker.objects.filter(emails_sent=[]).count() == 4
    send_emails_task()
    assert DisablementTracker.objects.filter(emails_sent=[]).count() == 3
    # the record that was switched to active should reset its' emails_sent back to empty
    dis.populate_tracker_records()
    assert DisablementTracker.objects.filter(emails_sent=[]).count() == 4
