from django.contrib import admin  # noqa: F401
from django.db.models import Count
from django.db.models.functions import Substr
from django.utils.html import format_html_join

from sat_automations.housing_automation.models import (
    ARLog,
    Clearance,
    RoomUseCode,
    UnmanagedBuilding,
)
from sat_automations.utils.mixins.admin import ExportCsvMixin

# Custom ListFilters


class ClearanceNamePrefixFilter(admin.SimpleListFilter):
    title = "Prefix"
    parameter_name = "Name"

    def lookups(self, request, model_admin):
        qs = (
            Clearance.objects.annotate(prefix=Substr("name", 1, 3))
            .values("prefix")
            .annotate(pre_count=Count("id"))
            .order_by("-pre_count")
        )
        res = [(k["prefix"], k["prefix"]) for k in qs]
        return res

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(name__startswith=self.value())
        return queryset


class NoClearancesFilter(admin.SimpleListFilter):
    title = "No Clearances"
    parameter_name = "name"

    def lookups(self, request, model_admin):
        return (("NC", "No Clearances Assigned"),)

    def queryset(self, request, queryset):
        unmanaged = list(UnmanagedBuilding.objects.all().values_list("prefix", flat=True))
        hide = []
        for prefix in unmanaged:
            hide.extend(
                list(
                    RoomUseCode.objects.filter(name__startswith=prefix).values_list(
                        "name", flat=True
                    )
                )
            )
        match (self.value()):
            case ("NC"):
                qs = queryset.filter(clearances__name="NC").exclude(name__in=hide)
                qs1 = RoomUseCode.objects.filter(clearances__isnull=True).exclude(name__in=hide)
                return qs | qs1


class MoreThanTwoEntriesFilter(admin.SimpleListFilter):
    title = "More Than Two Actions"
    parameter_name = "two_actions"

    def lookups(self, request, model_admin):
        return [(2, "> 2 actions"), (3, "> 3 actions")]

    def queryset(self, request, queryset):
        if self.value():
            return (
                ARLog.objects.all()
                .annotate(Count("campus_id"))
                .order_by()
                .filter(campus_id__count__gt=self.value())
            )


class ARLogAdmin(ExportCsvMixin, admin.ModelAdmin):
    list_display = [
        "campus_id",
        "get_ruc",
        "action_date",
        "action_type",
        "req_status",
        "action_success",
        "action_message",
        "peoplesoft_action_dttm",
        "created",
        "modified",
    ]
    ordering = ["-action_date"]

    search_fields = [
        "campus_id",
        "ruc__name",
    ]

    list_filter = ["action_type", "action_success", "action_message", MoreThanTwoEntriesFilter]

    actions = ExportCsvMixin.actions

    @admin.display(description="Room Use Code")
    def get_ruc(self, obj):
        return obj.ruc.name

    def has_module_permission(self, request, obj=None):
        if not request.user.is_staff:
            return False
        return True

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class UnmanagedBuildingAdmin(admin.ModelAdmin):
    list_display = ["prefix"]


class RoomUseCodeAdmin(ExportCsvMixin, admin.ModelAdmin):
    list_display = ["name", "assigned_clearances", "created", "modified"]
    list_filter = [NoClearancesFilter, "created", "modified"]
    filter_horizontal = ["clearances"]

    ordering = ["name"]
    search_fields = ["name"]
    actions = ExportCsvMixin.actions

    class Meta:
        model = RoomUseCode

    def assigned_clearances(self, obj):
        clearances = (c.name for c in obj.clearances.all())
        if len(obj.clearances.all()) > 1:
            return format_html_join("\n", "{}<br/>", ((c,) for c in clearances))
        if clearance := obj.clearances.first():
            return clearance.name


class ClearanceAdmin(ExportCsvMixin, admin.ModelAdmin):
    list_display = ["name", "created", "modified"]
    list_filter = [ClearanceNamePrefixFilter]
    actions = ExportCsvMixin.actions

    ordering = ["name"]
    search_fields = ["name"]

    class Meta:
        model = Clearance


# Register your models here.
admin.site.register(Clearance, ClearanceAdmin)
admin.site.register(RoomUseCode, RoomUseCodeAdmin)
admin.site.register(ARLog, ARLogAdmin)
admin.site.register(UnmanagedBuilding, UnmanagedBuildingAdmin)
