from typing import Optional

from dateutil.parser import parse
from django.db import models  # noqa: F401
from model_utils.models import TimeStampedModel
from pydantic import BaseModel

CCURE_COLUMN_HEADERS = [
    "Text1",  # campus_id
    "Text12",  # cidc
    "FirstName",
    "MiddleName",
    "LastName",
    "Text16",
    "Text14",
]


class ClearanceType(models.TextChoices):
    ASSIGN = "assign", "Assign"
    REVOKE = "revoke", "Revoke"


class StatusType(models.TextChoices):
    PASS = "pass", "Pass"
    FAIL = "fail", "Fail"
    MODIFIED = "modified", "Modified"


class LogMessageType(models.TextChoices):
    BUILDING_CODE_NOT_FOUND = "BC_NF", "Building code not found"
    BUILDING_CODE_NO_CLEARANCE = "BC_NC", "Building code has no clearance"
    ASSIGNMENT_FAILED = "FC", "Failed Clearance Assignment Operation"
    ASSIGNMENT_SUCCESS = "SC", "Successful Clearance Assignment Operation"


# Create your models here.


class CCurePersonnel(BaseModel):
    # the only field acslib requires is last_name
    campus_id: Optional[str] = None
    cidc: Optional[str] = None
    first_name: Optional[str] = None
    middle_name: Optional[str] = None
    last_name: str
    classification: Optional[str] = None
    email: Optional[str] = None


# Non-Django classes
class PeopleSoftProxyBase:
    def __init__(self, access_token=None) -> None:
        self.base_url = ""
        self.access_token = access_token

    def get_person(self, campus_id):
        pass

    def get_housing(self, days):
        pass


class AssignRevokeLog(TimeStampedModel):
    """A model to track the results of the HousingAssignRevoke automation."""

    campus_id = models.CharField(max_length=32)
    building_code = models.CharField(max_length=64)
    clearance_type = models.CharField(
        choices=ClearanceType.choices, max_length=max(len(x) for x in ClearanceType.values)
    )
    status = models.CharField(
        choices=StatusType.choices, max_length=max(len(x) for x in StatusType.values)
    )
    message = models.CharField(
        choices=LogMessageType.choices, max_length=max(len(x) for x in LogMessageType.values)
    )
    extra_info = models.TextField(blank=True, null=True)

    class Meta:
        verbose_name = "Assign/Revoke Log"
        verbose_name_plural = "Assign/Revoke Logs"
        indexes = [
            models.Index(fields=["campus_id", "building_code"]),
        ]

    def __str__(self):
        return f"{self.campus_id}: {self.building_code} - {self.clearance_type} - {self.status}"


class AssignRevokeTracker(TimeStampedModel):
    campus_id = models.CharField(max_length=32)
    move_in_date = models.DateField()
    move_out_date = models.DateField()
    building_code = models.CharField(max_length=64)
    status = models.CharField(
        choices=StatusType.choices, max_length=max(len(x) for x in StatusType.values), null=True
    )
    action_type = models.CharField(
        choices=ClearanceType.choices,
        max_length=max(len(x) for x in ClearanceType.values),
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name = "Assign/Revoke Tracker"
        verbose_name_plural = "Assign/Revoke Tracker"
        indexes = [
            models.Index(fields=["campus_id", "building_code", "action_type"]),
        ]

    def __str__(self):
        return f"{self.campus_id}: {self.action_type}"

    def match_peoplesoft_list(self, ps_list) -> bool:
        match = (
            self.campus_id,
            self.building_code,
            str(self.move_in_date),
            str(self.move_out_date),
        )
        for ps in ps_list:
            against = (
                ps.get("campus_id"),
                ps.get("room_second_description"),
                f'{parse(ps.get("move_in_date")).date()}',
                f'{parse(ps.get("move_out_date")).date()}',
            )
            if match == against:
                return True
        return False


class Clearance(TimeStampedModel):
    name = models.CharField(max_length=128)
    object_id = models.IntegerField()

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = "Clearance"
        verbose_name_plural = "Clearances"
        unique_together = ("object_id", "name")
        indexes = [
            models.Index(fields=["object_id", "name"]),
        ]


class RoomUseCode(TimeStampedModel):
    name = models.CharField(max_length=128)
    clearances = models.ManyToManyField(Clearance)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = "Room Use Code"
        verbose_name_plural = "Room Use Codes"


class UnmanagedBuilding(TimeStampedModel):
    prefix = models.CharField(max_length=30)

    class Meta:
        verbose_name = "Unmanaged Building"
        verbose_name_plural = "Unmanaged Buildings"


class ARLog(TimeStampedModel):
    campus_id = models.CharField(max_length=9)
    action_date = models.DateField(null=True)
    action_type = models.CharField(
        choices=ClearanceType.choices,
        max_length=max(len(x) for x in ClearanceType.values),
        null=True,
        blank=True,
    )
    peoplesoft_action_dttm = models.DateTimeField(null=True)
    ruc = models.ForeignKey(RoomUseCode, on_delete=models.PROTECT)
    action_success = models.BooleanField(null=True)
    action_message = models.CharField(
        choices=LogMessageType.choices,
        max_length=max(len(x) for x in LogMessageType.values),
        null=True,
        blank=True,
    )
    action_error_information = models.TextField(blank=True, null=True)
    req_status = models.CharField(max_length=10, null=True)

    def __str__(self):
        return f"{self.campus_id}::{self.action_type} - {self.action_date}"

    class Meta:
        indexes = [models.Index(fields=["campus_id"])]
        constraints = [
            models.UniqueConstraint(
                fields=["campus_id", "action_type", "ruc"], name="unique action"
            )
        ]
        verbose_name = "Housing Log"
