import time
from datetime import datetime
from typing import List

import dateutil.parser
import requests
from celery import shared_task
from django.conf import settings
from sat.db import ConnectionType, get_db_connection
from sat.logs import SATLogger

from sat_automations.housing_automation.config import (
    PeopleSoftProxyProvider,
)
from sat_automations.housing_automation.models import (
    CCURE_COLUMN_HEADERS,
    ARLog,
    CCurePersonnel,
    Clearance,
    ClearanceType,
    LogMessageType,
    RoomUseCode,
    StatusType,
    UnmanagedBuilding,
)
from sat_automations.manage_automations.models import SlackChannel
from sat_automations.utils.authenticate import GoogleAuthenticate
from sat_automations.utils.common.connection import CCURE_CONNECTION
from sat_automations.utils.common.queries import CCURE_ALL_IDS_QUERY
from sat_automations.utils.common.time import datetimestr_as_datetime
from sat_automations.utils.managers import BulkCreateManager

logger = SATLogger(__name__)

DT_FORMAT = "%Y-%m-%dT%H:%M:%S"


class HousingAssignRevoke:
    def __init__(self) -> None:
        self.slack = SlackChannel.objects.filter(channel_name="housing").first()
        self.bulk_create_manager = BulkCreateManager(
            update_fields=[
                "peoplesoft_action_dttm",
                "action_success",
                "action_message",
                "action_date",
                "action_error_information",
                "req_status",
            ]
        )
        self.authenticator = None
        self.access_token = None
        self.people_soft_proxy = None
        self.unmanaged_buildings = list(
            UnmanagedBuilding.objects.all().values_list("prefix", flat=True)
        )
        self.populate_errors = {}
        self.ccure_connection = get_db_connection(ConnectionType.SQL, conn_string=CCURE_CONNECTION)

    def init_connections(self):
        self.authenticator = GoogleAuthenticate("housing")
        self.authenticator.authenticate()
        if self.authenticator.response_json:
            self.access_token = self.authenticator.response_json.get("token")
        else:
            logger.error(
                f"Unable to authenticate with auth service. Error: {self.authenticator.error_message}"
            )
            raise ConnectionError("Unable to authenticate with auth service.")
        # When this is in a test environment it will give us test data rather than nothing.
        self.people_soft_proxy = PeopleSoftProxyProvider(access_token=self.access_token)

    def store_peoplesoft_records(self, lookbehind=90):
        # Queries the PeopleSoft API for housing that are marked either ASSIGN or REVOKE and
        # writes them to the database, for the other steps of the automation to process.
        housing_automation_list = self.people_soft_proxy.get_housing(days=lookbehind)
        if not housing_automation_list:
            logger.info("No records found")
        else:
            logger.info(f"Storing {len(housing_automation_list)} records.")
            self.store_actions(housing_automation_list)

    def _save_action(self, individual: dict, ruc: RoomUseCode):
        campus_id = individual["campus_id"]
        move_out_date = dateutil.parser.parse(individual["move_out_date"]).date()
        move_in_date = dateutil.parser.parse(individual["move_in_date"]).date()
        action_added_peoplesoft = individual["action_dttm"]
        req_status = individual["DESCR50"]
        move_in_log, mi_created = ARLog.objects.get_or_create(
            campus_id=campus_id,
            action_type=ClearanceType.ASSIGN,
            ruc=ruc,
        )
        move_out_log, mo_created = ARLog.objects.get_or_create(
            campus_id=campus_id,
            action_type=ClearanceType.REVOKE,
            ruc=ruc,
        )
        move_out_log.action_date = move_out_date
        move_in_log.action_date = move_in_date
        move_out_log.req_status = req_status
        move_in_log.req_status = req_status
        move_in_log.peoplesoft_action_dttm = datetimestr_as_datetime(
            action_added_peoplesoft, dt_format=DT_FORMAT
        )
        move_out_log.peoplesoft_action_dttm = datetimestr_as_datetime(
            action_added_peoplesoft, dt_format=DT_FORMAT
        )
        self.bulk_create_manager.update(move_in_log)
        self.bulk_create_manager.update(move_out_log)

    def store_actions(self, housing_list: list):
        logger.debug(f"Storing {len(housing_list)} records.")
        for each_individual in housing_list:
            try:
                ruc = each_individual["room_second_description"]
            except KeyError:
                logger.error(
                    f"cid: {each_individual.get('campud_id', 'campus_id_not_present')}: This person doesn't have a ruc field defined at this time. Skipping to next person"
                )
                continue
            if ruc.startswith(tuple(self.unmanaged_buildings)):
                logger.info(f"Skipping unmanaged room: {ruc}")
                continue
            if room_use_code := RoomUseCode.objects.filter(name=ruc).first():
                self._save_action(each_individual, room_use_code)
            else:
                logger.error(f"The room use code ({ruc}) does not exist.")
                logger.info(f"Creating a new Room Use Code: {ruc}")
                new_ruc = RoomUseCode.objects.create(name=ruc)
                self._save_action(each_individual, new_ruc)

    def process_housing_management_action(self, action: ClearanceType):
        logger.debug(f"Starting {action} Housing Automation.")
        today = datetime.now()
        for action_item in ARLog.objects.filter(action_date=today, action_type=action):
            self.apply_clearance_action(action_item)
        else:
            logger.info(f"No records found in the Log for {action} Clearances.")

    def apply_action_in_ccure(self, individual: ARLog, clearance: Clearance):
        logger.debug(f"IN APPLY ACTION IN CCURE: {individual.campus_id}")
        try:
            clearance_name = clearance.name
            clearance_id = clearance.object_id
            response = self.api_request(individual.campus_id, clearance_id, individual.action_type)
            if response.status_code != 200:
                individual.action_success = False
                individual.action_message = LogMessageType.ASSIGNMENT_FAILED
                individual.action_error_information = response.text
                self.bulk_create_manager.update(individual)
                self.slack.send_message(
                    f"FAILED: {individual.action_type} housing: "
                    f"Campus ID: {individual.campus_id}, clearance {clearance_id}: {clearance_name}"
                    f"\nReceived {response.status_code} response from clearance service."
                )
                return

            response_data = response.json()
            if response_data.get("failed"):
                individual.status = StatusType.FAIL
                logger.info(
                    f"Failed Clearance Assignment Operation [ "
                    f" campus_id: {individual.campus_id} |  building room code:"
                    f" {individual.ruc.name} | clearance: {clearance_name}]:"
                    f" {response_data.get('failed').get(individual.campus_id)} ]:"
                    f" {response.text}"
                )
                self.slack.send_message(
                    f"FAILED: {individual.action_type} housing: "
                    f"Campus ID: {individual.campus_id}, clearance {clearance_id}: {clearance_name}"
                    f"\n{response.text}"
                )
                individual.action_success = False
                individual.action_message = LogMessageType.ASSIGNMENT_FAILED
                individual.action_error_information = response.text
            else:
                individual.action_success = True
                individual.action_message = LogMessageType.ASSIGNMENT_SUCCESS
                logger.info(
                    "Successful Clearance Operation [ "
                    f"campus_id: {individual.campus_id} | building room"
                    f" code: {individual.ruc.name} | clearance: {clearance_name} ]"
                )
        except TimeoutError as te:
            logger.info(
                f"Failed Clearance Assignment Operation [ campus_id:"
                f" {individual.campus_id} | building room code: {individual.ruc.name}] due to Timeout:"
                f" {str(te)} ]"
            )
            individual.action_success = False
            individual.action_message = LogMessageType.ASSIGNMENT_FAILED
            individual.action_error_information = str(te)
        except Exception as e:
            logger.info(
                f"Failed Operation [ campus_id:"
                f" {individual.campus_id} | building room code: {individual.ruc.name}] due to Exception:"
                f" {str(e)} ]"
            )
            individual.action_success = False
            individual.action_message = LogMessageType.ASSIGNMENT_FAILED
            individual.action_error_information = str(e)
        finally:
            self.bulk_create_manager.update(individual)

    def apply_clearance_action(self, individual: ARLog):
        logger.debug(f"Searching and Assigning/Revoking clearance for {individual.campus_id}")
        clearances = Clearance.objects.filter(roomusecode=individual.ruc)
        if clearances:
            for clearance in clearances:
                logger.info(
                    f"Processing {clearance.name} for {individual.campus_id} in {individual.ruc.name}"
                )
                self.apply_action_in_ccure(individual, clearance)
        else:
            individual.action_success = False
            individual.action_message = LogMessageType.BUILDING_CODE_NO_CLEARANCE
            individual.save()
            logger.info(
                f"No action taken for {individual.campus_id} due to no clearance associated with the"
                f" building code {individual.ruc.name}."
            )

    def api_request(self, individual_id: int, clearance_id: int, action_type: ClearanceType):
        response = requests.post(
            f"{settings.CLEARANCE_SERVICE_URL}/assignments/{action_type}",
            headers={"Authorization": f"Bearer {self.access_token}"},
            json={"assignees": [str(individual_id).zfill(9)], "clearance_ids": [clearance_id]},
            timeout=600,
        )
        return response

    def get_ccure_ids(self):
        try:
            ccure_cursor = self.ccure_connection.cursor()
            query = CCURE_ALL_IDS_QUERY
            results = ccure_cursor.execute(query).fetchall()
            return results
        except Exception as e:
            logger.error(f"Failed to fetch records due to {str(e)}")

    def add_active_students_to_ccure(
        self, ccure_ids: List[int], active_students: List[dict]
    ) -> dict:
        """
        Add active students not in ccure to the CCURE database
        :param active_students: List of active students missing in CCURE
        :return: Number of records successfully updated in CCURE
        """
        update_results = {
            "success": 0,
            "failed_ccure_update": 0,
        }
        logger.info("Updating CCURE with missing records")

        for student in active_students:
            if student["campus_id"] not in ccure_ids:
                personnel = CCurePersonnel(**student)
                personnel.cidc = f"{personnel.campus_id}:0"
                ccure_response = self.add_to_ccure(personnel)
                if ccure_response.status_code != 200:
                    failure_message = (
                        f"Failed to persist record with campus_id {student['campus_id']} "
                        f"using CCURE api. Error: {ccure_response.text}"
                    )
                    logger.info(failure_message)
                    update_results["failed_ccure_update"] += 1
                else:
                    update_results["success"] += 1

        return update_results

    def add_to_ccure(self, active_student: CCurePersonnel) -> requests.Response:
        values = list(active_student.model_dump().values())
        response = requests.post(
            f"{settings.CLEARANCE_SERVICE_URL}/personnel/persist",
            headers={"Authorization": f"Bearer {self.access_token}"},
            json={"property_names": CCURE_COLUMN_HEADERS, "property_values": values},
            timeout=600,
        )
        return response


@shared_task
def housing_populate_tracker(lookbehind=90):  # pragma: no cover
    logger.info("RUNNING: Housing Populate Tracker.")
    har = HousingAssignRevoke()
    try:
        har.init_connections()
        har.store_peoplesoft_records(lookbehind)
        har.bulk_create_manager.done()
        logger.info("COMPLETED: Housing Populate Tracker.")
        return har.populate_errors
    except Exception as ex:
        logger.error(f"FAILED: Housing Populate Assign: {ex}")
        har.slack.send_message(f"FAILED: Housing Populate Assign: {ex}")
        raise ex


@shared_task
def process_active_students_into_ccure():  # pragma: no cover
    logger.info("RUNNING: Move missing students into CCURE.")
    assign = HousingAssignRevoke()
    try:
        assign.init_connections()
        start = time.time()
        ccure_ids = [id[0] for id in assign.get_ccure_ids()]
        logger.info(f"Time taken to fetch ids from CCURE: {time.time() - start}")
        start = time.time()
        active_students = assign.people_soft_proxy.get_active_students()
        logger.info(f"Time taken to fetch active student records: {time.time() - start}")
        start = time.time()
        results = assign.add_active_students_to_ccure(ccure_ids, active_students)
        logger.info(f"Time taken to process students missing from CCURE: {time.time() - start}")
        assign.bulk_create_manager.done()
        logger.info("COMPLETE: Move missing students into CCURE.")
        logger.info(f"Results: {results}")
    except Exception as ex:
        logger.error(f"FAILED: Move missing students into CCURE: {ex}")
        assign.slack.send_message(f"FAILED: Move missing students into CCURE: {ex}")
        raise ex


@shared_task
def housing_process_move_ins():
    logger.info("RUNNING: Housing Process Move Ins.")
    har = HousingAssignRevoke()
    try:
        har.init_connections()
        har.process_housing_management_action(ClearanceType.ASSIGN)
        har.bulk_create_manager.done()
    except Exception as ex:
        logger.error(f"FAILED: Housing Process Move Ins: {ex}")
        har.slack.send_message(f"FAILED: Housing Process Move Ins: {ex}")
        raise ex
    logger.info("COMPLETE: Housing Process Move Ins.")


@shared_task
def housing_process_move_outs():  # pragma: no cover
    logger.info("RUNNING: Housing Process Move Outs.")
    har = HousingAssignRevoke()
    try:
        har.init_connections()
        har.process_housing_management_action(ClearanceType.REVOKE)
        har.bulk_create_manager.done()
    except Exception as ex:
        logger.error(f"FAILED: Housing Process Move Outs: {ex}")
        har.slack.send_message(f"FAILED: Housing Process Move Outs: {ex}")
        raise ex
    logger.info("COMPLETE: Housing Process Move Outs.")
