import random
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime

import pytest
from faker import Faker

from sat_automations.housing_automation.models import (
    ARLog,
    ClearanceType,
    LogMessageType,
)
from sat_automations.housing_automation.tasks import HousingAssignRevoke
from sat_automations.housing_automation.tests.fixtures import (
    PEOPLESOFT_RESPONSE_EXAMPLES,
    PEOPLESOFT_RESPONSE_EXAMPLES_PLUS_DUPS,
    get_positive_delta,
)

Faker.seed(random.randint(0, 1000))

fake = Faker()

pytestmark = pytest.mark.django_db


@dataclass
class CCurePersonnel:
    # the only field acslib requires is last_name
    last_name: str
    campus_id: str = ""
    first_name: str = ""
    middle_name: str = ""
    classification: str = ""
    email: str = ""

    def __eq__(self, other):
        return self.campus_id == other.campus_id

    def render_person(self):
        return {
            "campus_id": self.campus_id,
            "first_name": self.first_name,
            "middle_name": self.middle_name,
            "last_name": self.last_name,
            "classification": self.classification,
            "email": self.email,
        }


def build_ccure_ids(count=4):
    ids = []
    for _ in range(count):
        ids.append(str(fake.random_int(min=0, max=999999999)))  # campus_id)
    return ids


FIXED_CCURE_IDS = build_ccure_ids()


def build_student_records(ccure_ids: list[int] = ()) -> list[CCurePersonnel]:
    """
    By default, this function will return a list of 20 CCurePersonnel objects with
    campus ids that match the CCURE records.
    :param campus_ids:
    :param count:
    :return:
    """
    if not ccure_ids:
        ccure_ids = [ccure_id for ccure_id in FIXED_CCURE_IDS]
    students = []
    for id in ccure_ids:
        student = CCurePersonnel(
            campus_id=id,
            first_name=fake.first_name(),
            middle_name=fake.random_letter().upper(),
            last_name=fake.last_name(),
            classification=random.choice(("10", "20", "30", "40", "SP", "GR", "UN")),
            email=fake.email(),
        )
        students.append(student)
    return students


MATCHED_STUDENT_RECORDS = build_student_records()


@pytest.fixture
def standard_patch(
    mocker, patch_google_authenticate, patch_ccure_connection, patch_peoplesoft_provider
):
    patch_google_authenticate("housing_automation")
    patch_ccure_connection("housing_automation")
    patch_peoplesoft_provider("housing_automation")


@pytest.fixture(autouse=True)
def use_test_peoplesoft_proxy(settings):
    settings.PEOPLESOFT_PROXY_URL = "TEST"


def test_housing_ar_store_no_records(caplog, standard_patch):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = []
    har.process_housing_management_action(ClearanceType.ASSIGN)
    assert (
        f"No records found in the Log for {ClearanceType.ASSIGN.value} Clearances." in caplog.text
    )


def test_housing_ar_store_all(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.all().count() == 8


def test_housing_ar_store_assign(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.ASSIGN).count() == 4


def test_with_duplicate_assign_no_change_in_ruc(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.ASSIGN).count() == 4
    modified = deepcopy(PEOPLESOFT_RESPONSE_EXAMPLES_PLUS_DUPS)
    modified[-1]["move_in_date"] = get_positive_delta()
    har.people_soft_proxy.get_housing.return_value = modified
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    # We should still get 4 even though there has been a date change for the assignment in one of the
    # records.
    assert ARLog.objects.filter(action_type=ClearanceType.ASSIGN).count() == 4


def test_with_duplicate_assign_change_in_ruc(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.ASSIGN).count() == 4
    modified = deepcopy(PEOPLESOFT_RESPONSE_EXAMPLES_PLUS_DUPS)
    modified[-1]["room_second_description"] = "BOWEN"
    har.people_soft_proxy.get_housing.return_value = modified
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.ASSIGN).count() == 5


def test_housing_ar_store_revoke(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.REVOKE).count() == 4


def test_housing_ar_store_revoke_no_change_in_ruc(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.REVOKE).count() == 4
    modified = deepcopy(PEOPLESOFT_RESPONSE_EXAMPLES_PLUS_DUPS)
    modified[-2]["move_out_date"] = get_positive_delta(5)
    har.people_soft_proxy.get_housing.return_value = modified
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.REVOKE).count() == 4


def test_housing_ar_store_revoke_change_in_ruc(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.REVOKE).count() == 4
    modified = deepcopy(PEOPLESOFT_RESPONSE_EXAMPLES_PLUS_DUPS)
    modified[-2]["room_second_description"] = "BOWEN"
    har.people_soft_proxy.get_housing.return_value = modified
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert ARLog.objects.filter(action_type=ClearanceType.REVOKE).count() == 5


def test_housing_ar_store_full_dups(standard_patch, setup_rucs):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.store_peoplesoft_records()
    # Same list same number of objects.
    assert ARLog.objects.all().count() == 8


def test_housing_ar_date_only(standard_patch, setup_rucs):
    import datetime

    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    assert type(ARLog.objects.first().action_date) == datetime.date


def test_housing_assign_revoke_today(standard_patch, setup_rucs):
    today = datetime.now().date()
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    queryset = ARLog.objects.filter(action_date=today, action_type=ClearanceType.REVOKE)
    assert queryset.count() == 1


def test_housing_data_automation_no_revoke(standard_patch, caplog):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = []
    har.process_housing_management_action(ClearanceType.REVOKE)
    assert f"No records found in the Log for {ClearanceType.REVOKE} Clearances." in caplog.text


def test_housing_data_automation_assign(standard_patch, setup_rucs, mocker):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    har.apply_clearance_action = mocker.MagicMock()
    har.apply_clearance_action.return_value = True
    har.process_housing_management_action(ClearanceType.ASSIGN)
    assert har.apply_clearance_action.call_count == 3


class MockRequestResponse:
    def __init__(self, status_code: int, success: bool):
        self.status_code = status_code
        self.success = success

    def json(self):
        return {
            "succeeded": {"123456789": {1234: True}} if self.success else {},
            "failed": {"123456789": {1234: "Didn't work."}} if not self.success else {},
            "pending": [],
        }


def test_housing_data_automation_assign_server_error(
    standard_patch, setup_rucs, monkeypatch, mocker
):
    monkeypatch.setattr(HousingAssignRevoke, "api_request", MockRequestResponse(500, False))
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    har.apply_clearance_action = mocker.MagicMock()
    har.apply_clearance_action.return_value = True
    har.process_housing_management_action(ClearanceType.ASSIGN)
    assert har.apply_clearance_action.call_count == 3
    records = ARLog.objects.all()
    assert not any(record.action_success for record in records)


def test_housing_data_automation_assign_failures(standard_patch, setup_rucs, monkeypatch, mocker):
    monkeypatch.setattr(HousingAssignRevoke, "api_request", MockRequestResponse(200, False))
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    har.apply_clearance_action = mocker.MagicMock()
    har.apply_clearance_action.return_value = True
    har.process_housing_management_action(ClearanceType.ASSIGN)
    assert har.apply_clearance_action.call_count == 3
    records = ARLog.objects.all()
    assert not any(record.action_success for record in records)


def test_housing_data_automation_revoke(standard_patch, setup_rucs, mocker):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = PEOPLESOFT_RESPONSE_EXAMPLES
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    har.apply_clearance_action = mocker.MagicMock()
    har.apply_clearance_action.return_value = True
    har.process_housing_management_action(ClearanceType.REVOKE)
    assert har.apply_clearance_action.call_count == 1


def test_housing_data_automation_no_ruc(standard_patch, caplog):
    har = HousingAssignRevoke()
    har.init_connections()
    ps_response = PEOPLESOFT_RESPONSE_EXAMPLES.copy()
    # Remove the first 3 responses
    for i in range(3):
        ps_response.pop(0)
    har.people_soft_proxy.get_housing.return_value = [ps_response[0]]
    har.people_soft_proxy.get_person.return_value = [ps_response[0]]
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    har.process_housing_management_action(ClearanceType.REVOKE)
    assert "The room use code (Grk Vlg-2M-335) does not exist." in caplog.text
    assert f"No records found in the Log for {ClearanceType.REVOKE} Clearances." in caplog.text
    assert "Creating a new Room Use Code: Grk Vlg-2M-335"
    har.bulk_create_manager.done()
    arlogs = ARLog.objects.all()
    # A single response will have 2 entries 1 revoke 1 assign
    assert arlogs.count() == 2


def test_housing_data_automation_no_clearance(standard_patch, caplog):
    har = HousingAssignRevoke()
    har.init_connections()
    har.people_soft_proxy.get_housing.return_value = [PEOPLESOFT_RESPONSE_EXAMPLES[2]]
    har.people_soft_proxy.get_person.return_value = [PEOPLESOFT_RESPONSE_EXAMPLES[2]]
    har.store_peoplesoft_records()
    har.bulk_create_manager.done()
    har.process_housing_management_action(ClearanceType.ASSIGN)
    assert (
        "No action taken for 200536406 due to no clearance associated with the building code ABC123"
        in caplog.text
    )
    har.bulk_create_manager.done()
    arl = ARLog.objects.filter(action_type=ClearanceType.ASSIGN).first()
    assert arl.action_message == LogMessageType.BUILDING_CODE_NO_CLEARANCE
    assert not arl.action_success


# TODO: Add tests for the apply_action_in_ccure method


def test_missing_active_students(mocker, standard_patch):
    mocker.patch("sat_automations.housing_automation.tasks.HousingAssignRevoke.add_to_ccure")
    har = HousingAssignRevoke()
    har.add_to_ccure.return_value = mocker.Mock(status_code=200)
    ccure = FIXED_CCURE_IDS
    students = [x.render_person() for x in MATCHED_STUDENT_RECORDS]
    # remove a student so that only one is missing from CCURE
    ccure.pop(0)
    results = har.add_active_students_to_ccure(ccure, students)
    assert results["success"] == 1
