import base64
import io
import time
from datetime import datetime
from pathlib import Path
from typing import Generator
from urllib.parse import quote_plus

from acslib.ccure import CcureAPI
from acslib.ccure.filters import FUZZ, PersonnelFilter
from celery import shared_task
from django.conf import settings
from PIL import Image
from sat.db import ConnectionType, SatDBException, get_db_connection
from sat.logs import SATLogger

from sat_automations.manage_automations.models import SlackChannel
from sat_automations.utils.common.connection import (
    CCURE_CONNECTION,
    GOLD_CONNECTION,
    get_acslib_connection,
)
from sat_automations.utils.common.queries import (
    CCURE_ACTIVE_PERSONNEL_NO_IMAGE,
    GOLD_IMAGE_ENTRY_COUNT,
    get_image_for_id,
    gold_image_all_select,
    return_last_image_update_query,
)

logger = SATLogger(__name__)

personnel_lookup = PersonnelFilter(lookups={"Text1": FUZZ})


class ImageUpdate:
    def __init__(self):
        try:
            self.gold_db_conn = get_db_connection(
                ConnectionType.PY_ORACLE, conn_dict=GOLD_CONNECTION
            )
            self.ccure_connection = get_db_connection(
                ConnectionType.SQL, conn_string=CCURE_CONNECTION
            )
            self.acslib_connection = CcureAPI(get_acslib_connection())
        except SatDBException as error:
            logger.error(f"{error}")
            raise ConnectionError(f"Unable to connect to database: {error}")
        self.slack = SlackChannel.objects.filter(channel_name="image_update").first()
        self.total_images = 0
        if not settings.IMAGE_VOLUME_BASE_PATH:
            raise ValueError("IMAGE_VOLUME_PATH not set in settings.")
        self.volume_path = Path(settings.IMAGE_VOLUME_BASE_PATH)

    def _get_recently_updated_images(self) -> list[tuple[int, io.BytesIO, datetime]]:
        """Get images from gold that have been updated in the last 24 hours"""
        cursor = self.gold_db_conn.cursor()
        query_results = cursor.execute(return_last_image_update_query())
        return query_results.fetchall()

    def get_ccure_personnel_w_no_images(self) -> list[str]:
        cursor = self.ccure_connection.cursor()
        query_results = cursor.execute(CCURE_ACTIVE_PERSONNEL_NO_IMAGE)
        return [x[0] for x in query_results.fetchall()]

    def _get_all_images(self, batch_size=5000) -> Generator[int, io.BytesIO, datetime]:
        """Get all images from gold"""
        cursor = self.gold_db_conn.cursor()
        with cursor.execute(GOLD_IMAGE_ENTRY_COUNT) as count_res:
            self.total_images = count_res.fetchone()[0]

        cursor = self.gold_db_conn.cursor()
        with cursor.execute(gold_image_all_select()):
            while self.total_images > 0:
                logger.info(f"Remaining images: {self.total_images}")
                results = cursor.fetchmany(batch_size)
                if len(results) == 0:
                    return
                self.total_images -= len(results)
                yield results

    def process_all_images(self):
        """Process all images from gold"""
        for images in self._get_all_images():
            start = time.time()
            self._store_images(images)
            logger.info(f"Time to store {len(images)}: {time.time() - start}")
        logger.info("All images processed")

    def set_empties(self):
        if res := self.get_ccure_personnel_w_no_images():
            cursor = self.gold_db_conn.cursor()
            for campus_id in res:
                query = get_image_for_id(campus_id)
                result = cursor.execute(query).fetchall()
                if result:
                    self._store_images(result)
        else:
            logger.info("No Personnel without images.")

    def _store_images(self, records: list[tuple[int, io.BytesIO, datetime]]):
        """Take a list of results from the gold query and store the images on disc.
        If the image already exists, it will not be overwritten.
        """
        for record in records:
            if not all(record):  # if any value in the record is missing or null:
                logger.info(f"Missing information from record {record}, could not store image.")
                continue
            try:
                campus_id, image, last_update = record
                logger.info(f"Storing image for: {campus_id}")
                filename = f"{last_update.strftime('%Y-%m-%d_%H-%M-%S')}.jpg"
                person_path = self.volume_path / str(campus_id)
                if not person_path.exists():
                    try:
                        person_path.mkdir(mode=0o777, parents=True)
                    except PermissionError:
                        logger.info(f"Could not create path for user {campus_id}")
                        continue
                image_path = person_path / filename
                if not image_path.exists():
                    buffer = io.BytesIO()
                    binary_image = image.read()
                    jpg_image = Image.open(io.BytesIO(binary_image))
                    jpg_image.save(image_path)
                    jpg_image.save(buffer, format="JPEG")
                    self._add_image_to_ccure_records([(campus_id, buffer, image_path)])
            except Exception as e:
                logger.info(f"Could not save image for {campus_id}: {e}")

    def _add_image_to_ccure_records(self, records: list):
        """
        Takes a list of tuples with the shape
        0: campus_id: int
        1: image_buffer: BytesIO
        2: image_path: Path

        The image_buffer should be encoded as a JPEG
        """
        for campus_id, image_buffer, image_path in records:
            if personnel_record := self.acslib_connection.personnel.search(
                [f"{campus_id}"], search_filter=personnel_lookup
            ):
                object_id = personnel_record[0].get("ObjectID")
                encoded_image = quote_plus(base64.b64encode(image_buffer.getvalue()).decode())
                logger.info(f"Setting Primary Portrait for: {object_id}::{campus_id}")
                try:
                    response = self.acslib_connection.action.personnel.add_image(
                        personnel_id=object_id,
                        image=encoded_image,
                        image_name=f"{campus_id}_{image_path.name}",
                    )
                    return response
                except Exception as e:
                    logger.error(f"Error saving image to CCURE for campus_id {campus_id}: {e}")
            else:
                logger.info(f"Can't find a record for: {campus_id}")

    def update_images(self):
        """
        1. Get person images updated in the last 24 hours
        2. Store the new images on disc
        3. Add those images to the personnel records in CCure
        """
        # get images updated in the last 24 hours
        logger.info("Getting recently updated images from gold")
        records = self._get_recently_updated_images()
        logger.info(f"Storing {len(records)} recently updated images from gold")

        # save images to the volume
        self._store_images(records)


@shared_task
def daily_image_update():
    logger.info("RUNNING: Image Update Automation.")
    try:
        image_update = ImageUpdate()
        image_update.update_images()
        logger.info("COMPLETE: Image Update automation")
    except Exception as e:
        logger.error(f"FAILED: Image Update automation: {e}")
        image_update.slack.send_message(f"FAILED: Image_Update automation: {e}")
        raise e


@shared_task
def all_images():
    logger.info("RUNNING: Image Update Automation.")
    try:
        image_update = ImageUpdate()
        image_update.process_all_images()
        logger.info("COMPLETE: Image Update automation")
    except Exception as e:
        logger.error(f"FAILED: Image Update automation: {e}")
        image_update.slack.send_message(f"FAILED: Image_Update automation: {e}")
        raise e


@shared_task
def set_empty_portraits():
    logger.info("RUNNING: Set Empty Portraits")
    try:
        image_update = ImageUpdate()
        image_update.set_empties()
        logger.info("COMPLETE: Set Empty Portraits")
    except Exception as e:
        logger.error(f"FAILED: Set Empty Portraits: {e}")
        try:
            image_update.slack.send_message(f"FAILED: Set Empty Portraits: {e}")
        except UnboundLocalError:
            pass
        raise e
