import enum
from datetime import datetime, timedelta
from typing import List, Union
from zoneinfo import ZoneInfo

TupleTime = Union[tuple[datetime, datetime], tuple[str, str]]
SplitTime = List[TupleTime]

DATETIME_FMT_MICRO = "%Y-%m-%d %H:%M:%S.%f"
DATETIME_FMT = "%Y-%m-%d %H:%M:%S"
DATE_FMT = "%Y-%m-%d"
HUMAN_FMT = "%B %d, %Y"

UTC = ZoneInfo("UTC")
# Eastern Prevailing Time
# Takes into account DST when converting from or to UTC.
EPT = ZoneInfo("US/Eastern")


class SATTimeZones(enum.Enum):
    UTC = UTC
    EPT = EPT


def bound_time(
    dt: datetime, nearest_day=False, nearest_hour=False, nearest_minute=False, nearest_second=False
) -> datetime:
    """
    Limit a datetime to nearest hour, minute, or second.
    Default is to return the end of the day.
    Nearest day will return the start of the day.
    Nearest hour will return the start of the hour.
    Nearest minute will return the start of the minute.
    Nearest second will return the start of the second. I doubt this will get much use, but it's here for completeness.
    :param dt: datetime
    :param nearest_day: bool
    :param nearest_hour: bool
    :param nearest_minute: bool
    :param nearest_second: bool
    :return: datetime
    """
    if nearest_day:
        return datetime.replace(dt, hour=0, minute=0, second=0, microsecond=0)
    if nearest_hour:
        return datetime.replace(dt, minute=0, second=0, microsecond=0)
    if nearest_minute:
        return datetime.replace(dt, second=0, microsecond=0)
    if nearest_second:
        return datetime.replace(dt, microsecond=0)
    return datetime.replace(dt, hour=23, minute=59, second=59, microsecond=999999)


def date_as_datetime(dt: str, tz: SATTimeZones = SATTimeZones.UTC) -> datetime:
    """
    Convert a date string to a datetime object.
    :param dt: str(YYYY-MM-DD)
    :param tz: SATTimeZones
    :return: datetime
    """
    return datetime.strptime(dt, DATE_FMT).replace(tzinfo=tz.value)


def datetimestr_as_datetime(
    dt: str, tz: SATTimeZones = SATTimeZones.UTC, dt_format: str = DATETIME_FMT
) -> datetime:
    """
    Convert a datetime string to a datetime object.
    :param dt: str(YYYY-MM-DD HH:MM:SS)
    :param tz: SATTimeZones
    :param dt_format: str
    :return: datetime
    """
    return datetime.strptime(dt, dt_format).replace(tzinfo=tz.value)


def datetime_as_datestr(dt: datetime) -> str:
    """
    Convert a datetime object to a date string.
    :param dt: datetime
    :return: str
    """
    return dt.strftime(DATE_FMT)


def datetime_as_timestr(dt: datetime, trunc_micro=0) -> str:
    """
    Convert a datetime object to a time string.
    :param dt: datetime
    :param trunc_micro: int
        trunc_micro will truncate the microseconds to the specified number. This is useful for MSSQL, which expects a 3 digit
        microsecond.
    :return: str
    """
    if trunc_micro:
        return dt.strftime(DATETIME_FMT_MICRO)[:-trunc_micro]
    return dt.strftime(DATETIME_FMT_MICRO)


def yield_times(start_time: datetime, end_time: datetime, incrementor: int = 1) -> SplitTime:
    """
    For ease of reasoning and programming filters we will assume incoming datetimes are in UTC.
    Yield a range of dates beginning at start datetime by set number of hours.
    :param start_time: datetime
    :param end_time: datetime
    :param incrementor: int
    :return: str, str
    """

    left_dt = start_time
    first_loop = True
    while left_dt < end_time:
        right_dt = left_dt + timedelta(hours=incrementor)
        if first_loop:
            right_dt = bound_time(right_dt, nearest_hour=True)
            first_loop = False
        yield left_dt.strftime(DATETIME_FMT_MICRO), right_dt.strftime(DATETIME_FMT_MICRO)
        left_dt = right_dt
