from collections import defaultdict

from django.apps import apps
from sat.logs import SATLogger

logger = SATLogger(__name__)


class BulkCreateManager:
    """
    This helper class keeps track of ORM objects to be created for multiple
    model classes, and automatically creates those objects with `bulk_create`
    when the number of objects accumulated for a given model class exceeds
    `chunk_size`.

    Upon completion of the loop that's `add()`ing objects, the developer must
    call `done()` to ensure the final set of objects is created for all models.
    """

    def __init__(self, chunk_size=100, update_fields=None):
        self._create_queues = defaultdict(list)
        self._update_queues = defaultdict(list)
        self.chunk_size = chunk_size
        self.update_fields = update_fields

    def _commit_add(self, model_class):
        logger.debug(
            f"Committing {len(self._create_queues[model_class._meta.label])} {model_class._meta.label} objects"
        )
        logger.debug(f"Last object: {self._create_queues[model_class._meta.label][-1]}")
        model_key = model_class._meta.label
        model_class.objects.bulk_create(self._create_queues[model_key])
        self._create_queues[model_key] = []

    def _commit_update(self, model_class):
        logger.debug(
            f"Committing {len(self._update_queues[model_class._meta.label])} {model_class._meta.label} objects"
        )
        model_key = model_class._meta.label
        model_class.objects.bulk_update(self._update_queues[model_key], self.update_fields)
        self._update_queues[model_key] = []

    def add(self, obj):
        """
        Add an object to the queue to be created, and call bulk_create if we
        have enough objs.
        """
        model_class = type(obj)
        model_key = model_class._meta.label
        self._create_queues[model_key].append(obj)
        if len(self._create_queues[model_key]) >= self.chunk_size:
            self._commit_add(model_class)

    def update(self, obj):
        """
        Add an object to the queue to be updated, and call bulk_update if we
        have enough objs.
        :param obj:
        :return:
        """
        model_class = type(obj)
        model_key = model_class._meta.label
        self._update_queues[model_key].append(obj)
        if len(self._update_queues[model_key]) >= self.chunk_size:
            self._commit_update(model_class)

    def done(self):
        """
        Always call this upon completion to make sure the final partial chunk
        is saved.
        """
        for model_name, objs in self._create_queues.items():
            if len(objs) > 0:
                self._commit_add(apps.get_model(model_name))

        for model_name, objs in self._update_queues.items():
            if len(objs) > 0:
                self._commit_update(apps.get_model(model_name))
