import csv
from datetime import datetime

from django.contrib import admin
from django.http import StreamingHttpResponse


class Echo:
    """An object that implements just the write method of the file-like interface."""

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value


def iter_items(items, pseudo_buffer, headers):
    writer = csv.DictWriter(pseudo_buffer, fieldnames=headers)
    yield writer.writeheader()

    for value in items.values():
        yield writer.writerow(value)


class ExportCsvMixin(admin.ModelAdmin):
    """Mixin for exporting to CSV."""

    actions = ["export_as_csv"]

    @admin.action(description="Export as CSV")
    def export_as_csv(self, request, queryset):
        """Export filtered records as CSV."""
        filename = f"{self.model._meta.verbose_name_plural.replace(' ', '_').lower()}_export_{datetime.now()}.csv"
        fieldnames = [x.name for x in self.model._meta.fields]
        return StreamingHttpResponse(
            streaming_content=(iter_items(queryset, Echo(), fieldnames)),
            content_type="text/csv",
            headers={"Content-Disposition": f'attachment; filename="{filename}"'},
        )
